#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include <float.h>                      // DBL_MAX DBL_MIN
#include <stdlib.h>                     // atoi()
#include "..\types\types.h"             // Bit32 Bit8
#include "..\pci\pci.h"                 // pci_get_device_handle() pci_get_base_address_n()
#include "..\dpmi32\phys.h"             // map_physical()
#include "..\intel\timestmp.h"          // class PentiumClock
#include "..\apptel\point.h"            // class AtmCard



// manifest constants
//
#define DOWNLOAD_BUFFER_SIZE            (512*400)
#define MAX_START_TIME                  0.5             // 1/2 S
#define MAX_VERIFY_TIME                 0.005           // 5  mS (1 mS too fast for PPro200 1/15/98)
#define MAX_DONE_TIME                   0.5             // 1/2 S


// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define SEGREGATE_BYTES                 FALSE           // TRUE screws up at whitespace!



// global variables
//
Bit8 buffer[ DOWNLOAD_BUFFER_SIZE ];



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *card_index_string = argv[ 2 ];
    FILE *in_handle;
    size_t result;
    Bit32 byte_count = 0;
    Bit32 byte_offset;
    Bit8 program_byte;
    Bit8 readback_byte;
    Bit16 card_index;
    PentiumClock start;
    PentiumClock now;
    double elapsed;
    double verify_min = DBL_MAX;
    double verify_max = DBL_MIN;
    double verify_total = 0.0;
    AtmCard *card;
    CpldControl1_register *cpld;
    union { CpldControl1_register cpld_temp; Bit16 cpld_temp_alias; };
    Bit8 *tx_fpga;

    if( argc != 3 )
    {
        fprintf( stderr, "downloads a binary FPGA image to the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s infile cardindex\n", program );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

    // convert the card index from string to binary
    //
    card_index = atoi( card_index_string );

    card = new AtmCard( card_index );
    if( !card )
        return 3;
    if( card->init_result )
        return card->init_result;
    cpld = &card->aim_and_cpld->Cpld.Control1;
    tx_fpga = &card->both_fpgas->Transmit.Download;

    // assert the FPGA program line, then wait
    //
    cpld_temp_alias = 0;
    cpld_temp.ProgramNot = 0;
    *cpld = cpld_temp;
    PentiumClock::wait( CLOCK_RATE * MAX_START_TIME );

    // make sure the program line is asserted (low),
    // so we know the jumpers are set properly on the card
    //
    if( cpld->ProgramNot )
    {
        fprintf( stderr, "cannot assert program line; jumpers are probably wrong\n" );
        return 8;   // should this be a different number than program line deassert failure???
    }

    // deassert the FPGA program line, then wait
    //
    cpld_temp.ProgramNot = 1;
    *cpld = cpld_temp;
    PentiumClock::wait( CLOCK_RATE * MAX_START_TIME );

    // make sure the program line is deasserted (high),
    // so we know the jumpers are set properly on the card
    //
    if( !cpld->ProgramNot )
    {
        fprintf( stderr, "cannot deassert program line; jumpers are probably wrong\n" );
        return 8;
    }

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        // read one buffer's worth of data from file
        //
        result = fread( buffer, 1, DOWNLOAD_BUFFER_SIZE, in_handle );
        if( result < DOWNLOAD_BUFFER_SIZE )
            if( ferror( in_handle ) )
            {
                fprintf( stderr, "error %d reading input file\n", errno );
                break;
            }

        // for all bytes in the buffer...
        //
        for( byte_offset=0; byte_offset<result; byte_offset++ )
        {
            program_byte = buffer[ byte_offset ];

            // output the byte to the TX FPGA
            //
            *tx_fpga = program_byte;

            // remember what time we started verifying
            //
            start.set();

            do
            {
                // read byte from TX FPGA
                //
                readback_byte = *tx_fpga;

                // check that bits 0-6 were high
                //
                if( (readback_byte & 0x7F) != 0x7F )
                {
                    fprintf( stderr, "after writing %dth value 0x%.2X,"
                        " low 6 bits of TX FPGA were 0x%.2X, not 0x7F\n",
                        byte_count + byte_offset + 1,
                        readback_byte & 0x7F
                    );

                    return 9;
                }

                // get the current time
                //
                now.set();

                // calculate the elapsed time in seconds
                //
                elapsed = PentiumClock( now - start ).get_seconds();

            // while bit 7 was read as 0
            // and not too much time has elapsed
            //
            } while( (readback_byte & 0x80) == 0
                 && elapsed < MAX_VERIFY_TIME );

            // check that timeout didn't occur
            //
            if( elapsed >= MAX_VERIFY_TIME )
            {
                fprintf( stderr, "waited more than %g seconds for FPGA to"
                    " accept %dth value 0x%.2X\n",
                    elapsed,
                    byte_count + byte_offset + 1,
                    program_byte
                );

                return 10;
            }

            // keep track of the verify timing
            //
            verify_total += elapsed;
            if( verify_min > elapsed )
                verify_min = elapsed;
            if( verify_max < elapsed )
                verify_max = elapsed;

            // check that config_error line is deasserted (high)
            //
            if( !cpld->ConfigErrorNot )
            {
                fprintf( stderr, "config error line was asserted (low) after"
                    " %dth value 0x%.2X was verified\n",
                    byte_count + byte_offset + 1,
                    program_byte
                );

                return 11;
            }
        }

        byte_count += result;

        // show some status to the user
        //
        fprintf( stderr, "%lu ", byte_count );
    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
    else
#endif
    {
        // wait for done signal to be asserted (high)
        // or for config error signal to be asserted (low)
        // or for too much time to elapse
        //
        start.set();
        do
        {
            // get the DONE and CONFIG ERROR lines
            //
            cpld_temp = *cpld;

            // get the current time
            //
            now.set();

            // calculate the elapsed time in seconds
            //
            elapsed = PentiumClock( now - start ).get_seconds();

        // while bit 7 was read as 0
        // and not too much time has elapsed
        //
        } while( cpld_temp.ConfigErrorNot
             && !cpld_temp.Done
             && elapsed < MAX_DONE_TIME );

        // check that timeout did not occur
        //
        if( elapsed >= MAX_DONE_TIME )
        {
            fprintf( stderr, "error!  waited more than %g seconds for done line to be asserted\n", MAX_DONE_TIME );
            return 12;
        }

        // check that error did not occur
        //
        if( !cpld_temp.ConfigErrorNot )
        {
            fprintf( stderr, "configuration error reported after all %d bytes transferred to FPGA's\n", byte_count );
            return 13;
        }
    }

    // show statistics
    //
    // 105538 bytes for 2 Xilinx 4025E FPGA's
    // (346 bits per frame per FPGA) in series
    //
    fprintf( stdout, "%lu bytes output\n", byte_count );

    // 4/2/97 1:42PM saw 5.86 microseconds average
    //
    fprintf( stdout, "verify took min/max/avg %.3f/%.3f/%.3f uS\n",
        1000000*verify_min,
        1000000*verify_max,
        1000000*(verify_total/byte_count)
    );

    delete card;

    fclose( in_handle );

    // return successfully
    //
    return 0;
}

