#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/fpgatest.h"         // fpga_test() ERROR_FPGA_xxxx



// conditional compile switches
//
#define FPGA_WALK_VERBOSE       TRUE
#define FPGA_TESTS_TRANSMIT     TRUE
#define FPGA_TESTS_AIM          TRUE
#define MAX_ADDRESS_PARANOIA    TRUE



// manifest constants
//
#define REGISTER_ADDRESS_LSB    4



static int fpga_read_after_write( Bit8 *address, Bit8 new_value, const char *state, const char *device, int error_base )
{
    Bit8 value_read;
    int result;

#if FPGA_WALK_VERBOSE
    fprintf( stdout, " 0x%.2X", new_value );
#endif

    *address = new_value;

    value_read = *address;

    if( value_read != new_value )
    {
        fprintf( stderr, "read 0x%.2X instead of %s 0x%.2X at %s FPGA address 0x%.8X\n", value_read, state, new_value, device, address );
        return error_base;
    }

    return 0;
}



#if 0
// not used by anyone, but I implemented it in case we ever
// want to do the checkerboard test we did for SRAM data
//
static int fpga_write_after_read( Bit8 *address, Bit8 old_value, Bit8 new_value, const char *state, const char *device, int error_base )
{
    Bit8 value_read;
    int result;

    value_read = *address;

    if( value_read != old_value )
    {
        fprintf( stderr, "read 0x%.2X instead of %s 0x%.2X at %s FPGA address 0x%.8X\n", value_read, state, old_value, device, address );
        return error_base;
    }

    *address = new_value;

    return 0;
}
#endif



static int fpga_walk_1_thru_0_data( Diagnostic_register *address, const char *device )
{
    Bit32 count;
    int result;
    Bit8 fpga_data;

    printf( "walking data 1 through 0's at %s FPGA 8-bit address 0x%.8X\n", device, &address->Value );

    // initial 0's
    //
    result = fpga_read_after_write( &address->Value, 0, "initial", device, ERROR_FPGA_TEST_INIT_0_MISCOMP );
    if( result )
        return result;

    // walk 1 through 0's
    //
    fpga_data = 1;
    for( count=0; count<8; count++ )
    {
        result = fpga_read_after_write( &address->Value, fpga_data, "walking", device, ERROR_FPGA_TEST_WALK_1_MISCOMP );
        if( result )
            return result;

        fpga_data <<= 1;
    }

    // final 0's
    //
    result = fpga_read_after_write( &address->Value, 0, "final", device, ERROR_FPGA_TEST_FINAL_0_MISCOMP );
#if FPGA_WALK_VERBOSE
    fprintf( stdout, "\n" );
#endif
    if( result )
        return result;

    return 0;
}



static int fpga_walk_0_thru_1_data( Diagnostic_register *address, const char *device )
{
    Bit32 count;
    int result;
    Bit8 fpga_data;

    printf( "walking data 0 through 1's at %s FPGA 8-bit address 0x%.8X\n", device, &address->Value );

    // initial 1's
    //
    result = fpga_read_after_write( &address->Value, -1, "initial", device, ERROR_FPGA_TEST_INIT_1_MISCOMP );
    if( result )
        return result;

    // walk 0 through 1's
    //
    fpga_data = ~1;
    for( count=0; count<8; count++ )
    {
        result = fpga_read_after_write( &address->Value, fpga_data, "walking", device, ERROR_FPGA_TEST_WALK_0_MISCOMP );
        if( result )
            return result;

        fpga_data = (fpga_data << 1) | 1;
    }

    // final 1's
    //
    result = fpga_read_after_write( &address->Value, -1, "final", device, ERROR_FPGA_TEST_FINAL_1_MISCOMP );
#if FPGA_WALK_VERBOSE
    fprintf( stdout, "\n" );
#endif
    if( result )
        return result;

    return 0;
}



// turn off warning about the unused assignment
// of value_read to itself
//
#pragma warn -aus

// it is OK to use this when the XOR register in the FPGA
// supports fewer than 8 bits, as long as the caller takes
// care of zeroing the unused bits of offset beforehand
//
static int fpga_address_xor( Bit8 *base, Bit16 offset, AddressXor_register *xor_reg, const char *state, const char *device, int error_base )
{
    Bit8 value_read;
    Bit32 expected_xor;
    Bit32 actual_xor;
    int result;

#if FPGA_WALK_VERBOSE
    fprintf( stdout, " 0x%.3X", offset );
#endif

    expected_xor = ((offset >> 2)
                 ^  (offset >> 3)
                 ^  (offset >> 4)
                 ^  (offset >> 5)
                 ^  (offset >> 6)
                 ^  (offset >> 7)
                 ^  (offset >> 8)
                 ^  (offset >> 9)) & 1;

    value_read = base[ offset ];


    // we must keep the compiler from optimizing away the
    // previous statement due to the result being unused
    // so we do an extra assignment here
    //
    value_read = value_read;

    actual_xor = xor_reg->Value;
    if( actual_xor != expected_xor )
    {
        fprintf( stderr, "address XOR diagnostic was %d instead of %d after reading %s %s FPGA address 0x%.8X + 0x%.3X\n", actual_xor, expected_xor, state, device, base, offset );
        return error_base;
    }

    return 0;
}

// turn on warnings about unused variable assignments
// if they were on initially
//
// this must be done after the end of the function,
// not immediately after the assignment itself
//
#pragma warn .aus



// fills 0 with 1's starting at the lsb
//
static int fpga_walk_1_thru_0_address( Bit32 bits_to_test, Bit8 *base, AddressXor_register *xor_reg, const char *device )
{
    Bit32 count;
    int result;
    Bit16 fpga_offset;

    printf( "walking address 1 through 0's at %s FPGA 8-bit base address 0x%.8X\n", device, base );

    // initial 0's
    //
    result = fpga_address_xor( base, 0, xor_reg, "initial", device, ERROR_FPGA_TEST_INIT_0_ADDR_XOR );
    if( result )
        return result;

    // walk 1 through 0's
    //
    fpga_offset = REGISTER_ADDRESS_LSB;
    for( count=0; count<bits_to_test; count++ )
    {
        result = fpga_address_xor( base, fpga_offset, xor_reg, "walking", device, ERROR_FPGA_TEST_WALK_1_ADDR_XOR );
        if( result )
            return result;

        fpga_offset = (fpga_offset << 1) | REGISTER_ADDRESS_LSB;
    }

#if FPGA_WALK_VERBOSE
    fprintf( stdout, "\n" );
#endif
    return 0;
}



// fills 1 with 0's starting at lsb
//
static int fpga_walk_0_thru_1_address( Bit32 bits_to_test, Bit8 *base, AddressXor_register *xor_reg, const char *device )
{
    Bit32 count;
    int result;
    Bit16 fpga_offset;
    Bit32 max_address = (((1 << bits_to_test) * REGISTER_ADDRESS_LSB) - 1) & (~(REGISTER_ADDRESS_LSB - 1));

#if MAX_ADDRESS_PARANOIA
    if( bits_to_test == 8 )
    {
        if( max_address != 0x3FC )
        {
            printf( "internal error: max_address should be 0x3FC, not 0x%.3X, when bits_to_test==8\n", max_address );
            return ERROR_FPGA_TEST_BAD_MAX_ADDRESS;
        }
    }
    else if( bits_to_test == 5 )
    {
        if( max_address != 0x7C )
        {
            printf( "internal error: max_address should be 0x7C, not 0x%.3X, when bits_to_test==5\n", max_address );
            return ERROR_FPGA_TEST_BAD_MAX_ADDRESS;
        }
    }
    else
    {
        printf( "internal error: bits_to_test should be either 8 or 5, not %d\n", bits_to_test );
        return ERROR_FPGA_TEST_BAD_MAX_ADDRESS;
    }

#endif

    printf( "walking address 0 through 1's at %s FPGA 8-bit base address 0x%.8X\n", device, base );

    // initial 1's
    //
    result = fpga_address_xor( base, max_address, xor_reg, "initial", device, ERROR_FPGA_TEST_INIT_1_ADDR_XOR );
    if( result )
        return result;

    // walk 0 through 1's
    //
    fpga_offset = max_address - REGISTER_ADDRESS_LSB;
    for( count=0; count<bits_to_test; count++ )
    {
        result = fpga_address_xor( base, fpga_offset, xor_reg, "walking", device, ERROR_FPGA_TEST_WALK_0_ADDR_XOR );
        if( result )
            return result;

        fpga_offset = (fpga_offset << 1) & max_address;
    }

#if FPGA_WALK_VERBOSE
    fprintf( stdout, "\n" );
#endif
    return 0;
}






int fpga_test( Fpga_registers *both_fpgas, AimAndCpld_registers *aim_and_cpld )
{
    int result;

    result = fpga_walk_1_thru_0_data( &both_fpgas->Receive.Diagnostic, "RX" );
    if( result )
        return result;

    result = fpga_walk_0_thru_1_data( &both_fpgas->Receive.Diagnostic, "RX" );
    if( result )
        return result;

    result = fpga_walk_1_thru_0_address( 8, (Bit8 *)&both_fpgas->Receive.dword_0, &both_fpgas->Receive.AddressXor, "RX" );
    if( result )
        return result;

    result = fpga_walk_0_thru_1_address( 8, (Bit8 *)&both_fpgas->Receive.dword_0, &both_fpgas->Receive.AddressXor, "RX" );
    if( result )
        return result;

#if FPGA_TESTS_TRANSMIT
    result = fpga_walk_1_thru_0_data( &both_fpgas->Transmit.Diagnostic, "TX" );
    if( result )
        return result;

    result = fpga_walk_0_thru_1_data( &both_fpgas->Transmit.Diagnostic, "TX" );
    if( result )
        return result;

    result = fpga_walk_1_thru_0_address( 8, (Bit8 *)&both_fpgas->Transmit.dword_0, &both_fpgas->Transmit.AddressXor, "TX" );
    if( result )
        return result;

    result = fpga_walk_0_thru_1_address( 8, (Bit8 *)&both_fpgas->Transmit.dword_0, &both_fpgas->Transmit.AddressXor, "TX" );
    if( result )
        return result;
#endif

#if FPGA_TESTS_AIM
    result = fpga_walk_1_thru_0_data( &aim_and_cpld->Aim.Diagnostic, "AIM" );
    if( result )
        return result;

    result = fpga_walk_0_thru_1_data( &aim_and_cpld->Aim.Diagnostic, "AIM" );
    if( result )
        return result;

    result = fpga_walk_1_thru_0_address( 5, (Bit8 *)&aim_and_cpld->Aim.dword_0, &aim_and_cpld->Aim.AddressXor, "AIM" );
    if( result )
        return result;

    result = fpga_walk_0_thru_1_address( 5, (Bit8 *)&aim_and_cpld->Aim.dword_0, &aim_and_cpld->Aim.AddressXor, "AIM" );
    if( result )
        return result;
#endif

    return 0;
}



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    int result;
    AtmCard *card;
    const char *program_name = argv[0];

    // parse command-line arguments
    //
    if( argc != 2 )
    {
        fprintf( stderr, "tests TX and RX and AIM FPGA address and data buses on the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex\n", program_name );
        return ERROR_FPGA_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_SRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    sram_init_stats();

    result = fpga_test( card->both_fpgas, card->aim_and_cpld );

    //sram_print_stats();

    delete card;

    if( result )
        return result;

    return 0;
}
