#ifndef _BUFFER_H
#define _BUFFER_H



#include <stdio.h>                      // printf()
#include <mem.h>                        // memset()
#include <dos.h>                        // sound() nosound()
#include <conio.h>                      // cprintf()
#include <except.h>                     // class xalloc
#include "..\types\types.h"             // Bit8 Bit16
#include "..\dpmi32\flat.h"             // flat_alloc()
#include "..\integrat\constant.h"       // COMPL_RING_ENTRIES
#include CARD_CLOCK_HEADER              // class CardClock



// conditional compilation switches
//
#define BUFFER_FG_SAFETY_CHECKS         TRUE        // routines called by foreground
#define BUFFER_BG_SAFETY_CHECKS         TRUE        // routines called by background
#define BUFFER_ALLOCATE_NONNIL_NEXT     FALSE
#define BUFFER_AMOUNT_FREE_VERBOSE      FALSE
#define BUFFER_AMOUNT_FILLED_VERBOSE    FALSE
#define BUFFER_INIT_VERBOSE             FALSE
#define BUFFER_CONSTANT_NEXT_OFFSET     FALSE
#define BUFFER_DUMP_TICKS               TRUE
#define CLEAR_MEM_DURING_INIT           TRUE        // shouldn't affect execution
#define ESSENTIAL_TIME_STAMPS           TRUE
#define THROW_IS_LEGAL                  TRUE
#define THROW_IS_SAFE_IN_INTERRUPT      FALSE
#define BEEP_IF_YOU_GET_HERE            FALSE
#define WAIT_FOR_KEYSTROKE              TRUE




// manifest constants
//
// it is okay for these to have forward references
// because the string substitution doesn't occur
// until the macro is used in the code far below
//
#define RX_FREE_RING_ALIGNMENT      4096            // only 1024 used by h/w
#define RX_COMPL_RING_ALIGNMENT     4096
#define RX_BUFFER_ALIGNMENT         4               // 4 might be OK for FORE card

#define SW_RINGS                    all_buffers->total_blocks

//#define SW_RINGS_WANTED           500
#define SW_RINGS_WANTED             (flat_remaining() / COMPL_RING_ENTRIES / BUFFER_SIZE)

#define BUFFER_COUNT                (SW_RINGS * COMPL_RING_ENTRIES)

#define CLEAR_MEM_VALUE             0x87654321L // 32 bits wide

#define QUEUE_MAX_NIL_PTRS          2

// assumes that partial blocks are only put onto filled list during cleanup
#define QUEUE_MAX_MISSING_BLKS      (0 * MAX_ATM_NICS)
#define QUEUE_MAX_MISSING_BUFS      ((QUEUE_MAX_MISSING_BLKS * COMPL_RING_ENTRIES) + (MAX_ATM_NICS * (COMPL_RING_ENTRIES - 1)))



// error return codes
//
#define ERROR_QUEUE_BASE                1000
#define ERROR_QUEUE_NOT_FILLED          (ERROR_QUEUE_BASE + 0)
#define ERROR_QUEUE_NOT_MY_ENTRY        (ERROR_QUEUE_BASE + 1)
#define ERROR_QUEUE_NOT_FREE            (ERROR_QUEUE_BASE + 2)
#define ERROR_QUEUE_FILE_CANNOT_OPEN    (ERROR_QUEUE_BASE + 3)
#define ERROR_QUEUE_FILE_CANNOT_WRITE   (ERROR_QUEUE_BASE + 4)
#define WARNING_QUEUE_FILE_CANNOT_CLOSE (-ERROR_QUEUE_BASE - 5)
#define ERROR_QUEUE_WRONG_FLAT_ADDR     (ERROR_QUEUE_BASE + 6)
#define ERROR_QUEUE_WRONG_USED_COUNT    (ERROR_QUEUE_BASE + 7)
#define ERROR_QUEUE_TOO_MANY_NIL_PTRS   (ERROR_QUEUE_BASE + 8)
#define ERROR_QUEUE_PTR_OUT_OF_BOUNDS   (ERROR_QUEUE_BASE + 9)
#define ERROR_QUEUE_DEALLOCATED         (ERROR_QUEUE_BASE + 10)
#define ERROR_QUEUE_STARTS_AT_ZERO      (ERROR_QUEUE_BASE + 11)
#define ERROR_QUEUE_TOO_MANY_MISSING    (ERROR_QUEUE_BASE + 12)
#define ERROR_QUEUE_STAMP_FREE_CIRCULAR (ERROR_QUEUE_BASE + 13)
#define ERROR_QUEUE_BAD_ALLOC_COUNT     (ERROR_QUEUE_BASE + 14)
#define WARNING_QUEUE_FILE_KEYPRESS     (-ERROR_QUEUE_BASE - 15)
#define ERROR_QUEUE_FILE_ALREADY_OPEN   (ERROR_QUEUE_BASE + 16)
#define ERROR_QUEUE_FILE_NOT_OPEN       (ERROR_QUEUE_BASE + 17)
#define ERROR_QUEUE_STATE_CANNOT_FILL   (ERROR_QUEUE_BASE + 18)
#define ERROR_QUEUE_STATE_CANNOT_FREE   (ERROR_QUEUE_BASE + 19)
#define ERROR_QUEUE_DISK_FULL           (ERROR_QUEUE_BASE + 20)
#define ERROR_QUEUE_BAD_CARDINDEX       (ERROR_QUEUE_BASE + 21)
#define ERROR_QUEUE_NOT_SENDABLE        (ERROR_QUEUE_BASE + 22)



enum BufferBlockState
{
    is_freed,                   // on the free list (rx or tx)
    is_filling,                 // being filled by a card (rx) or host (tx)
    is_filled,                  // on filled list (rx)
    is_consuming,               // being consumed for statistics (rx)
    is_sendable,                // on sendable list (tx) for specific card
    is_sending                  // being sent (tx)
};



// one block of COMPL_RING_ENTRIES buffers and the info necessary
// to put it on a linked list and know who filled it
//
// this data structure has no time stamps, since the
// FORE card will attach timestamps to each buffer
//
class BufferBlock
{

public:
    // next BufferBlock on linked list
    //
    BufferBlock *next;

    // flat (but not physical) pointer to first
    // of COMPL_RING_ENTRIES contiguous buffers
    // each of which is BUFFER_SIZE bytes long
    //
    Bit8 *buffers;

    // which card did we expect to fill in this block?
    //
    Bit16 cardnum;
    Bit8 cardindex;

    // how many of buffers in this block are filled in?
    //
    Bit32 used;

    // how many of the buffers in this block have already
    // been queried by the foreground statistics code?
    //
    Bit16 queried;

    // how many of the buffers in this block have already
    // been freed by the foreground cell reassembly code?
    //
    Bit16 freed;

    // who is doing what with this block?
    //
    // helps detect attempts to crosslink the lists
    //
    BufferBlockState state;

    // constructor
    //
    BufferBlock()
    {
        // it is OK to allocate this memory here, per
        // block, rather than in a larger aggregation,
        // because flat_alloc() does not have a per-allocation
        // memory overhead
        //
        buffers = (Bit8 *) flat_alloc( BUFFER_SIZE * COMPL_RING_ENTRIES );

        // check that buffer is aligned the way we want,
        // but do not attempt to deal with the problem
        //
        if( align( RX_BUFFER_ALIGNMENT, buffers ) != buffers )
        {
            printf( "internal error: BufferBlock at 0x%p got buffers at flat"
                " 0x%8.8lX, which is not %d bytes aligned!\n",
                this, buffers, RX_BUFFER_ALIGNMENT
            );

            exit( 3 );
        }

        // no buffers have been filled in yet
        //
        used = 0;

        // foreground hasn't dealt with any buffers yet
        //
        queried = 0;
        freed = 0;

        // BufferQueue constructor will put us on the free list
        //
        state = is_freed;

#if CLEAR_MEM_DURING_INIT
        next = (BufferBlock *) CLEAR_MEM_VALUE;
        cardnum = (Bit16) CLEAR_MEM_VALUE;
        cardindex = (Bit8) CLEAR_MEM_VALUE;

        // fill all the buffers with zero
        //
        // this makes it easier to tell the filled data
        // from the unfilled data, when looking at a core dump
        //
        // should not affect program execution, unless
        // there is some data structure we forgot to fill
        // in in our code!
        //
        memset( buffers, CLEAR_MEM_VALUE, BUFFER_SIZE*COMPL_RING_ENTRIES );
#endif
    }

    CardClock &first_timestamp( void )
    {
        return * (CardClock *) buffers;
    }

    CardClock &last_timestamp( void )
    {
        return * (CardClock *) (buffers + used*BUFFER_SIZE);
    }
};



// holds a queue of buffer blocks, each of which can be in one of
// 4 states: free, allocated, filled, or reading
//
// we do not keep track of buffers in the allocated and reading
// states to save execution time, although that can be implemented
// transparently at a later time if desired
//
// freed buffers are on a last-in-first-out (LIFO) linked list
// for execution speed, since there is no reason to maintain
// order among freed buffers
//
// filled buffers are on a first-in-first-out (FIFO) linked list
// so that a consumer can empty them in the same order they were
// filled
//
// singly-linked lists are used so that buffers can be filled in
// a different order than they reside in memory, and freed in a
// different order than they were filled
//
//
//
// only one of these resides in a machine at a time
//
class BufferQueue
{

private:
    // pointer to first buffer managed by first control block
    //
    // only used by check() and block_from_record()
    //
    Bit8 *start_buffers;

    // pointer to header of first control block we allocated
    //
    BufferBlock *start;

    // pointer to header of last+1 control block
    //
    BufferBlock *end;

    // pointer to header of next free control block
    // be given to SAR chip to fill in at its leisure
    //
    // newly-freed elements get inserted at front
    //
    BufferBlock *free_head;

    // pointers to headers of first and last entries in a
    // linked list of blocks that have been filled in by SAR
    //
    BufferBlock *filled_head;
    BufferBlock *filled_tail;

    // pointers to headers of first and last entries in several
    // linked lists of blocks that are ready to send
    //
    BufferBlock *sendable_head[ MAX_ATM_NICS ];
    BufferBlock *sendable_tail[ MAX_ATM_NICS ];

    // how many blocks and records on the filled list?
    //
    Bit32 filled_block_count;
    Bit32 filled_record_count;

    // how many blocks and records on the sendable lists?
    //
    Bit32 sendable_block_count[ MAX_ATM_NICS ];
    Bit32 sendable_record_count[ MAX_ATM_NICS ];

    // how many blocks on the free list?
    //
    Bit32 free_block_count;

    Bit32 prev_allocate_failure_count;
    Bit32 prev_read_tx_failure_count[ MAX_ATM_NICS ];

    // C library file handle to dump file
    //
    int dump_handle;

    // name of file we are dumping to
    //
    // only used for diagnostic messages
    //
    const char *dump_filename;

public:
    // how many blocks are we managing (static)?
    //
    // (also known as SW_RINGS)
    //
    Bit32 total_blocks;

    // how many times has allocate() failed?
    //
    Bit32 allocate_failure_count;

    // has allocate() failed since the last call to this function?
    //
    int has_allocate_failed_lately( void )
    {
        int result = allocate_failure_count > prev_allocate_failure_count;
        prev_allocate_failure_count = allocate_failure_count;
        return result;
    }

    // how many times has read_tx() failed for this cardindex?
    //
    Bit32 read_tx_failure_count[ MAX_ATM_NICS ];

    // has read_tx() failed for this card since the last call to this function?
    //
    int has_read_tx_failed_lately( Bit8 cardindex )
    {
        int result = read_tx_failure_count[cardindex] > prev_read_tx_failure_count[cardindex];
        prev_read_tx_failure_count[cardindex] = read_tx_failure_count[cardindex];
        return result;
    }

    // constructor
    //
    BufferQueue();

    // destructor
    //
    ~BufferQueue()
    {
        delete[] start;
    }

    // tell the caller where the buffer starts
    //
    // should only be used for debug messages;
    // there is no need for caller to have this
    // info during normal execution!
    //
    const BufferBlock *get_start( void )
    {
        return start;
    }

    // calculates which BufferBlock owns a piece of memory
    //
    // assumes the memory manager hasn't inserted any bytes
    // between the memory it hands out for successive calls
    //
    // this is only true for flat_alloc() !!!
    // so if we start calling malloc() again, the memory that
    // the BufferBlock's manage will have to be allocated all
    // at once by somebody else, not their constructors
    //
    BufferBlock *block_from_record( void *interior )
    {
        return (((Bit8 *)interior - start_buffers) / BUFFER_SIZE / COMPL_RING_ENTRIES) + start;
    }

    // does this entry belong in this queue?
    //
    // check that address is within range
    // and at an even 20-byte offset from start
    //
    int belongs( BufferBlock *pointer )
    {
        return pointer>=start && pointer<end
            && (((Bit32)pointer - (Bit32)start) % sizeof(BufferBlock))==0;
    }

    Bit32 amount_filled( void )
    {
        Bit32 dummy_block_count;
        return amount_filled( dummy_block_count );
    }

    Bit32 amount_filled( Bit32 &block_count )
    {
        block_count = filled_block_count;
        return filled_record_count;
    }

    Bit32 amount_filled( Bit16 cardnum )
    {
        Bit32 dummy_block_count;
        return amount_filled( cardnum, dummy_block_count );
    }

    Bit32 amount_filled( Bit16 cardnum, Bit32 &block_count );

    Bit32 amount_sendable( Bit8 cardindex )
    {
        Bit32 dummy_block_count;
        return amount_sendable( cardindex, dummy_block_count );
    }

    Bit32 amount_sendable( Bit8 cardindex, Bit32 &block_count )
    {
        if( cardindex >= MAX_ATM_NICS )
        {
            block_count = 0;
            return 0;
        }
        block_count = sendable_block_count[ cardindex ];
        return sendable_record_count[ cardindex ];
    }

    Bit32 amount_free( void )
    {
        Bit32 dummy_block_count;
        return amount_free( dummy_block_count );
    }

    Bit32 amount_free( Bit32 &block_count )
    {
        block_count = free_block_count;
        return block_count * COMPL_RING_ENTRIES;
    }

    int dump_open( const char *const filename );
    int dump_close( void );
    int dump_one( BufferBlock *block );
    int dump( const char *const filename );

    int check( void );

    // make one entry available to SAR chip
    //
    // no need to enable/disable interrupts, because
    // this function is only called by interrupt handler
    //
    int allocate( BufferBlock **pointer )
    {
        // keep another interrupt handler from altering
        // the free list head pointer while we are
        // manipulating it
        //
        disable();

        // caller gets the element at head of list
        //
        *pointer = free_head;

        // check that there is free space
        //
        if( !free_head )
        {
            enable();
            allocate_failure_count++;
            return ERROR_QUEUE_NOT_FREE;
        }

        // update counter
        //
        free_block_count--;

        // move the head pointer forwards one element
        //
        free_head = free_head->next;

        // we are done messing with free list
        //
        enable();

#if BUFFER_ALLOCATE_NONNIL_NEXT
        // force next pointer to be non nil, so BufferQueue::check()
        // will not be confused by in-use blocks
        //
        // but set the pointer to an invalid value so a protection
        // violation will occur if somebody is stupid enough to
        // dereference it
        //
        (*pointer)->next = (BufferBlock *) 1;
#endif

        // remember this block is being filled by a card or host, on no list
        //
        (*pointer)->state = is_filling;

        return 0;
    }

    // consumer is done with an entry
    //
    int free( BufferBlock *pointer )
    {
    #if BUFFER_FG_SAFETY_CHECKS
        // check that entry belongs on this queue
        //
        if( !belongs( pointer ) )
            return ERROR_QUEUE_NOT_MY_ENTRY;

        if( pointer->state != is_consuming
        &&  pointer->state != is_sending )
            return ERROR_QUEUE_STATE_CANNOT_FREE;
    #endif

        // forget which of the buffers have been used, queried, or freed
        //
        pointer->used = 0;
        pointer->queried = 0;
        pointer->freed = 0;

        // remember this block is on the free list
        //
        pointer->state = is_freed;

        // keep any interrupt handler from altering
        // the free list head pointer while we are
        // manipulating it
        //
        disable();

        // point the given element at the current head of list
        // (safe to do even if head pointer is NIL)
        //
        pointer->next = free_head;

        // move the head pointer backwards one element
        //
        free_head = pointer;

        // update counter
        //
        free_block_count++;

        // we are done messing with free list
        //
        enable();

        return 0;
    }

    // returns a pointer to the BufferBlock at the designated index
    // into the filled list, or NULL if index is invalid
    //
    BufferBlock *read( Bit16 cardnum, unsigned index );

    // consumer wants an entry from one of the sendable lists
    //
    // theoretically no need to enable/disable interrupts, because
    // this function is only called by interrupt handler, but we
    // do it anyways
    //
    int read_tx( BufferBlock **pointer, Bit8 cardindex )
    {
    #if BUFFER_BG_SAFETY_CHECKS
        // check that the cardindex is in range
        //
        if( cardindex >= MAX_ATM_NICS )
            return ERROR_QUEUE_BAD_CARDINDEX;
    #endif

        // keep any other interrupt handlers from altering the
        // sendable linked list while we are doing so
        //
        disable();

    #if BUFFER_BG_SAFETY_CHECKS
        // check that there is some data filled in
        //
        if( !sendable_head[ cardindex ] )
        {
            read_tx_failure_count[cardindex]++;
            enable();
            return ERROR_QUEUE_NOT_SENDABLE;
        }
    #endif

        *pointer = sendable_head[ cardindex ];

        // update counters
        //
        sendable_block_count[cardindex]--;
        sendable_record_count[cardindex] -= sendable_head[cardindex]->used;

        // if we just took the last or last+1 element off the list...
        //
        if( sendable_head[cardindex] == sendable_tail[cardindex] )
        {
            // both head and tail pointer should know about this
            //
            sendable_tail[cardindex] = 0;
            sendable_head[cardindex] = 0;
        }
        else
        {
            // move the head pointer forwards one element
            //
            sendable_head[cardindex] = sendable_head[cardindex]->next;

            // no special processing is required, even if
            // we moved head forward to point at tail
            //
        }

        // we are done altering the sendable linked list for this card
        //
        enable();

        // remember that this block is being processed
        // for statistics, on no list
        //
        (*pointer)->state = is_sending;

        return 0;
    }

    // consumer wants an entry from filled list
    //
    int read( BufferBlock **pointer )
    {
        // keep any interrupt handler from altering the
        // filled linked list while we are doing so
        //
        disable();

    #if BUFFER_FG_SAFETY_CHECKS
        // check that there is some data filled in
        //
        if( !filled_head )
        {
            enable();
            return ERROR_QUEUE_NOT_FILLED;
        }
    #endif

        *pointer = filled_head;

        // update counters
        //
        filled_block_count--;
        filled_record_count -= filled_head->used;

        // if we just took the last or last+1 element off the list...
        //
        if( filled_head == filled_tail )
        {
            // both head and tail pointer should know about this
            //
            filled_tail = 0;
            filled_head = 0;
        }
        else
        {
            // move the head pointer forwards one element
            //
            filled_head = filled_head->next;

            // no special processing is required, even if
            // we moved head forward to point at tail
            //
        }

        // we are done altering the filled linked list
        //
        enable();

        // remember that this block is being processed
        // for statistics, on no list
        //
        (*pointer)->state = is_consuming;

        return 0;
    }

    // user chip has filled in an entry
    //
    int fill_tx( BufferBlock *pointer, Bit8 cardindex )
    {
    #if BUFFER_FG_SAFETY_CHECKS
        // check that the cardindex is in range
        //
        if( cardindex >= MAX_ATM_NICS )
            return ERROR_QUEUE_BAD_CARDINDEX;

        if( !belongs( pointer ) )
            return ERROR_QUEUE_NOT_MY_ENTRY;

        if( pointer->state != is_filling )
            return ERROR_QUEUE_STATE_CANNOT_FILL;
    #endif

        // enforce the block's cardindex to match the list it will be on
        //
        pointer->cardindex = cardindex;

        // remember that this block is filled with data, on sendable list
        //
        pointer->state = is_sendable;

        // mark the given element as the tail of the list
        //
        pointer->next = 0;

        // keep interrupt handlers from altering the
        // filled linked list while we are doing so
        //
        disable();

        // if there was something on the list before...
        //
        if( sendable_tail[cardindex] )
        {
            // point the previously-last element at the given element
            //
            sendable_tail[cardindex]->next = pointer;
        }
        else
        {
            // linked list head pointer should also point at given element
            //
            sendable_head[cardindex] = pointer;
        }

        // move the tail pointer forwards one element
        //
        sendable_tail[cardindex] = pointer;

        // update counters
        //
        sendable_block_count[cardindex]++;
        sendable_record_count[cardindex] += sendable_tail[cardindex]->used;

        // we are done altering the sendable linked list
        //
        enable();

        return 0;
    }

    // SAR chip has filled in an entry
    //
    // theoretically no need to enable/disable interrupts, because
    // this function is only called by interrupt handler, but we
    // do it anyways
    //
    int fill( BufferBlock *pointer )
    {
    #if BUFFER_BG_SAFETY_CHECKS
        if( !belongs( pointer ) )
            return ERROR_QUEUE_NOT_MY_ENTRY;

        if( pointer->state != is_filling )
            return ERROR_QUEUE_STATE_CANNOT_FILL;
    #endif

        // remember that this block is filled with data, on filled list
        //
        pointer->state = is_filled;

        // mark the given element as the tail of the list
        //
        pointer->next = 0;

        // keep other interrupt handlers from altering the
        // filled linked list while we are doing so
        //
        disable();

        // if there was something on the list before...
        //
        if( filled_tail )
        {
            // point the previously-last element at the given element
            //
            filled_tail->next = pointer;
        }
        else
        {
            // linked list head pointer should also point at given element
            //
            filled_head = pointer;
        }

        // move the tail pointer forwards one element
        //
        filled_tail = pointer;

        // update counters
        //
        filled_block_count++;
        filled_record_count += filled_tail->used;

        // we are done altering the filled linked list
        //
        enable();

        return 0;
    }
};



extern BufferQueue *all_buffers;



#endif      // not included yet
