// real-mode program to test throughput of BIOS's disk write command
//
// destroys data at a few of the cylinders of the second physical hard disk
//



#include <bios.h>                           // biosdisk()
#include <stdio.h>                          // printf()
#include <dos.h>                            // MK_FP() FP_OFF() FP_SEG()
#include <string.h>                         // stricmp()
#include <stdlib.h>                         // atoi()
#include "../types/types.h"                 // Bit8 Bit16 Bit32
#include "../dpmi32/realmode.h"             // real_to_physical() physical_to_real()
#include "../intel/timestmp.h"              // class PentiumClock



// manifest constants that should have been in \bc45\include
//
#define _DISK_GETPARMS      8
#define _DISKERR_BAD_CMD    1
#define MBR_SIGNATURE       0xAA55          // byteswapped for Intel
#define MBR_SIGNATURE_OFF   0x1FE           // from start of master boot record
#define MBR_PARTITION_OFF   0x1BE           // also from start of MBR, room for 4 total
#define MBR_PARTITION_CNT   4
#define PARTITION_BOOT_FLG  0x80



// data structures that should have been in \bc45\include
//
struct diskparms_t
{
    Bit16 tracks_high : 2;                  // numbered from 0, so add 1 for count
    Bit16 sectors : 6;                      // numbered from 1, so is count
    Bit16 tracks_low : 8;
    Bit8 drives;                            // numbered from 1, so is count
    Bit8 heads;                             // numbered from 0, so add 1 for count
};

enum PartitionSystem
{
    Unknown = 0,
    DosFat12 = 1,
    DosFat16 = 4,
    DosExtFat16 = 5,
    DosExtFat16Pri = 6,
    W95Fat32 = 11,
    CoralSave = 123
};

struct partition_t
{
    Bit8 bootable;                          // 0x80 if yes
    Bit8 start_head;
    Bit8 start_sector : 6;
    Bit8 start_cylinder_high : 2;
    Bit8 start_cylinder_low;
    Bit8 system_id;
    Bit8 end_head;
    Bit8 end_sector : 6;
    Bit8 end_cylinder_high : 2;
    Bit8 end_cylinder_low;
    Bit32 lba_start_sector;
    Bit32 lba_sector_count;
};

struct master_boot_record_t
{
    Bit8 boot_program[ MBR_PARTITION_OFF ];
    struct partition_t partitions[ MBR_PARTITION_CNT ];
    Bit16 signature;
};





// constants that all disks have in common
//
#define FIRST_DISK      0x80
#define DISK_TOTAL      16
#define FIRST_HEAD      0
#define FIRST_SECTOR    1
#define FIRST_CYLINDER  0
#define SECTOR_SIZE     512
#define ITERATIONS      1000
#define BUFFER_ALIGN    65536               // avoids BIOS DMA boundary errors



// manifest constants that indicate choices we made
//

// 128 is maximum for EIDE drives (129 causes lockup in BIOS)
// 255 works for SCSI drives with Adaptec BIOS (256 made disk inaccessible until reboot)
//
#define BUFFER_SIZE     (255 * (Bit32)SECTOR_SIZE)



// global variables
//
Bit8 huge buffer0[ BUFFER_SIZE + BUFFER_ALIGN - 1 ];
Bit8 huge *buffer;



// forces a pointer forward until it is aligned on
// an N byte boundary, where N is a power of 2
//
Bit32 align( Bit32 boundary, Bit32 pointer )
{
    return ((pointer) + (boundary - 1)) & (~(boundary-1));
}



int main( int argc, char *argv[] )
{
    // cmdline parms
    //
    const char *program = argv[ 0 ];
    const char *user_sector_offset_string = argv[ 1 ];
    const char *user_sector_length_string = argv[ 2 ];
    const char *want_write_string = argv[ 3 ];
    long user_sector_offset;
    long user_sector_length;
    int want_write = FALSE;
    int operation;
    char *operation_name;

    // disk, partition, and geometry we found out
    //
    int disk_count;
    int partition_count;
    int start_cylinder, end_cylinder;
    int start_head, end_head;
    int start_sector, end_sector;
    Bit32 lba_start_sector, lba_end_sector;
    struct diskparms_t parms;
    struct master_boot_record_t mbr;
    int tracks_per_disk;
    Bit8 heads_per_track;
    Bit8 sectors_per_head;
    int sectors_per_track;              // actually only needs to be Bit16
    Bit32 sectors_per_disk;
    double bytes_per_disk;

    // loop control variables and statistics we calculated
    //
    int sector_count;
    int sectors_this_head;
    int heads_this_track;
    int head_count;
    int track_count;
    Bit32 true_end_sector;
    Bit32 lba_sector;
    int result;
    Bit16 sectors_per_buffer = BUFFER_SIZE / SECTOR_SIZE;
    PentiumClock before, after;
    double processing_seconds;
    Bit32 sectors_processed;
    Bit32 calls_to_bios;
    double bytes_processed;
    double bytes_per_second;

    if( argc == 4 )
    {
        want_write = !stricmp( want_write_string, "WRITE" );
    }

    if( (argc!=3 && argc!=4) || (argc==4 && !want_write) )
    {
        printf( "usage: %s sector-offset-within-partition length-in-sectors {WRITE}\n"
            "you must spell out the word 'WRITE' if you want the disk to be written to\n", program );

        return 1;
    }

    if( want_write )
    {
        operation = _DISK_WRITE;
        operation_name = "write";
        printf( "write is enabled\n" );
    }
    else
    {
        operation = _DISK_READ;
        operation_name = "read";
        printf( "write is disabled; reads will be used instead\n" );
    }

    user_sector_offset = atol( user_sector_offset_string );
    if( user_sector_offset < 0 )
    {
        printf( "sorry, offset must not be negative; you typed '%s' which converted to %ld\n",
            user_sector_offset_string, user_sector_offset );
        return 1;
    }

    user_sector_length = atol( user_sector_length_string );
    if( user_sector_length < 1 )
    {
        printf( "sorry, sector length must be 1 or greater; you typed '%s' which converted to %ld\n",
            user_sector_length_string, user_sector_length );
        return 1;
    }

    printf( "you asked to access %ld sectors starting %ld into the first CoralSave partition\n",
        user_sector_length, user_sector_offset );

    buffer = (Bit8 huge *) physical_to_real( align( BUFFER_ALIGN, real_to_physical( (Bit32)buffer0 ) ) );
    printf( "0x%.4X:%.4X buffer before alignment\n", FP_SEG(buffer0), FP_OFF(buffer0) );
    printf( "0x%.4X:%.4X buffer after  alignment\n", FP_SEG(buffer), FP_OFF(buffer) );



    // first section:
    // determine which disk to use, and where on that disk
    //

    // for all possible disks...
    //
    for( disk_count=FIRST_DISK; disk_count<FIRST_DISK+DISK_TOTAL; disk_count++ )
    {
        // get the master boot record
        //
        result = biosdisk( _DISK_READ, disk_count, FIRST_HEAD, FIRST_CYLINDER, FIRST_SECTOR, 1, &mbr );
        if( result != 0 )
        {
            printf( "error %d getting master boot record for disk 0x%.2X\n", result, disk_count );
            if( result == _DISKERR_BAD_CMD )
            {
                printf( "assuming no such disk exists\n" );
                continue;
            }
            return result;
        }

        // check that the MBR has a proper signature
        //
        if( mbr.signature != MBR_SIGNATURE )
        {
            printf( "warning: found 0x%.4X instead of expected signature 0x%.4X at"
               " offset 0x%.4X into MBR found at CHS %d/%d/%d on drive 0x%.2X\n",
               mbr.signature,
               MBR_SIGNATURE,
               MBR_SIGNATURE_OFF,
               FIRST_CYLINDER, FIRST_HEAD, FIRST_SECTOR,
               disk_count
            );
            printf( "assuming disk is not partitioned yet\n" );

            continue;
        }

        // rummage around for our partition
        //
        for( partition_count=0; partition_count<MBR_PARTITION_CNT; partition_count++ )
            if( mbr.partitions[ partition_count ].system_id == CoralSave )
                break;

        if( partition_count < MBR_PARTITION_CNT )
            break;

    }   // end for all disks

    if( disk_count >= FIRST_DISK + DISK_TOTAL )
    {
        printf( "sorry, no partition with system ID 0x%.2X was found on any"
            " of %d hard disks starting with 0x%.2X\n",
            CoralSave, DISK_TOTAL, FIRST_DISK
        );

        return 1;
    }

    start_cylinder = mbr.partitions[partition_count].start_cylinder_high * 256
                   + mbr.partitions[partition_count].start_cylinder_low;
    end_cylinder   = mbr.partitions[partition_count].end_cylinder_high * 256
                   + mbr.partitions[partition_count].end_cylinder_low;
    start_head     = mbr.partitions[partition_count].start_head;
    end_head       = mbr.partitions[partition_count].end_head;
    start_sector   = mbr.partitions[partition_count].start_sector;
    end_sector     = mbr.partitions[partition_count].end_sector;
    lba_start_sector = mbr.partitions[partition_count].lba_start_sector;
    lba_end_sector = mbr.partitions[partition_count].lba_sector_count
                   + lba_start_sector - 1;

    printf( "CoralSave partition (system ID = 0x%.2X) found as %dth on disk"
        " 0x%.2X starting at CHS %d/%d/%d (LBA %lu) and ending at CHS %d/%d/%d"
        " (LBA %lu) for a total sector count of %lu\n",
        CoralSave, partition_count+1, disk_count,
        start_cylinder, start_head, start_sector,
        lba_start_sector,
        end_cylinder, end_head, end_sector,
        lba_end_sector,
        mbr.partitions[partition_count].lba_sector_count
    );

    // get the disk geometry
    //
    result = biosdisk( _DISK_GETPARMS, disk_count, FIRST_HEAD, FIRST_CYLINDER, FIRST_SECTOR, 1, &parms );
    if( result != 0 )
    {
        printf( "error %d getting disk parms\n", result );
        return result;
    }

    tracks_per_disk = parms.tracks_high*256 + parms.tracks_low + 1 - FIRST_CYLINDER;
    heads_per_track = parms.heads + 1 - FIRST_HEAD;
    sectors_per_head = parms.sectors + 1 - FIRST_SECTOR;
    sectors_per_track = sectors_per_head * heads_per_track;
    sectors_per_disk = (Bit32)tracks_per_disk * (Bit32)heads_per_track * (Bit32)sectors_per_head;
    bytes_per_disk = (double)sectors_per_disk * SECTOR_SIZE;
    printf( "for disk=0x%.2X, tracks=%d sectors=%d heads=%d drives=%d\n",
        disk_count, tracks_per_disk, sectors_per_head, heads_per_track, parms.drives );
    printf( "so total capacity before partitioning and high-level formatting\n"
        " is %lu sectors = %.0f bytes\n", sectors_per_disk, bytes_per_disk );

    printf( "host memory buffer is %lu bytes = %u sectors\n",
        (Bit32)BUFFER_SIZE, sectors_per_buffer );
    if( sectors_per_buffer < sectors_per_head )
    {
        // does this really apply now that I am using sectors
        // independently of heads???
        //
        printf( "sorry, buffer cannot hold as many sectors as one head, so this simplistic program cannot run\n" );
        return 1;
    }

    if( lba_start_sector + user_sector_offset > lba_end_sector )
    {
        printf( "whoops, sector offset %ld you requested is beyond the end of the partition!\ntry again...\n", user_sector_offset );
        return 1;
    }

    // avoid overrunning the end of disk, the end of partition,
    // and the end of user-allowed accesses
    //
    true_end_sector = min( lba_end_sector, sectors_per_disk-1 );
    true_end_sector = min( true_end_sector, lba_start_sector
        +user_sector_offset+user_sector_length-1 );
    printf( "you will actually be accessing sectors %lu through %lu on disk 0x%.2X\n",
        lba_start_sector + user_sector_offset, true_end_sector, disk_count );


    // finally, we're gonna really access the disk!
    //

    before.set();

    sectors_processed = 0;
    calls_to_bios = 0;
    for( lba_sector=lba_start_sector+user_sector_offset;
         lba_sector<=true_end_sector;
         lba_sector+=sectors_this_head )
    {
        // from absolute LBA sector, calculate absolute
        // track, head, and sector to use for start of BIOS call
        //
        track_count = (lba_sector / sectors_per_track);
        sector_count = lba_sector - (track_count * sectors_per_track);
        head_count = sector_count / sectors_per_head;
        sector_count = sector_count - (head_count * sectors_per_head);

        track_count  += FIRST_CYLINDER;
        head_count   += FIRST_HEAD;
        sector_count += FIRST_SECTOR;

        if( sectors_per_buffer > true_end_sector + 1 - lba_sector )
            sectors_this_head = true_end_sector + 1 - lba_sector;
        else
            sectors_this_head = sectors_per_buffer;

        result = biosdisk( operation, disk_count, head_count, track_count, sector_count, sectors_this_head, buffer );
        if( result != 0 )
        {
            printf( "error %d from biosdisk(%s,drive=0x%.2X,head=%d,cyl=%d,sec=%d,cnt=%d)\n",
                result, operation_name, disk_count, head_count, track_count, sector_count, sectors_this_head );
            goto error;
        }

        // count sectors we just did
        //
        calls_to_bios++;
        sectors_processed += sectors_this_head;
    }
error:

    after.set();
    after -= before;
    processing_seconds = after.get_seconds();
    printf( "elapsed time %g seconds\n", processing_seconds );

    bytes_processed = (double)sectors_processed * SECTOR_SIZE;
    printf( "processed %lu sectors * %u bytes = %.0f bytes\n",
        sectors_processed, SECTOR_SIZE, bytes_processed );

    bytes_per_second = bytes_processed / processing_seconds;
    printf( "%g bytes per second\n", bytes_per_second );

    printf( "called BIOS %lu times total, %g per second\n",
        calls_to_bios, calls_to_bios/processing_seconds );

    return 0;
}
