/* expects input to be in one of these formats:
 *   "Status codes: s suppressed, d damped, h history, * valid, > best, i - internal"
 *   "Origin codes: i - IGP, e - EGP, ? - incomplete"
 *   ""
 *   "   Network          Next Hop          Metric LocPrf Weight Path"
 *   "*  1.0.0.0          192.41.177.235         5     99      0 174 i"
 *   "*>i                 204.70.7.53            5    100      0 174 i"
 *   "*  4.20.0.0/18      192.41.177.190       150     99      0 1849 i"
 *   " h 5.20.0.0         192.41.177.190               99      0 1849 i"
 *   "*>i                 204.70.7.53            5    100      0 174 i"
 *   " s 5.21.0.0         192.41.177.190       150     99      0 1849 i"
 *   "*>i                 204.70.7.53            5    100      0 174 i"
 *   " d 5.22.0.0         192.41.177.190       150     99      0 1849 i"
 *   "*>i                 204.70.7.53            5    100      0 174 i"
 *
 * where the leading star indicates a non-heading line
 *    except for history, suppressed, or dampened routes,
 *    which are identified by all of the first 3 columns
 * and the ">" in 2nd column means this is a preferred route
 * and fields are always separated from each other by at least one space
 * and lines without a destination IP address use IP from closest line above
 * and the dest IP field always starts at the 4th char of a line
 * and the dest IP field is padded with spaces on right to 16 characters,
 * but can get as long as 18 with "255.255.255.255/32",
 * and the nexthop IP field is padded with spaces on right to 16 characters
 * and lines without metric use 0 for it
 * and metric is always padded on left to 8 characters
 *
 * if the IP field has no mask bit count specified,
 * we are expected to use class-full (A, B, C, D, or E)
 * interpretation of the IP address itself (i.e. based
 * on the upper bits)
 *
 *
 *
 * assumes that AS path always has a 2-character trailer
 * like " i" or " ?" or " e"
 *
 * assumes that AS numbers are always nonzero
 *
 *
 *
 * the first number in the AS path is the organization that
 * handed us the route, i.e. UUNet or SprintLink or ANS
 *
 * the last number in the AS path is the source AS,
 * i.e. the AS who owns the subnet talked about
 *
 * when there is only one number in the AS path, the route is
 * owned by that AS and was advertized directly to us
 *
 * when there are no numbers in the AS field, the route is
 * assumed to have originated within your network, which means
 * use your own AS number (145 for NSF vBNS, 3561 for MCI BIPP)
 *
 * a set of AS numbers whose subnets have been aggregated
 * into a single route by a third party (the AS appearing
 * to the left of them in the AS path string) will appear
 * like "{1325,3672,1681}"
 *
 *
 *
 * ftp://ftp.isi.edu/in-notes/iana/assignments/as-numbers
 * AS number 0 is reserved
 * AS numbers 1 to 32767 are assigned by the IANA
 * AS numbers 32768 to 64511 are reserved
 * AS numbers 64512 to 65534 are reserved
 * AS number 65535 is reserved
 *
 *
 *
 * calls InsertSubnet() for each parsed line
 */



/* conditional compile switches
 */
#define READBGP_TALKS           FALSE
#define READBGP_PARANOIA        TRUE
#define READBGP_DEBUG_HELPER    FALSE
#define READBGP_SHOWS_DEF_AS    TRUE
#define READBGP_SHOWS_AGGREGS   TRUE
#define READBGP_STRIPS_CR       TRUE
#define READBGP_STRIPS_LF       TRUE
#define READBGP_IGNORES_NONPREF TRUE
#define READBGP_OUTPUTS_BINARY  TRUE
#define READBGP_OUTPUTS_DECIMAL TRUE
#define READBGP_OUTPUTS_FLAGS   FALSE



#include <stdio.h>                  /* printf() fprintf() gets() sscanf() */
#include <stdlib.h>                 /* atoi() getenv() */
#include <string.h>                 /* strcpy() strchr() strlen() */
#include <errno.h>                  /* errno _sys_errlist */
#ifdef __unix__
#define _sys_errlist sys_errlist
extern int sys_nerr;
extern char *sys_errlist[];
#include "../integrat/itoa.h"       /* my own itoa() */
#endif
#include "../types/types.h"
#include "../integrat/readbgp.h"    /* our own prototypes and error messages */
#include "../wattcp/tcp.h"          /* inet_ntoa() */
#include "../integrat/constant.h"   /* DO_FOREGROUND_STATS */
#ifndef __unix__
#include "../intel/byteswap.h"      /* htonl() */
#endif
#include "../dpmi32/flat.h"         /* flat_alloc() flat_remaining() */



/* none of this code gets compiled in
 * if it's isn't going to be called
 */
#if DO_FOREGROUND_STATS



/* manifest constants
 */
#define MAX_TEXTLINE_SIZE       160     /* typically 90 */
#define MAX_DEST_IP_WIDTH       (sizeof("255.255.255.255/32")-1)
#define DEF_DEST_IP_WIDTH       (sizeof("255.255.255.0/32")-1)
#define DEF_NEXTHOP_IP_OFFSET   20
#define DEF_METRIC_OFFSET       36
#define METRIC_WIDTH            8



static Bit32 mask_from_length[ 33 ];                     /* host byte order */
static unsigned length_from_high_nibble[ 16 ];
static unsigned subnet_parm_count = 0;
static unsigned line = 0;
#if READBGP_DEBUG_HELPER
    static unsigned stopline = 42;
#endif



/* sets up global variables used when parsing BGP files
 */
static int init_bgp( void )
{
    unsigned count;
    unsigned mask;

    /* fill in the table that maps from
     * high nibble to mask bit count
     */
    for( count=0; count<16; count++ )
        if( (count & 8) == 0 )                          /* class A */
            length_from_high_nibble[ count ] = 8;
        else if( (count & 4) == 0 )                     /* class B */
            length_from_high_nibble[ count ] = 16;
        else if( (count & 2) == 0 )                     /* class C */
            length_from_high_nibble[ count ] = 24;
        else if( (count & 1) == 0 )                     /* class D (multicast) */
            length_from_high_nibble[ count ] = 32;
        else                                            /* class E (reserved) */
            length_from_high_nibble[ count ] = 999;

    /* fill in the table that maps from
     * mask bit count to mask word
     *
     * need more than 1 statement so gdb (GNU debugger)
     * will be able to single-step
     *
     * note that on DEC alpha with OSF/1 compiler,
     * shifting an unsigned integer right will preserve
     * the high (sign) bit, but we do not depend on that
     * (nonportable, buggy) behavior
     */
    mask_from_length[ 0 ] = 0;
    for( count=1,mask=1<<31;
         count<33;
         count++,mask=mask|mask>>1 )
    {
        mask_from_length[ count ] = mask;
        mask = mask;
    }

    /* successful!
     */
    return 0;
}



// I know that the length parameter isn't used by this function,
// and I don't want the compiler nagging me about it
//
#pragma warn -par
static int parse_bgp_line( char *buffer, unsigned length )
{
    char *preferred = buffer + 1;
    char *dest = buffer + 3;
    char *nexthop_addr_string = buffer + DEF_NEXTHOP_IP_OFFSET;
    char *metric = buffer + DEF_METRIC_OFFSET;
    char trashed_for_dest;
    int dest_width;
    int converted;
    unsigned count;
    unsigned addr_octets[ 4 ];
    static unsigned mask_count;
    static Bit32 mask = 0;                          /* host byte order */
    static Bit32 addr = 0;
    char dest_binary[ 33 ];
    char dest_binary2[ 33 ];

    /* if the route isn't valid, and it also isn't
     * history, suppressed, or dampened...
     */
    if(  buffer[0] != '*'
    && ( buffer[0] != ' '
    ||   buffer[2] != ' '
    ||  (buffer[1] != 'h'
    &&   buffer[1] != 'd'
    &&   buffer[1] != 's')))

        /* ignore this input line
         */
        return 0;

    /* if there is a destination IP address...
     */
    if( dest[0] != ' ' )
    {
        /* make sure we don't misinterpret next hop IP address
         * as the destination IP address when destination IP
         * address is missing
         *
         * but remember what character we are trashing, so we
         * can restore it after parsing dest, so we can then
         * parse next hop router
         */
        trashed_for_dest = dest[ MAX_DEST_IP_WIDTH ];
        dest[ MAX_DEST_IP_WIDTH ] = '\0';

        /* try to read the destination IP address and its mask width
         *
         * also remember how many characters the field required
         *
         * note that we grab count of characters parsed ("%n" in format
         * string) both before and after looking for the mask width,
         * in case mask width is not found
         */
        converted = sscanf( dest, "%u.%u.%u.%u%n/%u%n", addr_octets, addr_octets+1, addr_octets+2, addr_octets+3, &dest_width, &mask_count, &dest_width );

#if READBGP_PARANOIA
        /* make sure we got at least the IP address
         *
         * since we already checked (above) that the first character
         * of destination IP address is not space, this is really
         * a paranoia check of malformed or addresses
         */
        if( converted < 4 )
        {
            fprintf( stderr, "warning on line %d: skipping malformed dest IP address '%s'\n", line, dest );
            return 0;
        }
#endif

        /* make sure that the 4 pieces of the IP address
         * are each in the range 0 to 255
         */
        addr = 0;
        for( count=0; count<4; count++ )
#if READBGP_PARANOIA
            if( addr_octets[count] > 255 )
            {
                fprintf( stderr, "error on line %d: byte %d >255\n", line, count );
                return ERROR_READBGP_OCTET_TOO_BIG;
            }
            else
#endif
                addr = (addr << 8) | addr_octets[count];

        /* if the mask bit count was specified...
         */
        if( converted == 5 )
        {
            /* make sure the mask bit count is in the range 0 to 32
             */
            if( mask_count > 32 )
            {
                fprintf( stderr, "error on line %d: mask bit count >32\n", line );
                return ERROR_READBGP_MASK_TOO_BIG;
            }
        }
        else
        {
            /* derive mask bit count from the class lookup table
             */
            mask_count = length_from_high_nibble[ (addr_octets[ 0 ] >> 4) & 15 ];

            if( mask_count > 32 )
            {
                fprintf( stderr, "error on line %d: cannot determine mask bit count; possible class E address\n", line );
                return ERROR_READBGP_MASK_UNKNOWN;
            }
        }

        /* look up the mask from its bit count
         */
        mask = mask_from_length[ mask_count ];

        /* restore the character we nulled to terminate dest IP address
         */
        dest[ MAX_DEST_IP_WIDTH ] = trashed_for_dest;

        /* if the destination IP address was wider than the default...
         */
        if( dest_width > DEF_DEST_IP_WIDTH )
        {
            /* adjust nexthop IP address pointer to compensate for longer dest IP address
             */
            nexthop_addr_string += dest_width - DEF_DEST_IP_WIDTH;

            /* adjust metric pointer to compensate for longer dest IP address
             */
            metric += dest_width - DEF_DEST_IP_WIDTH;
        }
    }

#if READBGP_IGNORES_NONPREF
    /* ignore the line if this is not a preferred route
     */
    if( *preferred != '>' )
        return 0;
#endif

#if READBGP_DEBUG_HELPER
    if( line == stopline )
        printf( "debugger wants to stop here\n" );
#endif

    /* if the metric is completely blank...
     */
    if( !strncmp( metric, "                          ", METRIC_WIDTH ) )

        /* place a zero on the right of the field
         */
    metric[ METRIC_WIDTH - 1 ] = '0';

    /* spit output
     */
#if READBGP_OUTPUTS_FLAGS
    printf( "*%c%c", *preferred, preferred[1] );
#endif
#if READBGP_OUTPUTS_BINARY
    /* convert destination address to binary
     */
    itoa( addr, dest_binary, 2 );

    /* find out how many binary digits it really bothered to fill in
     *
     * I sure wish I could just tell itoa() to use leading zeroes!
     * If I were converting to hex or octal, sprintf() would have
     * done the whole job.
     */
    count = strlen( dest_binary );

    /* insert leading zeroes
     */
    sprintf( dest_binary2, "%.*s%s", 32-count, "00000000000000000000000000000000", dest_binary );

    /* truncate based upon mask length
     */
    dest_binary2[ mask_count ] = '\0';

    printf( "%-32s ", dest_binary2 );
#endif
#if READBGP_OUTPUTS_DECIMAL
    sprintf( dest_binary2, "%s/%d", inet_ntoa(dest_binary,addr), mask_count );
    printf( "%-*s ", DEF_DEST_IP_WIDTH, dest_binary2 );
#endif
    printf( "%s\n", nexthop_addr_string );

    return 0;
}
#pragma warn .par



int read_bgp_file( const char *infile_name )
{
    char buffer[ MAX_TEXTLINE_SIZE ];
    int result;                                     /* successful by default */
    int result2;
    FILE *infile;
    unsigned count;
    unsigned line_before = line;

    result = init_bgp();
    if( result != 0 )
    {
        fprintf( stderr, "read_bgp_file(): error %d initializing BGP parsing\n", result );
        return result;
    }

    /* try to open the file
     */
    infile = fopen( infile_name, "rt" );
    if( infile == NULL )
    {
        fprintf( stderr, "readbgp: error %d '%s' opening file '%s'\n",
            errno, _sys_errlist[errno], infile_name );
        return ERROR_READBGP_FILE_OPEN;
    }

    /* as long as there is data...
     */
    while( !result && fgets( buffer, MAX_TEXTLINE_SIZE, infile ) )
    {
        line++;

        /* find out how long the text line is
         */
        count = strlen( buffer );

#if READBGP_STRIPS_LF
        /* strip off the trailing newline, if any,
         * by replacing it by a terminating NUL
         */
        if( buffer[ count - 1 ] == '\n' )
        {
            buffer[ count - 1 ] = '\0';
            count--;
        }
#endif

#if READBGP_STRIPS_CR
        /* also strip off any trailing carriage return,
         * the same way we did the newline above
         */
        if( buffer[ count - 1 ] == '\r' )
        {
            buffer[ count - 1 ] = '\0';
            count--;
        }
#endif

        result = parse_bgp_line( buffer, count );

    }   /* end for */

    /* show statistics for the parse
     */
    fprintf( stderr, "read_bgp_file( '%s' ) saw %d lines\n",
        infile_name,
        line - line_before
    );

    /* if the parse went OK
     * but file had an error...
     */
    if( !result && ferror( infile ) )
    {
        fprintf( stderr, "readbgp: error %d '%s' reading file '%s'\n",
            errno, _sys_errlist[errno], infile_name );

        result = ERROR_READBGP_FILE_READ;
    }

    /* close the input file
     */
    result2 = fclose( infile );
    if( result2!=0 && result==0 )
    {
        fprintf( stderr, "readbgp: error %d '%s' closing file '%s'\n",
            errno, _sys_errlist[errno], infile_name );

        result = ERROR_READBGP_FILE_CLOSE;
    }

    return result;
}



/* subnet structures created by this module will use
 * line numbers 1 through N
 *
 * this function tells you what that N is
 */
unsigned get_bgp_line_count( void )
{
    return line;
}



#define DUMP_BUF_SIZE       (2 * 1024 * 1024)     /* 2 MB */


int main( int argc, char *argv[] )
{
    int result;
    unsigned count;
    char *dump_buf;

    if( argc!=2 || argv[1][0]==0 )
    {
        printf( "usage: %s bgp-file-from-router\n", argv[0] );
        return ERROR_READBGP_BAD_PARMS;
    }

    /* initialize access to flat memory
     */
    flat_init( TRUE );

    result = read_bgp_file( argv[1] );
    if( result )
        return result;

    /* now how do I free up all the subnet radix trie?
     */

    return 0;
}



#endif  /* DO_FOREGROUND_STATS */


