/* conditional compile switches all go here
 *
 * manifest constants are still sprinkled
 * all through the code (sorry)
 */
#define STUB_0TH_LEVEL_STATS    FALSE           // just for debugging
#define STUB_2ND_LEVEL_STATS    FALSE           // just for debugging
#define STUB_DO_RECORD          FALSE           // just for debugging
#define STUB_FLOWS_EXPIRE       FALSE           // just for debugging
#define FORCE_1_BUCKET          FALSE           // just for debugging
#define FORCE_1_FLOW_PER_BUCKET FALSE           // just for debugging
#define ASSUME_OBJECT_IS_PQ     TRUE
#define ASSUME_ETHERTYPE_IS_IP  FALSE
#define TIME_ALREADY_SWAPPED    TRUE
#define BYTESWAP_FIELDS_EARLY   FALSE
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define USE_OLD_HASH            FALSE
#define FOLD_IS_JUST_MODULUS    FALSE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define DEBUG_CHAIN_LENGTH      FALSE
#define BUFFER_USED_PARANOIA    FALSE
#define BUFFER_PTR_PARANOIA     FALSE
#define FLOAT_EXCEPT_PARANOIA   FALSE
#define HASHINDEX_PARANOIA      FALSE
#define SUMMARIZE_ONE_PARANOIA  FALSE
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_COPY_CELL         FALSE           // stupid debugger!
#define DEBUG_FILLIN            FALSE
#define DEBUG_HOUSECLEAN        FALSE
#define FAIL_ON_PACKET_COUNT    FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE           // for debugging when overflow occurs only!  too slow
#define DELTA_TIME_PARANOIA     FALSE           // worry about negative delta time since last flows_expire()
#define DELTA_TIME_FIXUP        FALSE           // truncate large time since last flows_expire() to 1 second
#define DELTA_TIME_STATS        TRUE            // track min/max/mean for delta time since last flows_expire()
#define DEBUG_QUERY_TOTALS      TRUE
#define DEBUG_MAX_BPS           FALSE
#define MAX_BPS_PARANOIA        TRUE
#define QUERY_CALLS_EXPIRE      FALSE
#define DEBUG_ACTIVE_FEED       FALSE
#define DEBUG_ACTIVE_FEED_DATA  FALSE
#define EXPIRE_OVERHEAD_STATS   TRUE
#define REPORT_OVERHEAD_STATS   FALSE
#define DEBUG_EXTRA_TCP_STATS   FALSE
#define DEBUG_REASSEMBLY        FALSE
#define DEBUG_ORPHAN_FRAGS      FALSE
#define DEBUG_OVERLAP_FRAGS     FALSE
#define DEBUG_HOLE_FRAGS        FALSE
#define SHOW_CPU_UTILIZATION    TRUE
#define DEBUG_NON_IP            TRUE
#define DEBUG_NON_IP_PER_VPI    TRUE
#define COUNT_NON_IP_PER_VPI    TRUE
#define ASSUME_NNI_ADDRESSING   FALSE
#define RESELECTABLE_FLOWTRACE  TRUE
#define DEBUG_BLOCK_LOCKUP      FALSE
#define DEBUG_BLOCK_LOCKUP_2    FALSE



/*
OVERVIEW

this module is designed to be integrated into a
data capture system and called in realtime to convert
blocks of raw packet traces into "flows"

when we determine that a flow has finished (we haven't
seen any packets which hash the same in a user-defined
number of seconds) we pass it off to the statistics
module for further processing

occasionally, we receive a message on an out-of-band
channel (Ethernet when we are monitoring ATM) and we
tell the statistics module to finalize any info it has
for the flows we have already given it, and we send
those results back to whoever queried us, and then we
reset the statistics module




CREDITS

Written by Kevin Fall (kfall@sdsc.edu) 4/94 to 5/94
based on an earlier program by Hans-Werner Braun

change 4 mar 94 to verify whether protocol TCP (6) or UDP (17) and just
proceeds is same is the case. If not, it verifies whether the protocol
is ICMP (1), in which case only the two eight bit Type/Code fields are
retained. With all other protocols no information above the IP layer
is retained.

modified 12/5/95 by Joel Apisdorf:

to do hash chain length and concurrent flow count statistics

also changed hash algorithm to use a smaller, prime, number of buckets
and to use XOR and modulus to include all bits of all fields, rather
than using low 4 bits of each field and not combining them

also added -k command line parameter
so fillin messages can be turned off

also added -p command line parameter
so new flows messages can be turned off

modified 2/15/96 by Joel Apisdorf to use blocks of 128 packets
in the format captured by OC3MON.  Also changed the code
which ignores MAC-layer packets and non-first IP fragments to
understand the MAC-layer representation, rather than expecting
the fake flags seen in FR (FDDI receive) -generated traces.

modified 3/8/96 by Joel Apisdorf to coexist in same executable
and be called by data capture module, rather than be driven by
trace file on disk

modified 3/12/96 by Joel Apisdorf to support multiple queries
within the same run, so flows and stats can run continuously
without losing data, and so interval is controlled remotely,
like the NNSTAT agents

modified 10/30/96 by Kevin Thompson for a new option to spit
out active flow updates based on activity. Uses a 3rd linked list




10/28/97 got rid of end_block_stamp[] by changing tstamp into an array

*/



#include <sys/types.h>

#if __unix__
#   include <sys/time.h>
#   include <sys/resource.h>
#endif
#include "../wattcp/tcp.h"

#include <stdio.h>
#include <stdlib.h>             /* atoi() getenv() min() ?others? */
#include <fcntl.h>
#include <errno.h>
#include <float.h>              /* DBL_MIN DBL_MAX */

#include "../types/types.h"
#if !__unix__
#   include "..\intel\timestmp.h"
#   include "..\intel\byteswap.h"
#   include "..\intel\fpu.h"    /* get_fpu_status() clear_fpu_status() */
#endif

#include "tcp_seq.h"            /* SEQ_LT() */
#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"
#include "buffer.h"             /* class BufferBlock */
#include "fs2flows.h"           /* make sure we comply with our own prototypes */
#include "portstat.h"           /* porthash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "hoststat.h"           /* hosthash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "netstats.h"           /* nethash_alloc_failures       (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "fragstat.h"           /* fraghash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */



#define LARGEST_IP_SIZE         1500
#define UDP_PAYLOAD_SIZE        (LARGEST_IP_SIZE - 2*sizeof(struct ip) - sizeof(struct udphdr))
#define FLOWS_PER_UDP           (UDP_PAYLOAD_SIZE / sizeof(struct flowstats))
#if CARD_IS_OC12
#define EXPECTED_PACKETS        310000          // min per second (2 cells/packet, 1 occupied link)
#define EXPECTED_BYTES          68E6            // max per second
#else
#define EXPECTED_PACKETS        155000          // min per second (2 cells/packet, 1 occupied link)
#define EXPECTED_BYTES          17E6            // max per second
#endif
#define ACTIVE_SPEW_MILLISECS   1000    // how often should we spew active flows? (milliseconds)

#define FLOWTRACE_TCP_PORT         2000 // I made this up buddy, got a better idea?
#define CONTROL_TCP_PORT           23
#define QUERY_TCP_PORT             22   // use secure shell (ssh) port to get past firewall

#define QUERY_SOCKET_COUNT          2   // how many sockets to keep around

#define QUERY_BUFFER_SIZE       (3*1024*1024/2)     // 1.5 M bytes each

#define PKTFLOW_ARRAY_SIZE_MAX  1001    //klt (6/10/97) report #flows #pkts

//#define HASH_RECORDS 70000L       // fills after 40 seconds at core.wtn 9/19/96 4:00 PM
//#define HASH_RECORDS 100000L      // fills after 1'2" at core.wtn 9/19/96 4:07 PM
//#define HASH_RECORDS 140000L      // not full after 14'30" at core.wtn 9/19/96 4:23P but overflowed 10/5/96 3:10P - 3:25P
//#define HASH_RECORDS 200000L      // too little 10/22/96 4:34P
//#define HASH_RECORDS 400000L      // not even 375K + 60K fit when 32 MB allocated to DPMI server
//#define HASH_RECORDS 350000L      // fills at core.wtn 1/7/97 1:05 PM
//#define HASH_RECORDS 590000L      // 419K needed at core.wtn 1/24/97 11:50 AM
//#define HASH_RECORDS 550000L      // larger structure, cannot increase mem avail to malloc
//#define HASH_RECORDS 545000L      // fills at OC3 core.wtn 2/11/97 all day
//#define HASH_RECORDS 650000L      // how long will this last at WTN?
#define HASH_RECORDS 100000L        // use 100K (but 10K probably OK) for vBNS SCC, 200K for WOR BIPP, 650K for WTN BIPP

#if ASSUME_NNI_ADDRESSING
#   define MAX_VPI              4095
#   define ENTIRE_VPI(x)        ((x).GFC<<8 | (x).VPI)
#else
#   define MAX_VPI              255
#   define ENTIRE_VPI(x)        ((x).VPI)
#endif







// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



/*
 * user-selected object default
 */

#define DEFAULT_USER_PREF   "pq"


/*
 * how long to wait to time a flow out
 */

#define DEFAULT_FLOW_ACCESS_TIMEOUT      64.0   /* 64 seconds */
#define DEFAULT_FLOW_CREATE_TIMEOUT    3600.0   /* 1 hour */
static CardClock flow_access_timeout = DEFAULT_FLOW_ACCESS_TIMEOUT * CARD_CLOCK_RATE;
static CardClock flow_create_timeout = DEFAULT_FLOW_CREATE_TIMEOUT * CARD_CLOCK_RATE;


/*
 * housekeeping interval
 */

#define DEFAULT_HOUSEKEEP   1.0  /* seconds */
#define TVMULT          1.5 /* multiplier for "fillins" */
static CardClock housekeep_timeout = DEFAULT_HOUSEKEEP * CARD_CLOCK_RATE;

/* global status info
 */

#if DEBUG_BLOCK_LOCKUP_2
    unsigned reccount_trace = 3278;
#endif
#if DEBUG_HASHINDEX
    u_int hashindex_trace = -1;
#endif

u_int ip_length_usage[ IP_LEN_COUNT ];
u_int ip_precedence_usage[ IP_PREC_COUNT ];
Bit64 sum_bytes_per_packet[ MAX_ATM_NICS ];
Bit16 max_bytes_per_packet[ MAX_ATM_NICS ];
u_int sum_packets[ MAX_ATM_NICS ];
u_int max_packets[ MAX_ATM_NICS ];
u_int cleaned_packets[ MAX_ATM_NICS ];
Bit64 cleaned_bytes[ MAX_ATM_NICS ];
u_int cleaned_flows;
u_int all_packets_trace = -1;
u_int all_packets[ MAX_ATM_NICS ];
u_int ip_optioned_packets[ MAX_ATM_NICS ];
u_int non_ip_packets[ MAX_ATM_NICS ];
#if COUNT_NON_IP_PER_VPI
    u_int non_ip_packets_per_vpi[ MAX_ATM_NICS ][ MAX_VPI + 1 ];
#endif
#if REASM_IP_FRAGS
    u_int orphan_non_first_ip_fragments[ MAX_ATM_NICS ];
    u_int overlap_non_first_ip_fragments[ MAX_ATM_NICS ];
    u_int hole_non_first_ip_fragments[ MAX_ATM_NICS ];
#endif
u_int non_first_ip_fragments[ MAX_ATM_NICS ];
u_int first_ip_fragments[ MAX_ATM_NICS ];
u_int tcp_optioned_packets[ MAX_ATM_NICS ];
#if EXTRA_TCP_STATS && CELLS_PER_PACKET>1
    u_int unexpected_tcp_option_lengths[ MAX_ATM_NICS ];
#endif
Bit64 all_bytes[ MAX_ATM_NICS ];
u_int sum_new_flows;
Bit64 sum_bytes[ MAX_ATM_NICS ];
Bit64 max_bytes[ MAX_ATM_NICS ];
Bit16 instant_max_bytes[ MAX_ATM_NICS ];
u_int sum_flows;
u_int max_flows;
u_int max_new_flows;
int report_already_set;
int istamp_already_set;
int last_already_set[ MAX_ATM_NICS ];
u_int query_count;                /* how many times have we been queried? */
u_int total_flows;                /* total ever processed */
u_int deleted_flows;              /* total ever deleted */
u_int ndeleted_before_clean;      /* tot deleted before cleanup */
static CardClock istamp;          /* time on 1st record */
static CardClock tstamp[ MAX_ATM_NICS ];          /* current time stamp */

u_int pktflow[PKTFLOW_ARRAY_SIZE_MAX];

#if EXPIRE_OVERHEAD_STATS
    u_int sum_deleted_flows;
    u_int max_deleted_flows;
#endif

#if DELTA_TIME_STATS
    static double min_houseclean_interval;
    static double max_houseclean_interval;
    static double sum_houseclean_interval;
#endif

#if SHOW_CPU_UTILIZATION
    Bit32 blocks;
    PentiumClock before, elapsed;
    PentiumClock startup;
    PentiumClock min_elapsed, max_elapsed, sum_elapsed;
#endif

#if BUFFER_USED_PARANOIA
    static Bit32 blocks_overrun;
#endif

#if MAX_BPS_PARANOIA
    static Bit32 bitrate_overrun;
#endif

/* stuff related to command line flags */
u_int pflag = FALSE, rflag = FALSE; /* cmd line flags */
u_int active_feed = FALSE; /* klt - 10/30/96 */
u_int fillin_wanted = TRUE;
int objtype;
char *objname = DEFAULT_USER_PREF;
CardClock next_report_stamp;
CardClock last_report_stamp[ MAX_ATM_NICS ];

/* TCP/IP socket stuff
 */
tcp_Socket flowtrace_socket;
tcp_Socket control_socket;
tcp_Socket query_sockets[ QUERY_SOCKET_COUNT ];
Bit8 *query_output_buffers[ QUERY_SOCKET_COUNT ];
unsigned which_query_socket_to_use;
flowselector flowtrace_parms;
sockaddr flowtrace_caller;
int flowtrace_parms_are_filled = FALSE;
int flowtrace_is_connected = FALSE;
char *flowtrace_client_string = NULL;
Bit32 flowtrace_client_address = 0;

/* must measure time until socket is closed,
 * not time until flows_query() returns,
 * to know how slow network was, now that
 * TCP/IP stack buffers the entire report
 */
static unsigned which_query_socket_prev_used;
static PentiumClock query_start_time;
static double query_drainage_delta;

struct user_prefs {
    struct flowdata up_fd;  /* not completely used, but handy */
} prefs;

#define up_src      up_fd.fd_src
#define up_dst      up_fd.fd_dst
#define up_sport    up_fd.fd_sport
#define up_dport    up_fd.fd_dport
#define up_index    up_fd.fd_index

/*
 * general hashing routines
 *
 */


#if USE_OLD_HASH
#define NFLOWBUCKETS    (1024*1024)
#define flowhash(fd) ( (((u_int)(fd)->fd_src)&0xf << 16) | \
                  (((u_int)(fd)->fd_dst)&0xf << 12) | \
                  (((u_int)(fd)->fd_proto)&0xf << 8) | \
                  (((u_int)(fd)->fd_sport)&0xf << 4) | \
                  (((u_int)(fd)->fd_dport)&0xf) )
#else
/*
// primes
#define NFLOWBUCKETS    251
#define NFLOWBUCKETS    1021
#define NFLOWBUCKETS    4001
#define NFLOWBUCKETS    8191
#define NFLOWBUCKETS    32749
#define NFLOWBUCKETS    104729
#define NFLOWBUCKETS    131071
#define NFLOWBUCKETS    262139
#define NFLOWBUCKETS    262147

// powers of 2
#define NFLOWBUCKETS    65536
#define NFLOWBUCKETS    131072
#define NFLOWBUCKETS    262144
*/
#define NFLOWBUCKETS    65536
#define BUCKET_POOL_SIZE (NFLOWBUCKETS * sizeof(struct hashrec *))
static struct hashrec **flowtab = NULL;

#if DEBUG_CHAIN_LENGTH
#   define CHAIN_POOL_SIZE (NFLOWBUCKETS * sizeof(int))
    static int *chainlen = NULL;
    static int longest_chain = 0;
    static u_int chainlen_sum = 0;
    static int chains_used = 0;
#endif

#if FOLD_IS_JUST_MODULUS
#   define flowfold(x) ((x) % NFLOWBUCKETS)
#else
#if FOLD_RANDOMIZES
#   define flowfold(x) (((x)*1664525 + 1013904223) >> 16)
//#   define flowfold(x) (((x)*1664525 + 1013904223) >> 15)
//#   define flowfold(x) (((x)*1664525 + 1013904223) >> 14)
#else
#if FOLD_KEEPS_16_BITS
#   define flowfold(x) fold_32_to_16(x)
#else
#   define flowfold(x) fold_32_to_8(x)
#endif
#endif
#endif

#if BIT64_AVAILABLE
/* generate a 32-bit hash key using 64-bit intermediates
 */
#define flowhash(fd) fold_64_to_32( \
                  (*(Bit64 *)(fd)) ^ \
                  (*(Bit64 *)(8 + (Bit8 *)(fd)) >> 16) \
                  )
#else
/* generate a 32-bit hash key directly
 */
#define flowhash(fd) ( \
                  ((Bit32)(fd)->fd_src.s_addr) ^ \
                  ((Bit32)(fd)->fd_dst.s_addr) ^ \
                  (*(Bit32 *)(12 + (Bit8 *)(fd))) ^ \
                  ((Bit32)(fd)->fd_proto) \
                  )
#endif      /* generate 32 bit key */
#endif      /* use new hash */


#define HASH_POOL_SIZE (HASH_RECORDS * sizeof(struct hashrec))
struct hashrec *hash_free = NULL;
struct hashrec *hash_pool = NULL;
u_int hash_alloc_failures;

/* klt - global stuff for active feed */
    u_char mttl = 255;    /* jza - 11/13/96 multicast time to live */
    struct in_addr taddr = {0}; /* jza - 11/13/96 address to tunnel multicast through */
    struct in_addr maddr; /* klt - 10/30/96 mcast address */
    udp_Socket msock;   /* klt - 11/10/96 the socket we'll get on a udp_open */
#define ACTIVE_FEED_PORT 6507   // klt - 11-10-96 port to send active feed stuff out to


int flows_new( u_int indx, flowdata *fd, CardClock *timestamp, cellhashrec *rp );
void flows_delete( struct hashrec *hr );
void flows_usage( char *prog );
void flows_summarize_all( void );
void flows_summarize_one( char *cp, struct user_prefs *up, struct hashrec *hr );
int flows_clear( void );


#if !__unix__
/* UNIX-compatible command-line options processing
 */
char *optarg = 0;
unsigned optcount = 0;

int getopt(
    int count,
    char **strings,
    char *pattern
)
{
    char *patpos;
    int result;

    for( ; optcount<count; optcount++ )
    {
        // does the current arg start with dash?
        //
        if( strings[optcount][0] == '-' )
        {
            optarg = strings[optcount]+1;
            for( patpos=pattern; patpos; patpos++ )

                // if this is an option the caller wants...
                //
                if( *optarg == *patpos )
                {
                    // remember the option character
                    // so we can return it
                    //
                    result = *optarg;

                    // move forward to the next argument
                    // so we know our place next time
                    //
                    optcount++;

                    // if caller wants a value attached to this arg...
                    //
                    patpos++;
                    if( *patpos == ':' )
                    {
                        // use up an argument
                        //
                        optarg = strings[ optcount ];
                        optcount++;
                    }

                    return result;
                }
        }
        // else ignore the argument completely
        //
    }

    optcount = 0;
    return -1;
}
#endif  /* not unix */




static void parse_tcp_options( Bit8 interface, cellhashrec *rp, hashrec *hp )
{
    // if there are no TCP options in this packet, do nothing
    //
    unsigned options_sent = (rp->tr_tcplen - 5) * 4;
    if( options_sent == 0 )
        return;

    // count TCP packets with options
    //
    tcp_optioned_packets[ interface ]++;

#if EXTRA_TCP_STATS && CELLS_PER_PACKET>1
    // if the connection is not just being set up, do nothing
    //
    if( rp->tr_tcpflags & TH_SYN == 0 )
        return;

    // parse options we received
    //
    unsigned options_received = 0;
    unsigned options_parsed = 0;
    Bit8 options[ 40 ];
    Bit8 length;
    register Bit8 this_option;
    register Bit8 wscale;

    if( rp->tr_gotpay1to6 )
    {
        memcpy( options, rp->tr_pay1to6, 6 );
        options_received = min( options_sent, 6U );
    }

    if( rp->tr_gotpay7to8 )
    {
        memcpy( options+6, rp->tr_pay7to8, 2 );
        options_received = min( options_sent, 8U );
    }

    if( rp->tr_gotpay9up )
    {
        memcpy( options+8, rp->tr_pay9up, 32 );
        options_received = min( options_sent, 40U );
    }

    if( options_received < options_parsed )

        // remember that card not copying 2nd cell
        // lost us some TCP options
        //
        hp->hr_tcpopts |= fs_tcp_opt_truncated;

    while( options_parsed < options_received )
    {
        this_option = options[ options_parsed ];
        switch( this_option )
        {
            case TCPOPT_EOL:
                return;

            case TCPOPT_NOP:
                length = 1;
                break;

            default:
                if( options_parsed + 1 >= options_received )
                {
                    // remember that there was no room
                    // for an expected length
                    //
                    hp->hr_tcpopts |= fs_tcp_opt_truncated;
                    return;
                }

                length = options[ options_parsed + 1 ];
                if( length < 2 )
                {
                    hp->hr_tcpopts |= fs_tcp_opt_too_short;
                    return;
                }

                switch( this_option )
                {
                    case TCPOPT_WINDOW:
                        if( length != 3 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        if( length < 3 )
                        {
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_win_scale;
                        wscale = options[ options_parsed + 2 ];
                        if( wscale > TCP_MAX_WINSHIFT )
                            wscale = TCP_MAX_WINSHIFT + 1;
                        hp->hr_wscale = wscale;
                        break;

                    case TCPOPT_MAXSEG:
                        if( length != 4 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        if( length < 4 )
                        {
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_max_segment;
                        hp->hr_mss = (options[ options_parsed + 2 ] << 8)
                                   |  options[ options_parsed + 3 ];
                        break;

                    case TCPOPT_TSTAMP:
                        if( length != 10 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        if( length < 10 )
                        {
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_timestamp;
                        break;

                    case TCPOPT_SACKPERM:
                        if( length != 2 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        // must be 2 or greater since we tossed 0 and 1 above
                        hp->hr_tcpopts |= fs_tcp_opt_sack_permitted;
                        break;

                    case TCPOPT_SACK:
                        if( (length<10) || ((length - 10) % 8) )
                        {
                            unexpected_tcp_option_lengths[ interface ]++;
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_sack;
                        break;

                    case TCPOPT_CC:
                        if( length != 6 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        if( length < 6 )
                        {
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_ttcp_cc;
                        break;

                    case TCPOPT_CC_NEW:
                        if( length != 6 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        if( length < 6 )
                        {
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_ttcp_cc_new;
                        break;

                    case TCPOPT_CC_ECHO:
                        if( length != 6 )
                            unexpected_tcp_option_lengths[ interface ]++;
                        if( length < 6 )
                        {
                            hp->hr_tcpopts |= fs_tcp_opt_too_short;
                            return;
                        }
                        hp->hr_tcpopts |= fs_tcp_opt_ttcp_cc_echo;
                        break;

                    default:
                        hp->hr_tcpopts |= fs_tcp_opt_unknown;
                        break;

                }
                break;
        }

        options_parsed += length;
    }
#endif
}



/*
 * do_record
 *  1> look up record... if there update access time
 *  2> if not there, insert it
 */

static int do_record( struct cellhashrec *rp, Bit8 interface )
{
    u_int hashindex;
    struct hashrec *hp;
    struct flowdata fd;

    /*
     * set fields appropriately, based on user prefs
     * must be done before hash index is computed
     */
#if ASSUME_OBJECT_IS_PQ
    update_pq( rp, &fd, interface );
#else
    (*objtypes[objtype].update)( rp, &fd, interface );
#endif

#if FORCE_1_BUCKET
    fd.fd_index = 0;
#else
    fd.fd_index = flowhash( &fd );
#endif
    hashindex = flowfold( fd.fd_index );

#if HASHINDEX_PARANOIA
    if( hashindex >= NFLOWBUCKETS )
        printf( "internal error: flowhash %u > %d\n", hashindex, NFLOWBUCKETS );
#endif

    hp = flowtab[hashindex];

#if !ALWAYS_DUMP_HASH_INFO
    if( all_packets[ interface ] == all_packets_trace )
#endif
        fprintf( stderr, "interface %d packet %d hash index 0x%8.8lX from IP=%8.8X:%8.8X proto=%d TCP=%4.4X:%4.4X\n",
            interface, all_packets[interface], hashindex, fd.fd_src, fd.fd_dst, fd.fd_proto, fd.fd_sport, fd.fd_dport );

    if (pflag && rflag && flowcmp(&prefs.up_fd, &fd)) {
        CardClock tmp = tstamp[interface] - istamp;
        printf("FlowTrace:\t%.6f\t%d\t%x:%x\t%d:%d:%d\t%.6f\n",
            tmp.get_seconds(),
            fd.fd_len,
            fd.fd_src,
            fd.fd_dst,
            fd.fd_proto,
            fd.fd_sport,
            fd.fd_dport,
            tstamp[interface].get_seconds()
        );
    }

    /* look for match or collision on hash chain
     */
    while( hp != NULL )
    {
#if !FORCE_1_FLOW_PER_BUCKET
        // if we found an existing flow...
        //
        int cmp = flowcmp( &hp->hr_fd, &fd );
        if( cmp )
#endif
        {
            // update the fields of the old record
            //
            hp->hr_packets++;
            hp->hr_bytes += Bit32( fd.fd_len );

#if REASM_IP_FRAGS
            // if this flow is on the fragments list, take it off
            //
            if( hp->hr_flink.prev != (hashrec *)-1 )
                delete_fragment_queue( hp );

            // if this packet has more fragments following...
            //
            if( rp->tr_offset & IP_MF )
            {
                // put this flow on the fragments list
                //
                insert_fragment_queue( hp );

                // fill in the fields that matter to a fragment
                //
                hp->hr_len = fd.fd_len - sizeof( ip );
                hp->hr_ident = rp->tr_ident;
            }
#endif
#if KEEP_ATM_HEADER
            // no need to update the ATM header, since data coming
            // in on another interface would be another flow, and we
            // assume that all packets for a flow came in on the same
            // ATM VP/VC
            //
#endif
            // no need to deal with out-of-order timestamps
            // in the packets any more, since we define the
            // interface as part of the uniqueness of a flow
            //
            hp->hr_access = tstamp[ interface ];

            /* if object type (statistic to track) is "portquad"...
             */
#if !ASSUME_OBJECT_IS_PQ
            if( objindex == PORTQUAD_OBJ_INDEX )
#endif
            {
                // if this is a TCP flow...
                //
                if( fd.fd_proto == IPPROTO_TCP )
                {
                    parse_tcp_options( interface, rp, hp );

#if EXTRA_TCP_STATS
                    Bit8 flags = rp->tr_tcpflags;
                    Bit32 seq = ntohl( rp->tr_seq );
                    Bit32 ack = ntohl( rp->tr_ack );
                    Bit16 win = ntohs( rp->tr_win );


                    // if the acknowledge field is valid in this packet...
                    //
                    if( flags & TH_ACK )
                    {
                        // if we have seen an acknowledge on this flow before...
                        //
                        if( hp->hr_tcpflags & TH_ACK )
                        {
                            // if this is a duplicate acknowledge...
                            //
                            if( ack == hp->hr_prevack
                            &&  seq == hp->hr_prevseq
                            &&  win == hp->hr_prevwin )
                            {
                                saturated_inc_bit8( hp->hr_dupacks );
                            }
                            else
                            {
                                // remember the current sequence number and window,
                                // so we can detect duplicate ACK in next packet
                                //
                                hp->hr_prevack = ack;
                                hp->hr_prevseq = seq;
                                hp->hr_prevwin = win;

                                // track the minimum and maximum acknowledge numbers
                                //
                                if( SEQ_GT( hp->hr_minack, ack ) )
                                    hp->hr_minack = ack;
                                else if( SEQ_LT( hp->hr_maxack, ack ) )
                                    hp->hr_maxack = ack;
                            }
                        }
                        else
                        {
                            // remember the current sequence number and window,
                            // so we can detect duplicate ACK in next packet
                            //
                            hp->hr_prevack = ack;
                            hp->hr_prevseq = seq;
                            hp->hr_prevwin = win;

                            // this is definitely the minimum and maximum ack
                            //
                            hp->hr_minack = ack;
                            hp->hr_maxack = ack;

                        }   // end else first ack on this flow

                    }   // end if ack field is valid

#if DEBUG_EXTRA_TCP_STATS
                    if( hp == (struct hashrec *)0x1f584cc0 )
                    {
                        //dump_flow( stdout, hp );
                        printf( "IP_len=%u, TCP_SEQ=%u\n", fd.fd_len, seq );
                    }
#endif

                    // remember which TCP flags have been seen so far
                    //
                    hp->hr_tcpflags |= flags;

                    // calculate the right edge of this packet
                    //
                    // no need to look at IP header length, since
                    // flows_per_block already tossed packets with
                    // IP options
                    //
                    Bit32 maxseq = seq;
                    Signed32 tcp_len = fd.fd_len - sizeof(ip) - 4*rp->tr_tcplen;
                    if( tcp_len < 0 )
                        hp->hr_badlen++;
                    else
                        maxseq += tcp_len;

                    // count FIN and SYN as one byte each
                    //
                    // note that they both get a chance to increment
                    // maxseq, in case they both occur in same packet
                    //
                    if( flags & TH_SYN )
                        maxseq++;
                    if( flags & TH_FIN )
                        maxseq++;

                    // if there is no hole now...
                    //
                    if( hp->hr_minhole == hp->hr_maxhole )
                    {
                        // if this packet is far to the left of the flow...
                        //
                        if( SEQ_LT( maxseq, hp->hr_minseq ) )
                        {
                            // define a new hole between right of
                            // packet and left of flow
                            //
                            hp->hr_minhole = maxseq;
                            hp->hr_maxhole = hp->hr_minseq;

                            // count "single holes"
                            //
                            saturated_inc_bit8( hp->hr_1holes );
                        }

                        // otherwise, if this packet is far to the right of the flow...
                        //
                        else if( SEQ_GT( seq, hp->hr_maxseq ) )
                        {
                            // define a new hole between right of
                            // flow and left of packet
                            //
                            hp->hr_minhole = hp->hr_maxseq;
                            hp->hr_maxhole = seq;

                            // count "single holes"
                            //
                            saturated_inc_bit8( hp->hr_1holes );
                        }

                        // otherwise, if this packet overlaps the left
                        // or right edge of the flow...
                        //
                        else if( SEQ_LT( seq,    hp->hr_maxseq )
                        &&       SEQ_GT( maxseq, hp->hr_minseq ) )
                        {
                            // count the packet as a "duplicate"
                            //
                            saturated_inc_bit8( hp->hr_dupseqs );
                        }

                    }   // end if no hole yet

                    // otherwise, there must be one or more holes now
                    //
                    else
                    {
                        // if packet is outside of flow...
                        //
                        if( SEQ_LT( maxseq, hp->hr_minseq )         // left of flow
                        ||  SEQ_GT(    seq, hp->hr_maxseq ) )       // right of flow
                        {
                            // count "multiple holes"
                            //
                            saturated_inc_bit8( hp->hr_nholes );

                            // leave the hole where it was
                            //
                        }

                        // otherwise, if we have seen more than one hole...
                        //
                        else if( hp->hr_nholes > 0 )
                        {
                            // if this packet overlaps the left or right
                            // edge of the flow...
                            //
                            if( SEQ_LT( seq,    hp->hr_maxseq )
                            &&  SEQ_GT( maxseq, hp->hr_minseq ) )
                            {
                                // count the packet as a "backtrack", since
                                // we don't know whether it is a reorder or
                                // a retransmit (loss), since we don't know
                                // how thick the flow edges are
                                //
                                saturated_inc_bit8( hp->hr_backseqs );
                            }
                        }

                        // otherwise, we must know about exactly
                        // one hole, so if packet is inside hole...
                        //
                        else if( SEQ_LT( maxseq, hp->hr_maxhole )
                        &&       SEQ_GT(    seq, hp->hr_minhole ) )
                        {
                            // count "multiple holes"
                            //
                            saturated_inc_bit8( hp->hr_nholes );

                            // leave the hole where it was
                            //
                        }

                        // otherwise, packet must overlap or abut
                        // the edge of flow and/or of hole...
                        //
                        else
                        {
                            // if packet overlaps either edge
                            // of flow or hole...
                            //
                            // we need to check both sides of the packet,
                            // even though we looked at the other side
                            // above with almost the opposite comparison
                            // of the first of each pair here, since we
                            // didn't do the equality test above
                            //
                            // this sure is a lot of comparisons for a
                            // section of code in the "fast path".  Rats!
                            //
                            if( (SEQ_GT( maxseq, hp->hr_minseq ) &&
                                 SEQ_LT(    seq, hp->hr_minseq ))   // left  of flow
                            ||  (SEQ_LT(    seq, hp->hr_maxseq ) &&
                                 SEQ_GT( maxseq, hp->hr_maxseq ))   // right of flow
                            ||  (SEQ_GT( maxseq, hp->hr_minhole ) &&
                                 SEQ_LT(    seq, hp->hr_minhole ))  // left  of hole
                            ||  (SEQ_LT(    seq, hp->hr_maxhole ) &&
                                 SEQ_GT( maxseq, hp->hr_maxhole ))) // right of hole
                            {
                                // count the packet as a "duplicate"
                                //
                                saturated_inc_bit8( hp->hr_dupseqs );
                            }

                            // update the edges of the hole
                            //

                            // if packet covers some edge of hole...
                            //
                            if( SEQ_GT( maxseq, hp->hr_minhole )
                            &&  SEQ_LT(    seq, hp->hr_maxhole ) )
                            {
                                // if packet ends before right of hole...
                                //
                                if( SEQ_LT( maxseq, hp->hr_maxhole ) )
                                {
                                    // hole right edge moved
                                    //
                                    hp->hr_maxhole = maxseq;

                                    // count packets received in single holes
                                    // after the one that defined the hole
                                    // (that one is counted in hp->hr_1holes)
                                    // now that we know this packet will not
                                    // fill in the hole
                                    //
                                    saturated_inc_bit8( hp->hr_holepkts );
                                }

                                // otherwise, if packet starts after
                                // left of hole...
                                //
                                else if( SEQ_GT( seq, hp->hr_minhole ) )
                                {
                                    // hole left edge moved
                                    //
                                    hp->hr_minhole = seq;

                                    // count packets received in single holes
                                    // after the one that defined the hole
                                    // (that one is counted in hp->hr_1holes)
                                    // now that we know this packet will not
                                    // fill in the hole
                                    //
                                    saturated_inc_bit8( hp->hr_holepkts );
                                }

                                // otherwise, there isn't a hole anymore
                                //
                                else
                                    hp->hr_maxhole = hp->hr_minhole;

                            }   // end if packet covers an edge of hole

                        }   // end else packet must abut/overlap flow/hole

                    }   // end else one or more holes now

                    // update left and right edges of flow
                    //
                    if( SEQ_LT( seq, hp->hr_minseq ) )
                        hp->hr_minseq = seq;
                    if( SEQ_GT( maxseq, hp->hr_maxseq ) )
                        hp->hr_maxseq = maxseq;
#endif  // EXTRA_TCP_STATS

                }   // end if TCP flow

            }   // end if object is port quad

            // move the flow to the end of the
            // time-sorted queue of access times
            //
            // note that since packets could be out-of-order,
            // endque() ought to be stepping back through linked list
            // to make sure that this gets inserted where its access
            // time is less than next in list, but that isn't
            // implemented yet!
            //
            endque( hp );

            /* klt - 10/30/96 - update active list */
            /* if we're in active feed mode        */
            if( active_feed )
            {
#if DEBUG_ACTIVE_FEED
                fprintf(stderr,"do_record() calling update_active_queue()\n");
#endif
                update_active_queue( hp );
            }

            return 0;
        }
        hp = hp->hr_clink.next;
    }

    return flows_new( hashindex, &fd, tstamp+interface, rp );
}


/*
 *  establish a new flow
 *  1> insert into hash table [maintained as stack]
 *  2> increment chain length
 *  3> insert at end of access-time queue
 */

static int
flows_new(
    u_int indx,                  /* hash index */
    flowdata *fd,                /* record */
    CardClock *timestamp,
    cellhashrec *rp
)
{
    struct hashrec *hr, *tmp;

    if ((hr = hash_free) == NULL)
    {
        hash_alloc_failures++;
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new hashrec\n");
        return ERROR_FLOWS_CANNOT_SUBALLOC_FLOW;
#else
        return 0;
#endif
    }
    hash_free = hr->hr_clink.next;

    if (pflag && !rflag) {
        double tmp = CardClock(*timestamp - istamp).get_seconds();
        printf("N %.6f\t%x:%x\t0\t0\t0\t%d:%d:%d\t%.6f\n",
            tmp,
            fd->fd_src, fd->fd_dst, fd->fd_proto,
            fd->fd_sport, fd->fd_dport,
            timestamp->get_seconds()
        );
    }
    hr->hr_access = hr->hr_start = *timestamp;
    hr->hr_clink.next = hr->hr_clink.prev = NULL;
    hr->hr_alink.next = hr->hr_alink.prev = NULL;
    hr->hr_fd = *fd; /* copy from trace */
    hr->hr_bytes = Bit32( fd->fd_len );
    hr->hr_packets = 1;
#if REASM_IP_FRAGS
    hr->hr_len -= sizeof( ip );
    hr->hr_ident = rp->tr_ident;
    hr->hr_olaps = 0;
    hr->hr_holes = 0;
    if( rp->tr_offset & IP_MF )         // if more fragments...
    {
        insert_fragment_queue( hr );
    }
    else
    {
        hr->hr_flink.prev = (hashrec *) -1;
        hr->hr_flink.next = (hashrec *) -1;
    }
#endif
#if KEEP_ATM_HEADER
    hr->hr_atm.whole = rp->tr_atm.whole;
#endif
    if( hr->hr_fd.fd_proto == IPPROTO_TCP )
    {
#if EXTRA_TCP_STATS
        Bit8 flags = rp->tr_tcpflags;
        hr->hr_tcpflags = flags;
        hr->hr_prevseq = ntohl( rp->tr_seq );
#if DEBUG_EXTRA_TCP_STATS
        if( hr == (struct hashrec *)0x1f584cc0 )
        {
            //dump_flow( stdout, hr );
            printf( "IP_len=%u, TCP_SEQ=%u\n", fd->fd_len, hr->hr_prevseq );
        }
#endif
        hr->hr_minseq = hr->hr_prevseq;
        Signed32 tcp_len = fd->fd_len - sizeof(ip) - 4*rp->tr_tcplen;
        if( tcp_len < 0 )
        {
            hr->hr_badlen = 1;
            hr->hr_maxseq = hr->hr_minseq;
        }
        else
        {
            hr->hr_badlen = 0;
            hr->hr_maxseq = hr->hr_minseq + tcp_len;
        }
        if( flags & TH_SYN )                    // see do_record() for reasoning
            hr->hr_maxseq++;
        if( flags & TH_FIN )                    // see do_record() for reasoning
            hr->hr_maxseq++;
        hr->hr_minhole  = 1234567890;           // could be any value; this stands out to human eye
        hr->hr_maxhole  = hr->hr_minhole;
        hr->hr_nholes   = FALSE;
        hr->hr_1holes   = 0;
        hr->hr_holepkts = 0;
        hr->hr_backseqs = 0;
        hr->hr_dupseqs  = 0;
        if( flags & TH_ACK )
        {
            hr->hr_prevack = ntohl( rp->tr_ack );
            hr->hr_minack  = hr->hr_prevack;
            hr->hr_maxack  = hr->hr_prevack;
        }
        else
        {
            // not strictly necessary, but nice for printouts
            //
            hr->hr_prevack = 1234567890;
            hr->hr_minack = BIT32_MAX;
            hr->hr_maxack = 0;
        }
        hr->hr_dupacks = 0;
        hr->hr_prevwin = ntohs( rp->tr_win );
        hr->hr_maxwin  = hr->hr_prevwin;
#if CELLS_PER_PACKET > 1
        hr->hr_tcpopts = 0;
        hr->hr_wscale = 0;
        hr->hr_mss = 0;
#endif
#endif  // EXTRA_TCP_STATS

        parse_tcp_options( fd->fd_iface, rp, hr );

    }   // if protocol is TCP

    insque( hr );

    total_flows++;
    u_int pending_flows = total_flows - deleted_flows;
    if( max_flows < pending_flows )
        max_flows = pending_flows;

#if DEBUG_CHAIN_LENGTH
    /*
     * increment chain length for this hash index
     */
    chainlen[indx]++;
    chainlen_sum++;

#if DEBUG_HASHINDEX
    if( indx == hashindex_trace )
        fprintf( stderr, "hash index 0x%.8X chain length increased to %d\n", indx, chainlen[indx] );
#endif
#endif

    /* klt - 10/30/96 - add to active list */
    /* if we're in active feed mode        */
    if (active_feed)
    {
#if DEBUG_ACTIVE_FEED
        fprintf(stderr,"flows_new() calling update_active_queue()\n");
#endif
        update_active_queue( hr );
    }

    /*
     * handle the case where no flows existed
     * on this hash chain until now
     */
    if( (tmp = flowtab[indx]) == NULL )
    {
#if DEBUG_CHAIN_LENGTH
        chains_used++;
#endif
        flowtab[indx] = hr;
        return 0;
    }

    /*
     * at least one flow already there (we collided earlier)
     * insert at the head [this maintains a stack]
     */

    hr->hr_clink.next = tmp;
    tmp->hr_clink.prev = hr;
    hr->hr_clink.prev = NULL;
    flowtab[indx] = hr;

#if DEBUG_CHAIN_LENGTH
    /*
     * if this is the longest chain we have ever seen,
     * update our global, and notify the user
     *
     * notifications should be asymptotically decreasing
     */
    if( longest_chain < chainlen[indx] )
    {
        longest_chain = chainlen[indx];
        fprintf( stderr, "hash index 0x%8.8lX is now longest chain at %d\n", indx, longest_chain );
        fprintf( stderr, "calced from IP=%8.8X:%8.8X proto=%d TCP=%4.4X:%4.4X\n", fd->fd_src, fd->fd_dst, fd->fd_proto, fd->fd_sport, fd->fd_dport );
        fprintf( stderr, "average chain length is %.3f\n", chainlen_sum/double(chains_used) );
        fprintf( stderr, "%u (%.1f%%) of %u buckets used\n", chains_used,
            (100.0*chains_used)/NFLOWBUCKETS, NFLOWBUCKETS );
    }
#endif

    return 0;
}

/*
 * flows_delete
 *  decrement chain length for that hash index
 *  and delete an entry from the hash table
 */

static void flows_delete( struct hashrec *hr )
{
    u_int indx = flowfold( hr->hr_index );

#if DEBUG_CHAIN_LENGTH
    chainlen[indx]--;
    chainlen_sum--;
    if( chainlen[indx] < 0 )
    {
        fprintf( stderr, "warning: %d chain length for hash index 0x%8.8lX\n", chainlen[indx], indx );
    }

#if DEBUG_HASHINDEX
    if( indx == hashindex_trace )
        fprintf( stderr, "hash index 0x%8.8lX chain length decreased to %d\n", indx, chainlen[indx] );
#endif
#endif

#if REASM_IP_FRAGS
    // if this flow was on the IP fragments list, take it off
    //
    if( hr->hr_flink.prev != (hashrec *) -1 )
        delete_fragment_queue( hr );
#endif

    if (hr->hr_clink.prev != NULL)
        hr->hr_clink.prev->hr_clink.next = hr->hr_clink.next;
    else
    {
        flowtab[indx] = hr->hr_clink.next;
#if DEBUG_CHAIN_LENGTH
        if( !flowtab[indx] )
            chains_used--;
#endif
    }

    if (hr->hr_clink.next != NULL)
        hr->hr_clink.next->hr_clink.prev = hr->hr_clink.prev;

    hr->hr_clink.next = hash_free;
    hash_free = hr;

    deleted_flows++;
}



/*
 * fill in seconds that may not have been printed due
 * to lack of trace data.  Makes sure we have at least some
 * trace record every housekeep_timeout*TVMULT secs
 */

static void flows_fillin_seconds( CardClock *ostamp, CardClock *stamp )
{
    CardClock tmp, addto, j;
    addto = TVMULT * (double) housekeep_timeout;
    tmp = *ostamp + addto;


#if DEBUG_FILLIN
    printf( "LOOPSTUFF: %g, %g\n", tmp.get_seconds(), stamp->get_seconds() );
#endif

    if (tmp < *stamp) {
        addto.put_seconds( 1 );
        for (j = tmp; j<=*stamp; j+=addto)
        {
            CardClock t;
            t = j - istamp;
            printf("F at+ %.6f sec, %d on entry, 0 new, 0 deleted, %d remaining, %d c\n",
                t.get_seconds(),
                total_flows - ndeleted_before_clean,
                total_flows - deleted_flows,
                flow_access_timeout.get_seconds()
            );
        }
    }
}



// called each time the housekeep_timeout has expired
//
void flows_expire( CardClock *now )
{
    struct hashrec *hp, *next;
    int ndeleted_this_time;
    Bit16 new_bytes_per_packet;
    u_int new_packets;
    Bit64 new_bytes;
    u_int new_flows;
    double new_houseclean_interval;
    double local_max_houseclean_interval = DBL_MIN;
    unsigned interface;
    unsigned bucket_count;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        new_houseclean_interval = CardClock( tstamp[interface] - last_report_stamp[interface] ).get_seconds();

        // longest interval is best estimate for non-per-card flows per second
        // (see assignment to new_flows after this loop, for more comments)
        //
        if( local_max_houseclean_interval < new_houseclean_interval )
            local_max_houseclean_interval = new_houseclean_interval;

    #if DELTA_TIME_PARANOIA
        if( new_houseclean_interval > 10.0 )
        {
            printf( "new_houseclean_interval %f is way too big!\n", new_houseclean_interval );
            printf( "iterated interface=0x%.4X\n", interface );
            printf( "CardClock next high=%u low=%u\n", next_report_stamp.high, next_report_stamp.low );
            printf( "CardClock last high=%u low=%u\n", last_report_stamp[interface].high, last_report_stamp[interface].low );

    #if DELTA_TIME_FIXUP
            new_houseclean_interval = housekeep_timeout.get_seconds();
    #endif
        }
        else
    #endif
        {
    #if DELTA_TIME_STATS
            if( max_houseclean_interval < new_houseclean_interval )
                max_houseclean_interval = new_houseclean_interval;

            if( min_houseclean_interval > new_houseclean_interval )
                min_houseclean_interval = new_houseclean_interval;

            sum_houseclean_interval += new_houseclean_interval;
    #endif
        }

        // if any data has been received by this interface...
        //
        // otherwise, we want to avoid dividing by zero
        // and no point in adding zero to accumulators
        // and no point in comparing zero with maxes
        //
        if( new_houseclean_interval != 0.0 )
        {
            new_packets = (all_packets[ interface ] - cleaned_packets[ interface ]) / new_houseclean_interval;
            new_bytes = double(all_bytes[ interface ] - cleaned_bytes[ interface ]) / new_houseclean_interval;

            if( max_bytes[ interface ] < new_bytes )
            {
#if DEBUG_MAX_BPS
                fprintf( stderr, "%d max bytes per second increased from %.f to %.f\n", interface, (double)max_bytes[interface], (double)new_bytes );
                fprintf( stderr, "%d total=%.f cleaned=%.f interval=%g\n", interface, (double)all_bytes[interface], (double)cleaned_bytes[interface], new_houseclean_interval );
#endif

#if MAX_BPS_PARANOIA
                // if too many bytes per second for this interface...
                //
                if( (double)new_bytes > EXPECTED_BYTES )
                {
                    fprintf( stderr, "interface %d DANGER!  %.f is too many bytes per second!\n", interface, (double)new_bytes );
                    fprintf( stderr, "all_bytes[if]=%.0f cleaned_bytes[if]=%.0f new_houseclean_interval=%.1f\n",
                        double(all_bytes[interface]), double(cleaned_bytes[interface]), new_houseclean_interval );

#if 0
                    // truncate, in an attempt to fix max BPS
                    //
                    new_bytes = EXPECTED_BYTES;
#endif

                    // remember this error occurred
                    //
                    bitrate_overrun++;
                }
#endif
                max_bytes[ interface ] = new_bytes;
            }

            if( max_packets[ interface ] < new_packets )
                max_packets[ interface ] = new_packets;

            if( new_packets )
                new_bytes_per_packet = double( new_bytes ) / new_packets;
            else
                new_bytes_per_packet = Bit32( 0 );

            if( max_bytes_per_packet[ interface ] < new_bytes_per_packet )
                max_bytes_per_packet[ interface ] = new_bytes_per_packet;

            sum_bytes[ interface ] += new_bytes;
            sum_packets[ interface ] += new_packets;
            sum_bytes_per_packet[ interface ] += Bit32( new_bytes_per_packet );
        }
    }

    // this is better than assigning to new_flows when current_interface
    // is seen in loop above, because we want to use the best value for
    // interval when query is forced (current_interface >= MAX_ATM_NICS)
    // and flows were still seen since last expiration pass
    //
    // also, leaving the stack variable uninitialized tended to end up
    // with the value 4447288, which caused sum_new_flows to be way too
    // big, which caused variances in the ST line to be negative!
    //
    if( local_max_houseclean_interval > 0.0 )
        new_flows = (total_flows - cleaned_flows) / local_max_houseclean_interval;
    else
        new_flows = 0;

    if( max_new_flows < new_flows )
        max_new_flows = new_flows;

    sum_new_flows += new_flows;

    // small hack: flows_delete() will modify deleted_flows
    //
    ndeleted_before_clean = deleted_flows;

    // for all flows, least recently accessed first...
    //
    for( hp=Q.head; hp; hp=next )
    {
        // if this flow saw packets recently...
        //
        if( *now <= hp->hr_access + flow_access_timeout )

            // early out, because we have already seen any
            // less recently accessed flows, since they were
            // sorted to front of this list
            //
            break;

        // time to expire a flow
        //
        flows_summarize_one( "", &prefs, hp );

        // remember where the next flow is, since we are
        // about to alter this record's knowledge of it
        //
        next = hp->hr_qlink.next;

        // take this flow off the access time list
        //
        delque( hp );

        // take this flow off its collision-resolution list
        // and free up its memory
        //
        flows_delete( hp );
    }

    // for all hash chains...
    //
    for( bucket_count=0; bucket_count<NFLOWBUCKETS; bucket_count++ )
    {
        // for all hash records in this chain, least recently created first...
        //
        for( hp=flowtab[bucket_count]; hp; hp=next )
        {
            // if this flow was started recently enough...
            //
            if( *now <= hp->hr_start + flow_create_timeout )

                // early out, because we have already seen any
                // less recently created flows, since they were
                // sorted to front of this list
                //
                break;

            // time to expire a flow
            //
            flows_summarize_one( "", &prefs, hp );

            // take this flow off the access time list
            //
            delque( hp );

            // remember where the next flow is, since we are
            // about to alter this record's knowledge of it
            //
            next = hp->hr_clink.next;

            // take this flow off its collision-resolution list
            // and free up its memory
            //
            flows_delete( hp );
        }
    }

    ndeleted_this_time = deleted_flows - ndeleted_before_clean;
#if EXPIRE_OVERHEAD_STATS
    sum_deleted_flows += ndeleted_this_time;
    if( max_deleted_flows < ndeleted_this_time )
        max_deleted_flows = ndeleted_this_time;
#endif

    // pending flows stats
    //
    // note that we calculate max flows as flows are created,
    // not here, since it isn't a per-second measure and we want
    // it to reflect the true high-water mark, not a sampled one
    //

    sum_flows += total_flows - deleted_flows;

    if( fillin_wanted )
    {
        CardClock tmp = *now - istamp;
        flows_fillin_seconds(&last_report_stamp[interface], now);

        printf("F at %.6f sec, %d on entry, %d new, %d deleted, %d remaining, %.6f tmo, %.6f\n",
            tmp.get_seconds(),
            total_flows - ndeleted_before_clean,
            total_flows - cleaned_flows,
            ndeleted_this_time,
            total_flows - deleted_flows,
            flow_access_timeout.get_seconds(),
            now->get_seconds()
        );
    }

    // we can't simply add the housekeeping timeout
    // to the previous report timestamp, because if we
    // are very far behind, it would still end up less
    // than current time, causing this function to be
    // called as soon as the next packet is seen
    //
    // that would be wasteful of CPU because we would iterate
    // through all flows with nearly no time elapsed and hence
    // nearly no flows to expire
    //
    // more importantly, if the link is busy on more than one
    // virtual circuit, the second packet could be right in
    // the middle of the first one, causing us to calculate
    // too small a value for new_houseclean_interval (remember
    // that the timestamp the card attached is for the first
    // cell only), which would inflate all our per-second byte
    // and packet counts, possibly way beyond what is possible
    // for the link!
    //
    next_report_stamp = *now + housekeep_timeout;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        last_report_stamp[ interface ] = tstamp[ interface ];
        cleaned_packets[ interface ] = all_packets[ interface ];
        cleaned_bytes[ interface ] = all_bytes[ interface ];
    }
    cleaned_flows = total_flows;
}



void flows_usage( char *prog )
{
    fprintf(stderr, "Usage: %s [-pkh] [-o object]\n", prog );
    fprintf(stderr, "\t[-s create-timeout] [-t access-timeout]\n");
    fprintf(stderr,"\t[-a multicast-address-to-spew-active-flows-to]\n");
    fprintf(stderr,"\t[-u ip-address-to-tunnel-multicast-through]\n");
    fprintf(stderr,"\t[-l time-to-live-for-multicast-packets]\n");
    fprintf(stderr,"\t[-1 srcaddr] [-2 dstaddr] [-3 prot] [-4 sport] [-5 dport]\n");
    fprintf(stderr,"\t(p for printing, k for suppressing per second status \n");
    fprintf(stderr,"\th for help)\n");
}



int flows_init(
    int argc,
    char *argv[],
    PentiumClock *keystroke_interval,
    PentiumClock *query_interval,
    PentiumClock *active_interval,
    int *want_facility_loopback,
    int *want_rx,
    int *want_tx
)
{
    u_short rsp, rdp, rpr;
    struct in_addr rsa, rda;
    unsigned count;
    int option_flag;
    int result;
    extern char *optarg;
    unsigned interface;
    char tempstr[ 16 ];

    if (argc < 2) {
        flows_usage(argv[0]);
        return ERROR_FLOWS_TOO_FEW_ARGS;
    }

    while ((option_flag = getopt(argc, argv, "pkh:l:u:a:o:s:t:1:2:3:4:5:")) != -1) {
        switch (option_flag) {

        /* klt - 10/30/96 - option "a:<mcast addr>" for active feed */
        case 'a':
            /* multicast feed address */
            maddr.s_addr = resolve( optarg );
            active_feed = TRUE;
            break;
        /* jza - 11/13/96 - option "u:<multicast tunnel addr>" for active feed */
        case 'u':
            /* multicast tunnel address */
            taddr.s_addr = resolve( optarg );
            break;
        /* jza - 11/13/96 - option "l:<multicast time to live>" for active feed */
        case 'l':
            /* multicast time to live */
            mttl = atoi( optarg );
            break;
        case 'o':
            objname = optarg;
            break;
        case 'p':
            pflag = TRUE;
            break;
        case 'k':
            fillin_wanted = FALSE;
            break;
        case 's':
            flow_create_timeout.put_seconds( atof( optarg ) );
            break;
        case 't':
            flow_access_timeout.put_seconds( atof( optarg ) );
            break;
        case '1':
            /* source address */
            sscanf(optarg, "%x", &rsa.s_addr);
            prefs.up_dst = rsa;
            rflag = TRUE;
            break;
        case '2':
            /* dest address */
            sscanf(optarg, "%x", &rda.s_addr);
            prefs.up_src = rda;
            rflag = TRUE;
            break;
        case '3':
            /* protocol number */
            rpr = atoi(optarg);
            break;
        case '4':
            /* source port */
            rsp = atoi(optarg);
            prefs.up_sport = rsp;
            break;
        case '5':
            /* dest port */
            rdp = atoi(optarg);
            prefs.up_dport = rdp;
            break;
        case '?':
        case 'h': /* help */
        default:
            flows_usage(argv[0]);
            return ERROR_FLOWS_UNKNOWN_ARG;
        }
    }

    // need to hash the user prefs so comparisons will work
    //
    prefs.up_index = flowhash( &prefs.up_fd );

    printf( "flows time out when created >%.1f seconds ago or accessed >%.1f seconds ago\n",
        flow_create_timeout.get_seconds(), flow_access_timeout.get_seconds() );
    printf( "flows pflag=%s\n", pflag?"TRUE":"FALSE" );
    printf( "flows fillin %swanted\n", fillin_wanted?"":"not " );
    printf( "flows query output buffer size is %u bytes\n", QUERY_BUFFER_SIZE );

    // sorry, we cannot init TCP/IP library
    // see main() in capture.cpp for reasons
    //

    // force all our TCP/IP sockets to a known state
    //
    memset( &query_sockets, 0, sizeof(query_sockets) );
    memset( &control_socket, 0, sizeof(control_socket) );
    memset( &flowtrace_socket, 0, sizeof(flowtrace_socket) );

    // allocate output buffers for query sockets
    // so we get back to processing raw packets quickly,
    // even when the query connection is very slow
    //
    for( count=0; count<QUERY_SOCKET_COUNT; count++ )
    {
        query_output_buffers[ count ] = (Bit8 *) flat_alloc( QUERY_BUFFER_SIZE );
        if( query_output_buffers[ count ] == NULL )
        {
            printf( "error %d allocating %d bytes for query output buffer %d\n", errno, QUERY_BUFFER_SIZE, count );
            return ERROR_FLOWS_CANNOT_ALLOC_TCP_BUF;
        }
    }

    // start using the first query socket
    //
    which_query_socket_to_use = 0;

    // no query is pending yet
    //
    which_query_socket_prev_used = QUERY_SOCKET_COUNT;
    query_drainage_delta = 0.0;

    /* get a pointer to the string for the environment variable
     * that holds the IP address of the host who can get raw flows
     */
    flowtrace_client_string = getenv( FLOWTRACE_CLIENT_ENVAR );
    if( flowtrace_client_string == NULL )
    {
        printf( "flows_init() error: environment variable '%s' is not set\n",
            FLOWTRACE_CLIENT_ENVAR );
        return ERROR_FLOWS_NO_ENVAR;
    }
    printf( "environment variable %s='%s'", FLOWTRACE_CLIENT_ENVAR, flowtrace_client_string );

    /* convert the string to a 32-bit unsigned integer
     */
    flowtrace_client_address = resolve( flowtrace_client_string );
    if( 0 == flowtrace_client_address )
        printf( ", so any host can get raw expired flows\n" );
    else
        printf( ", so only host %s can get raw expired flows\n",
            inet_ntoa(tempstr,flowtrace_client_address) );

    /* klt - 10/30/96 - active feed */
    if (active_feed)
    {
        longword dport = ACTIVE_FEED_PORT;
        word iport;
        char maddr_string[ 17 ];
        char taddr_string[ 17 ];

        printf("Tunneling active flow updates to %s through %s with TTL %d\n",
            inet_ntoa(maddr_string, maddr.s_addr),
            inet_ntoa(taddr_string, taddr.s_addr),
            mttl);

        /* open a udp socket for sending it out */
        if (!udp_open(&msock,iport,maddr.s_addr,dport,NULL)) {
            printf("ERROR - unable to open udp socket\n");
            exit(1);
        }

#if DEBUG_ACTIVE_FEED
        printf("Opened UDP socket to send active feed to 0x%.8X\n", maddr.s_addr);
#endif
        sock_tunnel( &msock, taddr.s_addr );
        sock_ttl( &msock, mttl );

        // set an "alarm" time when we will
        // query the statistics
        //
        active_interval->put_seconds( ACTIVE_SPEW_MILLISECS / 1000 );
    }
    else
        printf( "active flows not feeding to anyone\n" );

    if (rflag) {
        printf("Tracing %x:%x\t%d:%d:%d\n", rsa.s_addr, rda.s_addr, rpr, rsp, rdp);
    }

    if ((objtype = findobj(objname)) < 0) {
        fprintf(stderr, "invalid option '%s' for -o\n", objname);
        fprintf(stderr, "valid types are: ");
        dumpobjtypes();
        return ERROR_FLOWS_UNKNOWN_OBJECT_TYPE;
    }
    printf( "flows using object name '%s'\n", objname );




    fprintf( stderr, "buffer blocks are %d bytes each\n", sizeof(struct BufferBlock) );
    fprintf( stderr, "cell 1 trace records are %d bytes each\n", sizeof(struct tracerec) );
    fprintf( stderr, "cell 2 trace records are %d bytes each\n", sizeof(struct tracerec_2) );
    fprintf( stderr, "AAL5 reassembly hash records are %d bytes each\n", sizeof(struct cellhashrec) );
    fprintf( stderr, "flow hash records are %d bytes each\n", sizeof(struct hashrec) );
    fprintf( stderr, "active flow stats are %d bytes each, so %d fit in each %d byte UDP payload\n",
        sizeof(struct flowstats), FLOWS_PER_UDP, UDP_PAYLOAD_SIZE );

    /* allocate buckets to point at each hash chain */
    flowtab = (struct hashrec **) flat_alloc( BUCKET_POOL_SIZE );
    if( flowtab == NULL )
    {
        printf( "error %d allocating %d bytes (%d pointers) for flow hash buckets\n", errno, BUCKET_POOL_SIZE, NFLOWBUCKETS );
        return ERROR_FLOWS_CANNOT_ALLOC_FLOW_BUCKETS;
    }

    /* nothing in any of the buckets yet */
    memset( flowtab, 0, BUCKET_POOL_SIZE );

#if DEBUG_CHAIN_LENGTH
    /* allocate chain length counters for each bucket */
    chainlen = (int *) flat_alloc( CHAIN_POOL_SIZE );
    if( chainlen == NULL )
    {
        printf( "error %d allocating %d bytes (%d integers) for chain length counters\n", errno, CHAIN_POOL_SIZE, NFLOWBUCKETS );
        return ERROR_FLOWS_CANNOT_ALLOC_CHAINLENS;
    }

    /* initialize chain length of all hashtable buckets to zero
     */
    for( count=0; count<NFLOWBUCKETS; count++ )
        chainlen[ count ] = 0;
#endif

    /* allocate a pool of hash records */
    hash_pool = (struct hashrec *) flat_alloc( HASH_POOL_SIZE );
    if( hash_pool == NULL )
    {
        printf( "error %d allocating %d bytes (%d records) for flow hash record pool\n", errno, HASH_POOL_SIZE, HASH_RECORDS );
        return ERROR_FLOWS_CANNOT_ALLOC_FLOWS;
    }

    /* capture module should enable RX but not TX on all cards */
    *want_rx = TRUE;
    *want_tx = FALSE;

    /* capture module should enable facility loopback on all cards */
    *want_facility_loopback = TRUE;

    /* capture module should ask me about query every chance it gets */
    query_interval->put_seconds( 0 );

    /* capture module only needs to check for keystrokes once a second */
    keystroke_interval->put_seconds( 1.0 );

    /* put all the hash records on the singly-linked free list */
    hash_free = hash_pool;
    for( count=0; count<HASH_RECORDS-1; count++ )
    {
        hash_pool[ count ].hr_clink.next = hash_pool + count + 1;
    }
    hash_pool[count].hr_clink.next = NULL;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        last_report_stamp[ interface ] = 0;
        last_already_set[ interface ] = FALSE;
    }
    report_already_set = FALSE;
    cleaned_flows = 0;
    deleted_flows = 0;
    total_flows = 0;
    query_count = -1;

    result = cells_init();
    if( result )
        return result;

    return flows_clear();
}



// clears all statistics after being queried remotely
// or during initialization of module
//
// note that we do not touch the report timestamp
// so that querying will not delay timing out a flow
//
int flows_clear( void )
{
    int result;
    unsigned interface;
#if COUNT_NON_IP_PER_VPI
    unsigned vpi;
#endif

    /* initialize 2nd-level stats
     */
    result = init_stats();
    if( result )
        return result;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        // updated when one or more flows expire
        //
        sum_bytes[ interface ] = Bit32( 0 );
        max_bytes[ interface ] = Bit32( 0 );
        sum_packets[ interface ] = 0;
        max_packets[ interface ] = 0;
        sum_bytes_per_packet[ interface ] = Bit32( 0 );
        max_bytes_per_packet[ interface ] = 0;
        cleaned_bytes[ interface ] = Bit32( 0 );
        cleaned_packets[ interface ] = 0;

        // updated when a packet comes in
        //
        all_bytes[ interface ] = Bit32( 0 );
        instant_max_bytes[ interface ] = 0;
        all_packets[ interface ] = 0;
        non_ip_packets[ interface ] = 0;
        ip_optioned_packets[ interface ] = 0;
        tcp_optioned_packets[ interface ] = 0;
        non_first_ip_fragments[ interface ] = 0;
        first_ip_fragments[ interface ] = 0;
#if REASM_IP_FRAGS
        orphan_non_first_ip_fragments[ interface ] = 0;
        overlap_non_first_ip_fragments[ interface ] = 0;
        hole_non_first_ip_fragments[ interface ] = 0;
#endif
#if EXTRA_TCP_STATS && CELLS_PER_PACKET>1
        unexpected_tcp_option_lengths[ interface ] = 0;
#endif

#if COUNT_NON_IP_PER_VPI
        for( vpi=0; vpi<=MAX_VPI; vpi++ )
            non_ip_packets_per_vpi[ interface ][ vpi ] = 0;
#endif
    }
    max_flows = 0;
    sum_flows = 0;
    max_new_flows = 0;
    sum_new_flows = 0;
#if EXPIRE_OVERHEAD_STATS
    max_deleted_flows = 0;
    sum_deleted_flows = 0;
#endif

#if DELTA_TIME_STATS
    sum_houseclean_interval = 0;
    max_houseclean_interval = DBL_MIN;
    min_houseclean_interval = DBL_MAX;
#endif

    hash_alloc_failures = 0;
    istamp_already_set = FALSE;
    query_count++;

    memset( ip_length_usage, 0, sizeof(ip_length_usage) );
    memset( ip_precedence_usage, 0, sizeof(ip_precedence_usage) );
    memset( pktflow, 0, sizeof(pktflow) );

#if SHOW_CPU_UTILIZATION
    sum_elapsed = Bit32( 0 );
    max_elapsed = Bit32( 0 );
    min_elapsed.high = -1;
    min_elapsed.low = -1;
    blocks = 0;
    startup.set();
#endif

#if BUFFER_USED_PARANOIA
    blocks_overrun = 0;
#endif

#if MAX_BPS_PARANOIA
    bitrate_overrun = 0;
#endif

    return 0;
}



int flows_per_packet( cellhashrec *chp, Bit8 interface )
{
    u_short just_offset;        // without flags
#if REASM_IP_FRAGS
    hashrec *fragment;
    u_short new_len;
#endif

    all_packets[ interface ]++;

    /* shortcut variable
     */
    tstamp[ interface ] = chp->tr_when;
    tstamp[ interface ].strip_bitfields();

    /* if we haven't seen any data since last query...
     */
    if( !istamp_already_set )
    {
        istamp_already_set = TRUE;

        // remember when the first packet was seen
        //
        istamp = tstamp[ interface ];
    }

    // if we never set the report timestamp...
    //
    if( !report_already_set )
    {
        report_already_set = TRUE;

        // do housekeeping one timeout from now
        //
        next_report_stamp = tstamp[ interface ] + housekeep_timeout;
    }

    // if we haven't seen any data on this interface yet...
    //
    if( !last_already_set[ interface ] )
    {
        last_already_set[ interface ] = TRUE;

        last_report_stamp[ interface ] = tstamp[ interface ];
    }

#if !STUB_FLOWS_EXPIRE
    /* report results if the interval is correct
     * (each 'housekeep_timeout' time quanta)
     *
     * no point in doing this unless the timestamp
     * could have changed, outside the
     */
    if( tstamp[ interface ] >= next_report_stamp )
    {
        flows_expire( &tstamp[interface] );
    }
#endif

#if !ASSUME_ETHERTYPE_IS_IP && (LINK_HAS_LLC_SNAP || LINK_HAS_NLPID)
    /* if MAC protocol is IP...
#if LINK_HAS_LLC_SNAP
     *
     * assumes that if LLC or SNAP OUI are wrong,
     * it is likely that ethertype is also wrong
#endif
     */
    if( chp->tr_etype != ETHERTYPE_IP_SWAPPED )
    {
        non_ip_packets[ interface ]++;
#if COUNT_NON_IP_PER_VPI
        non_ip_packets_per_vpi[ interface ][ ENTIRE_VPI(chp->tr_atm.bits) ]++;
#endif
#if DEBUG_NON_IP
#if DEBUG_NON_IP_PER_VPI && COUNT_NON_IP_PER_VPI
        // if this is the first time we saw non-ip on this VP...
        //
        if( 1 == non_ip_packets_per_vpi[ interface ][ ENTIRE_VPI(chp->tr_atm.bits) ] )
#endif
        {
            // avoid infinite loop inside TCP stack!
            //
            if( sock_established( &control_socket ) )
            {
                sock_printf( &control_socket, "non-IP packet seen!\n" );
                dump_packet( &control_socket, chp );
            }
        }
#endif
        return 0;
    }
#endif

    chp->tr_len = ntohs( chp->tr_len );
    all_bytes[ interface ] += Bit32( chp->tr_len );
    if( instant_max_bytes[ interface ] < chp->tr_len )
        instant_max_bytes[ interface ] = chp->tr_len;
    ip_length_usage[ chp->tr_len >> IP_LEN_SHIFT ]++;

    ip_precedence_usage[ (chp->tr_tos >> IP_PREC_SHIFT) & IP_PREC_MASK ]++;

    chp->tr_offset = ntohs( chp->tr_offset );

    /* if object type (statistic to track) is "portquad"
     * then we must toss packets with options and fragments
     * that are not the first
     */
#if !ASSUME_OBJECT_IS_PQ
    if( objindex == PORTQUAD_OBJ_INDEX )
#endif
    {
        if( chp->tr_iphl != (sizeof(ip) >> 2) )
        {
            ip_optioned_packets[ interface ]++;
            return 0;
        }

        just_offset = (chp->tr_offset & IP_OFF_MASK) << 3;
        if( just_offset != 0 )
        {
            non_first_ip_fragments[ interface ]++;

#if REASM_IP_FRAGS
            // we keep state in the flow and have no way to create a flow
            // using only a fragment other than the first, so we will sadly
            // (1) discard all fragments as orphans until the zero-offset
            // fragment is received, and
            // (2) leave the flow on the fragment list if the highest-
            // offset fragment is not received last
            //
            // one way to fix this is to (1) keep state somewhere
            // outside the flow, as AAL5.C does for ATM cell reassembly, but
            // that is a major undertaking that I am unwilling to do just to
            // track an attack that is already fixed in the internet...
            //
            // even if we allowed fake flows to be created (differentiable
            // by not being on other lists and/or left edge of data seen
            // being greater than zero) when only non-first have been seen,
            // we would still need to time them out so memory gets recovered
            // when one fragment is permanently dropped
            //
            // how to maintain state without variable record length?
            // keep left, right, hole left, hole right variables
            // like we do for TCP hole/retransmit tracking, but again
            // that's code (human time upfront, CPU time per fragment)
            // and storage (3 more Bit16's per flow) that may not be
            // worth it now, or ever
            //

            // for all flows with outstanding IP fragments,
            // most recently seen first...
            //
            for( fragment=F.tail; fragment; fragment=fragment->hr_flink.prev )
            {
                // if this flow has the same interface, IP address pair and IP ID...
                //
                if( fragment->hr_iface == interface
                &&  fragment->hr_ident == chp->tr_ident
                &&  fragment->hr_dst == chp->tr_dst
                &&  fragment->hr_src == chp->tr_src )
                {
                    break;
                }
            }

            if( fragment )
            {
                // if the beginning of this packet overlaps the end of
                // fragments seen so far for this packet, this is a
                // possible "bonk" attack
                //
                if( fragment->hr_len > just_offset )
                {
                    overlap_non_first_ip_fragments[ interface ]++;
#if DEBUG_OVERLAP_FRAGS
                    printf( "IP frag %d@%d overlaps prev\n", chp->tr_len-sizeof(ip), just_offset );
                    dump_flow( stdout, fragment );
#endif
                    saturated_inc_bit8( fragment->hr_olaps );

                    // a "perfect" reassembly engine would deal with
                    // the overlapping by keeping the uniquely new bytes,
                    // so figure how much that is
                    //
                }
                else if( fragment->hr_len < just_offset )
                {
                    hole_non_first_ip_fragments[ interface ]++;
#if DEBUG_HOLE_FRAGS
                    printf( "IP frag %d@%d leaves hole after prev\n", chp->tr_len-sizeof(ip), just_offset );
                    dump_flow( stdout, fragment );
#endif
                    saturated_inc_bit8( fragment->hr_holes );

                    // now act as if we got all fragments before this
                    // to avoid triggering on the next fragment that
                    // fits properly on the right of this one
                    //
                    // of course, that assumes that packets were sent
                    // from low offsets to high offsets, and they haven't
                    // been reordered in the network
                    //
                }

                // calculate the right edge of this fragment
                //
                new_len = just_offset + chp->tr_len - sizeof( ip );

                // saturate if it wrapped
                //
                if( new_len < just_offset )
                    new_len = 65535;

                // don't change packet right edge if this isn't
                // to the right of previous data
                //
                if( new_len > fragment->hr_len )
                {
                    fragment->hr_bytes += new_len - fragment->hr_len;
                    fragment->hr_len = new_len;
                }

                // any fragment counts as an access to the flow
                //
                fragment->hr_access = tstamp[ interface ];

                // if there are no more fragments...
                //
                if( chp->tr_offset & IP_MF == 0 )

                    // take this flow off the fragments list
                    //
                    delete_fragment_queue( fragment );
            }
            else
            {
#if DEBUG_ORPHAN_FRAGS
                char scratch1[ 20 ];
                char scratch2[ 20 ];

                printf( "interface %d cannot find src/dst/ident %s/%s/%d for frag %d@%d\n",
                    interface,
#if BYTESWAP_FIELDS_EARLY
                    inet_ntoa( scratch1, chp->tr_src ),
                    inet_ntoa( scratch2, chp->tr_dst ),
#else
                    inet_ntoa( scratch1, ntohl( chp->tr_src ) ),
                    inet_ntoa( scratch2, ntohl( chp->tr_dst ) ),
#endif
                    ntohs( chp->tr_ident ),
                    chp->tr_len - sizeof( ip ),
                    just_offset
                );
#endif
                orphan_non_first_ip_fragments[ interface ]++;
            }
#endif
            return 0;
        }

        if( chp->tr_offset & IP_MF )
            first_ip_fragments[ interface ]++;
    }

#if BYTESWAP_FIELDS_EARLY
    chp->tr_src=ntohl(chp->tr_src);
    chp->tr_dst=ntohl(chp->tr_dst);
    chp->tr_sport=ntohs(chp->tr_sport);
    chp->tr_dport=ntohs(chp->tr_dport);
#endif

#if STUB_DO_RECORD
    return 0;
#else
    /* insert new chp */
    return do_record( chp, interface );
#endif
}


/* flows_per_block() moved to blk2pkt.c, 19 Aug 98 */

// called in foreground by main data capture program
//
// int flows_per_block( BufferBlock *block )
//


// called to determine if program has been queried remotely
//
int flows_is_query_wanted( void )
{
    int result;
    char scratch[ 20 ];

    // give the TCP stack some time, and have it check the flowtrace socket
    //
    // if it is closed...
    //
    if( !tcp_tick( &flowtrace_socket ) )
    {
        // log that connection was closed, if we aren't just initializing
        //
        if( flowtrace_is_connected )
            sock_printf( stdout, "flowtrace disconnected from IP address %s, port %d\n",
                inet_ntoa(scratch,flowtrace_caller.s_ip), flowtrace_caller.s_port );

        // we don't know what flows should be traced yet
        //
        flowtrace_parms_are_filled = FALSE;

        // we haven't gotten a connection yet
        //
        flowtrace_is_connected = FALSE;

        // listen for a TCP/IP connection from any port on desired host,
        // without timing out, with no data handling callback
        // function and no busy-loop callback function
        //
        // no point in checking result; always succeeds
        //
        tcp_open( &flowtrace_socket, FLOWTRACE_TCP_PORT );
        tcp_listen( &flowtrace_socket, flowtrace_client_address, 0, NULL, 0 );
    }

    // otherwise, if someone has connected to us
    //
    else if( sock_established( &flowtrace_socket ) )
    {
        if( !flowtrace_is_connected )
        {
            int len = sizeof( sockaddr );
            int result;

            result = getpeername( &flowtrace_socket, &flowtrace_caller, &len );
            if( result < 0 )
                sock_printf( stdout, "cannot get peer name for flowtrace connection\n" );
            else if( len != sizeof( sockaddr ) )
                sock_printf( stdout, "got wrong peer name length for flowtrace connection\n" );
            else
                sock_printf( stdout, "flowtrace connected from IP address %s, port %d\n",
                    inet_ntoa(scratch,flowtrace_caller.s_ip), flowtrace_caller.s_port );

            // remember that we already logged the connection
            //
            flowtrace_is_connected = TRUE;
        }

#if !RESELECTABLE_FLOWTRACE
        if( !flowtrace_parms_are_filled )
#endif

            // if enough data was received for our flowtrace parms...
            //
            if( sock_dataready( &flowtrace_socket ) >= sizeof( flowselector ) )
            {
                // overwrite the existing parameters
                //
                sock_read( &flowtrace_socket, (Bit8 *)&flowtrace_parms, sizeof(flowselector) );

                // byteswap the ignore and ATM fields first
                //
                flowtrace_parms.behavior = ntohs( flowtrace_parms.behavior );
    #if KEEP_ATM_HEADER
                flowtrace_parms.atm.whole = ntohl( flowtrace_parms.atm.whole );
    #endif

                // log what parms they gave
                //
                // we do this before the conditional swapping,
                // since unconditional swapping here is easier to read
                //
                sock_printf( stdout, "flowtrace selection: " );

                if( flowtrace_parms.behavior & fsel_ignore_iface )
                    sock_printf( stdout, "ignore iface, " );
                else
                    sock_printf( stdout, "iface=%d, ", flowtrace_parms.iface );

                if( flowtrace_parms.behavior & fsel_ignore_proto )
                    sock_printf( stdout, "ignore proto, " );
                else
                    sock_printf( stdout, "proto=%d, ", flowtrace_parms.proto );

                if( flowtrace_parms.behavior & fsel_ignore_sport )
                    sock_printf( stdout, "ignore sport, " );
                else
                    sock_printf( stdout, "sport=%d, ", ntohs(flowtrace_parms.sport) );

                if( flowtrace_parms.behavior & fsel_ignore_dport )
                    sock_printf( stdout, "ignore dport, " );
                else
                    sock_printf( stdout, "dport=%d, ", ntohs(flowtrace_parms.dport) );

                if( flowtrace_parms.behavior & fsel_ignore_src )
                    sock_printf( stdout, "ignore src, " );
                else
                    sock_printf( stdout, "src=%s, ", inet_ntoa(scratch,ntohl(flowtrace_parms.src)) );

                if( flowtrace_parms.behavior & fsel_ignore_dst )
                    sock_printf( stdout, "ignore dst, " );
                else
                    sock_printf( stdout, "dst=%s, ", inet_ntoa(scratch,ntohl(flowtrace_parms.dst)) );

                if( flowtrace_parms.behavior & fsel_ignore_gfc )
                    sock_printf( stdout, "ignore GFC, " );
                else
                    sock_printf( stdout, "GFC=%d, ", flowtrace_parms.atm.bits.GFC );

                if( flowtrace_parms.behavior & fsel_ignore_vpi )
                    sock_printf( stdout, "ignore VPI, " );
                else
                    sock_printf( stdout, "VPI=%d, ", flowtrace_parms.atm.bits.VPI );

                if( flowtrace_parms.behavior & fsel_ignore_vci )
                    sock_printf( stdout, "ignore VCI\n" );
                else
                    sock_printf( stdout, "VCI=%d\n", flowtrace_parms.atm.bits.VCI );

                sock_printf( stdout, "flowtrace output will be in %s\n",
                    (flowtrace_parms.behavior & fsel_output_text) ? "text" : "binary" );

                // byteswap other parameters that will be in host byte order
                // when the flow expires, and which are bigger than 1 byte
                //
    #if BYTESWAP_FIELDS_EARLY
                flowtrace_parms.sport = ntohs( flowtrace_parms.sport );
                flowtrace_parms.dport = ntohs( flowtrace_parms.dport );
                flowtrace_parms.src   = ntohl( flowtrace_parms.src );
                flowtrace_parms.dst   = ntohl( flowtrace_parms.dst );
    #endif

                // remember that they gave us parms
                //
                flowtrace_parms_are_filled = TRUE;
            }
    }



    // give the TCP stack some time, and have it check the control socket
    //
    // if it is closed...
    //
    if( !tcp_tick( &control_socket ) )
    {
        // listen for a TCP/IP connection from any host/port,
        // without timing out, with no data handling callback
        // function and no busy-loop callback function
        //
        // no point in checking result; always succeeds
        //
        tcp_open( &control_socket, CONTROL_TCP_PORT );
        tcp_listen( &control_socket, 0, 0, NULL, 0 );
    }




    // now the code for the query sockets
    //

    // give the TCP/IP library some time
    //
    tcp_Socket *current_query_socket = &query_sockets[ which_query_socket_to_use ];
    Bit8 *current_query_buffer = query_output_buffers[ which_query_socket_to_use ];
    result = tcp_tick( current_query_socket );

    // if a query is draining and that previous socket has been closed...
    //
    if( which_query_socket_prev_used < QUERY_SOCKET_COUNT
    &&  !_chk_socket( query_sockets + which_query_socket_prev_used ) )
    {
        // remember that no query is draining anymore
        //
        which_query_socket_prev_used = QUERY_SOCKET_COUNT;

        // calculate how long it took to push the stats
        // through the damned sluggish network
        //
        PentiumClock now;
        now.set();
        query_drainage_delta = PentiumClock( now - query_start_time ).get_seconds();

        // put it on the console for debugging
        //
        printf( "previous query socket took %.3g seconds to drain through network\n", query_drainage_delta );
    }

    // if current query socket has been closed...
    //
    if( !result )
    {
        // listen for a TCP/IP connection from any host/port,
        // without timing out, with no data handling callback
        // function and no busy-loop callback function
        //
        // no point in checking result; always succeeds
        //
        tcp_open( current_query_socket, QUERY_TCP_PORT );

        // attach a much-larger output buffer to the socket
        //
        if( 0 == sock_set_tx_buf( current_query_socket, current_query_buffer, QUERY_BUFFER_SIZE ) )
            printf( "error: could not set TX buf for socket #%d\n", which_query_socket_to_use );

        tcp_listen( current_query_socket, 0, 0, NULL, 0 );
        return FALSE;
    }

    // otherwise if someone has connected to us
    //
    if( sock_established( current_query_socket ) )
    {
        // treat this as a query
        //
        // I know this is crude, but I don't want to
        // implement a whole request/response protocol yet
        //
        return TRUE;
    }

    return FALSE;
}



int flows_is_query_pending( void )
{
    return which_query_socket_prev_used != QUERY_SOCKET_COUNT;
}



// this can only work with the WATTCP library for DOS,
// since it lets us allocate the TX buffers,
// but UNIX allocates such things itself in the kernel
//
// this is useful from TRACEFLO.CPP which doesn't want
// to sit around waiting for someone to poll it, and
// wants to keep the results locally on the PC doing
// the analysis of a tracefile
//
void flows_print_query_results( void )
{
    Bit8 *current_query_buffer = query_output_buffers[ which_query_socket_to_use ];

    fprintf( stdout, "%s", current_query_buffer );
}



// called when program is queried remotely
//
int flows_query( const char *extra_msg )
{
    double trace_duration_ind;
    double trace_duration;
    int result;
    unsigned interface;
    unsigned count;
    tcp_Socket *sock = &query_sockets[ which_query_socket_to_use ];
#if REPORT_OVERHEAD_STATS
    PentiumClock start, now;
    double elapsed2, elapsed3;
#endif

#if SHOW_CPU_UTILIZATION
    // finalize flow processing time statistics
    // and print them
    //
    elapsed.set();
    elapsed -= startup;
#endif

    // remember when the query started, so we
    // can calculate network throughput later
    //
    query_start_time.set();

#if QUERY_CALLS_EXPIRE
#if DELTA_TIME_PARANOIA
    current_card = -1;
#endif

    // do one more housecleaning to trigger a timeout
    // on any "pregnant" flows, in case we haven't seen
    // a packet in more than housekeep_timeout ticks
    //
    // this fake timestamp will not cause a negative
    // delta time to be calculated during next flows_expire(),
    // even if it is later than the next packet we will see,
    // because flows_clear() will cause last_report_stamp
    // to be recalculated when we see the next packet
    //
    CardClock now2 = query_start_time;
    flows_expire( &now2 );

#if DELTA_TIME_PARANOIA
    last_card = -1;
    prev_card = -1;
    prev_stamp = now2;
#endif
#endif      // QUERY_CALLS_EXPIRE

    fprintf( stderr, "\n" );
    trace_duration = 0.0;
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        trace_duration_ind = CardClock( tstamp[interface] - istamp ).get_seconds();
        if( trace_duration < trace_duration_ind )
            trace_duration = trace_duration_ind;

#if DEBUG_QUERY_TOTALS
        fprintf( stderr, "interface %u %u/%u/%u/%u/%u"
#if REASM_IP_FRAGS
            "/%u/%u/%u"
#endif
            " all/non-IP/IP-opt/frag-1/frag-non-1"
#if REASM_IP_FRAGS
            "/orphan-frag-non-1/overlap-frag-non-1/hole-frag-non-1"
#endif
            " packets\n",
            interface,
            all_packets[interface],
            non_ip_packets[interface],
            ip_optioned_packets[interface],
            first_ip_fragments[interface],
            non_first_ip_fragments[interface]
#if REASM_IP_FRAGS
            ,
            orphan_non_first_ip_fragments[interface],
            overlap_non_first_ip_fragments[interface],
            hole_non_first_ip_fragments[interface]
#endif
        );

        fprintf( stderr, "interface %u IP bytes: %.f\n", interface, (double)all_bytes[interface] );
#endif

#if FAIL_ON_PACKET_COUNT
        if( all_packets[interface] < EXPECTED_PACKETS )
        {
            printf( "fatal error: expected at least %d packets on interface %u!\n", EXPECTED_PACKETS, interface );
#if WAIT_FOR_KEYSTROKE
            printf( "press a key and I will exit\n" );
            getch();
#endif
            return 1;
        }
#endif
    }





    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u IP bytes: %.0f\n", interface, double(all_bytes[interface]) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u all packets: %u\n", interface, all_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u non IP packets: %u\n", interface, non_ip_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u IP optioned packets: %u\n", interface, ip_optioned_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u TCP optioned packets: %u\n", interface, tcp_optioned_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u first IP fragments: %u\n", interface, first_ip_fragments[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u non-first IP fragments: %u\n", interface, non_first_ip_fragments[interface] );

#if REASM_IP_FRAGS
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u orphaned non-first IP fragments: %u\n", interface, orphan_non_first_ip_fragments[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u overlapping non-first IP fragments: %u\n", interface, overlap_non_first_ip_fragments[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u hole-producing non-first IP fragments: %u\n", interface, hole_non_first_ip_fragments[interface] );
#endif

    sock_printf( sock, "IP fragment allocation failures: %u\n", fraghash_alloc_failures );
    sock_printf( sock, "flow allocation failures: %u\n", hash_alloc_failures );
    sock_printf( sock, "port flow allocation failures: %u\n", porthash_alloc_failures );
    sock_printf( sock, "net flow allocation failures: %u\n", nethash_alloc_failures );
    sock_printf( sock, "host flow allocation failures: %u\n", hosthash_alloc_failures );

    sock_printf( sock, "following statistics were calculated at flow expiration time:"
        " (first packet >1 second after prev flow expiration)\n" );

    sock_printf( sock, "pending flows: %d max, %.0f avg\n",
        max_flows, sum_flows/nonzero(trace_duration) );

    sock_printf( sock, "new flows per second: %d max, %.0f avg\n",
        max_new_flows, sum_new_flows/nonzero(trace_duration) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u packets per second: %d max, %.0f avg\n",
            interface, max_packets[interface], sum_packets[interface]/nonzero(trace_duration) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u bits per second: %.0f max, %.0f avg\n",
            interface, 8*double(max_bytes[interface]), 8*double(sum_bytes[interface])/nonzero(trace_duration) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u avg bytes per packet: %d max, %.0f avg\n",
            interface, max_bytes_per_packet[interface], double(sum_bytes_per_packet[interface])/nonzero(trace_duration) );

    // print out the counts for all IP types of service we have seen
    //
    sock_printf( sock, "IP_PREC" );
    for( count=0; count<IP_PREC_COUNT; count++ )
        if( ip_precedence_usage[count] > 0 )
            sock_printf( sock, " %d:%d", count, ip_precedence_usage[count] );
    sock_printf( sock, "\n" );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u largest packet: %u bytes\n", interface, instant_max_bytes[interface] );

    // print out the counts for all IP lengths we have seen
    //
    sock_printf( sock, "IP_LEN range=%d", 1<<IP_LEN_SHIFT );
    for( count=0; count<IP_LEN_COUNT; count++ )
        if( ip_length_usage[count] > 0 )
            sock_printf( sock, " %d:%d", count<<IP_LEN_SHIFT, ip_length_usage[count] );
    sock_printf( sock, "\n" );

    // klt - print out the counts of flows for each pkt count
    //
    sock_printf(sock,"PKT_FLOW");
    for (count=1; count<PKTFLOW_ARRAY_SIZE_MAX;count++)
        if( pktflow[count] > 0)
            sock_printf(sock, " %d:%d", count, pktflow[count]);
    sock_printf(sock, " %d:%d", count, pktflow[0]);
    sock_printf(sock,"\n");

#if REPORT_OVERHEAD_STATS
    start.set();
#endif

    // output the stats that got calculated for expired flows
    //
    result = report_stats( sock, istamp.get_seconds(), trace_duration );

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed2 = PentiumClock( start - query_start_time ).get_seconds();
    elapsed3 = PentiumClock( now - query_start_time ).get_seconds();
    fprintf( stderr, "cumulative %.4f S before, %.4f after report_stats()\n", elapsed2, elapsed3 );
    sock_printf( sock, "cumulative %.4f S before, %.4f after report_stats()\n", elapsed2, elapsed3 );
#endif

    if( result )
        return result;

#if COUNT_NON_IP_PER_VPI
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        for( count=0; count<=MAX_VPI; count++ )
            if( non_ip_packets_per_vpi[ interface ][ count ] )
                sock_printf( sock, "interface %u VPI %u non-IP packets: %u\n", interface, count, non_ip_packets_per_vpi[ interface ][ count ] );
#endif

#if DELTA_TIME_STATS
    sock_printf( sock, "seconds since last flow housecleaning: %f min, %f max, %f avg\n",
        min_houseclean_interval, max_houseclean_interval,
        sum_houseclean_interval/nonzero(trace_duration) );
#endif

#if EXPIRE_OVERHEAD_STATS
    sock_printf( sock, "deleted flows: %d max, %.1f avg\n",
        max_deleted_flows, sum_deleted_flows/nonzero(trace_duration) );
#endif

#if SHOW_CPU_UTILIZATION
    if( blocks )
    {
        sock_printf( sock, "flows_per_block() was called %u times\n", blocks );
        sock_printf( sock, "flows_per_block() took min %.4f max %.4f avg %.4f mS\n",
            min_elapsed.get_seconds() * 1000,
            max_elapsed.get_seconds() * 1000,
            sum_elapsed.get_seconds() * 1000 / blocks
        );
        sock_printf( sock, "%.2f seconds (%.2f%% of %.2f) spent in flows_per_block() during run\n",
            sum_elapsed.get_seconds(),
            100.0*(double)sum_elapsed/(double)elapsed,
            elapsed.get_seconds()
        );
    }
    else
        sock_printf( sock, "flows_per_block() was never called\n" );
#endif

#if BUFFER_USED_PARANOIA
    sock_printf( sock, "flows_per_block() saw improper used>%u %u times\n", COMPL_RING_ENTRIES, blocks_overrun );
#endif

#if MAX_BPS_PARANOIA
    sock_printf( sock, "flows_expire() saw improper bytes/sec>%.0f %u times\n", EXPECTED_BYTES, bitrate_overrun );
#endif

#if EXTRA_TCP_STATS && CELLS_PER_PACKET>1
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u parse_tcp_options() saw improper length %u times\n",
            interface, unexpected_tcp_option_lengths[ interface ] );
#endif

    // packet reassembly routines should get first shot,
    // since they are lower-level than flows, but Kevin
    // wants all new output at bottom of report, after
    // the TCP stats
    //
    // this means that if cells_query() calls cells_expire()
    // the expired packets will not be reported until the
    // next query
    //
    cells_query( sock );

    // say how long it took to push the stats through
    // the network last time
    //
    sock_printf( sock, "%.3g seconds for previous query socket to drain"
        " through network\n", query_drainage_delta );

    // append the caller's message to the socket
    //
    if( extra_msg )
        sock_puts( sock, (byte *)extra_msg );

#if REPORT_OVERHEAD_STATS
    start.set();
#endif

    result = flows_clear();
    if( result )
        return result;

    result = cells_clear();
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed2 = PentiumClock( now - start ).get_seconds();
    fprintf( stderr, "clear:%.4f\n", elapsed2 );
    sock_printf( sock, "clear:%.4f\n", elapsed2 );
#endif

    // close the TCP/IP connection from this end
    //
    // when caller closes or we time out, we will
    // know to listen again
    //
    sock_close( sock );

    // move on to the next socket, so somebody can
    // connect to us with a different TCP source port
    // before TIME_WAIT elapses on the old TCP port#
    //
    which_query_socket_prev_used = which_query_socket_to_use;
    which_query_socket_to_use++;
    if( which_query_socket_to_use >= QUERY_SOCKET_COUNT )
        which_query_socket_to_use = 0;

    return 0;
}



/*
 * summarize_rec
 *  print out fascinating detail about hashrec
 */
static void flows_summarize_one(
    char *cp,
    struct user_prefs *up,
    struct hashrec *hr
)
{
    CardClock tmp, tmp2, latest;
    float   duration;
    unsigned interface;

#if DEBUG_HOUSECLEAN
    dump_flow( stdout, hr );
#endif

    /* if rflag set, and no match, early out */
    if (rflag && !flowcmp(&up->up_fd, &hr->hr_fd))
        return;

    tmp = hr->hr_access - hr->hr_start;
    duration = tmp.get_seconds();
#if SUMMARIZE_ONE_PARANOIA
    if( (int)tmp.high < 0 )
    {
        printf( "error: duration %.f > 1000\n", duration );
        dump_flow( stdout, hr );
    }
#endif

    /* klt - add pkts_per_flow record */
    if (hr->hr_packets < PKTFLOW_ARRAY_SIZE_MAX) {
        pktflow[hr->hr_packets]++;
    } else {
        pktflow[0]++;
    }

    if (pflag) {

        tmp2 = 0.0;
        for( interface=0; interface<MAX_ATM_NICS; interface++ )
        {
            tmp = tstamp[ interface ] - istamp;
            if( tmp2 < tmp )
            {
                tmp2 = tmp;
                latest = tstamp[ interface ];
            }
        }

        printf(
           "A%s: %.6f\t%x:%x\t%d\t%.0f\t%.3f\t%d:%d:%d\t%.6f\n",
            cp,
            tmp2.get_seconds(),
            hr->hr_fd.fd_src.s_addr,
            hr->hr_fd.fd_dst.s_addr,
            hr->hr_packets,
            (double) hr->hr_bytes,
            duration,
            hr->hr_fd.fd_proto,
            hr->hr_fd.fd_sport,
            hr->hr_fd.fd_dport,
            latest.get_seconds()
        );
    }

    if( flowtrace_parms_are_filled
    && sock_established( &flowtrace_socket )
    && ((flowtrace_parms.behavior & fsel_ignore_iface) || flowtrace_parms.iface==hr->hr_iface)
#if KEEP_ATM_HEADER
    && ((flowtrace_parms.behavior & fsel_ignore_gfc) || flowtrace_parms.atm.bits.GFC==hr->hr_atm.bits.GFC)
    && ((flowtrace_parms.behavior & fsel_ignore_vpi) || flowtrace_parms.atm.bits.VPI==hr->hr_atm.bits.VPI)
    && ((flowtrace_parms.behavior & fsel_ignore_vci) || flowtrace_parms.atm.bits.VCI==hr->hr_atm.bits.VCI)
#endif
    && ((flowtrace_parms.behavior & fsel_ignore_proto) || flowtrace_parms.proto==hr->hr_proto)

    // either match sources against sources, or match sources against dests
    //
    && (  (((flowtrace_parms.behavior & fsel_ignore_src)   || flowtrace_parms.src==hr->hr_src)
        && ((flowtrace_parms.behavior & fsel_ignore_dst)   || flowtrace_parms.dst==hr->hr_dst)
        && ((flowtrace_parms.behavior & fsel_ignore_sport) || flowtrace_parms.sport==hr->hr_sport)
        && ((flowtrace_parms.behavior & fsel_ignore_dport) || flowtrace_parms.dport==hr->hr_dport))

       || (((flowtrace_parms.behavior & fsel_ignore_src)   || flowtrace_parms.src==hr->hr_dst)
        && ((flowtrace_parms.behavior & fsel_ignore_dst)   || flowtrace_parms.dst==hr->hr_src)
        && ((flowtrace_parms.behavior & fsel_ignore_sport) || flowtrace_parms.sport==hr->hr_dport)
        && ((flowtrace_parms.behavior & fsel_ignore_dport) || flowtrace_parms.dport==hr->hr_sport)))
    )
    {
        if( flowtrace_parms.behavior & fsel_output_text )
            dump_flow( &flowtrace_socket, hr );
        else
            sock_write( &flowtrace_socket, (Bit8 *)&hr->hr_fs, sizeof(flowstats) );
    }

#if !STUB_2ND_LEVEL_STATS
    update_stats( hr, duration );
#endif
}



/*
 *  this is called when EOF is reached
 *  to walk through still-pending flows
 *  and force them to be accumulated statistically
 */
static void flows_summarize_all( void )
{
    extern struct queue Q;
    struct hashrec *hp = Q.head;
    struct hashrec *next;

    while (hp != NULL) {
        flows_summarize_one("-EXIT", &prefs, hp);

        next = hp->hr_qlink.next;
        delque(hp);
        flows_delete(hp);

        hp = next;
    }
}



// called when program is about to exit
//
int flows_cleanup( void )
{
    int result;

    // must flush pending cells before flushing pending flows
    //
    result = cells_cleanup();

#if DEBUG_CHAIN_LENGTH
    fprintf( stderr, "chain length is %d max, %.3f avg\n", longest_chain, chainlen_sum/double(chains_used) );
    fprintf( stderr, "%u (%.1f%%) of %u buckets used\n", chains_used,
        (100.0*chains_used)/NFLOWBUCKETS, NFLOWBUCKETS );
#endif

    flows_summarize_all();

    // since nobody has called us when we are cleaning up,
    // there is no point in doing one last query, so this
    // code is commented out:
    //
    //     result = flows_query( NULL );
    //

    return result;
}



/*
 * klt - 10/31/96
 * sends data on all hashrecs in "A" queue
 * if active feed mode is on
 */
void flows_send_active( void )
{
    /*printf("DEBUG --> active flows record...\n");*/
    /*dump_flow( stdout, hp );*/
    struct hashrec *hpp, *next;
    extern struct queue A;
    static struct flowstats packet[ FLOWS_PER_UDP ];
    unsigned packet_fullness = 0;

    /* rifle thru A */
#if DEBUG_ACTIVE_FEED
    printf("Process thru A queue...\n");
#endif
    for( hpp=A.head; hpp; hpp=next )
    {
        next = hpp->hr_alink.next;

#if DEBUG_ACTIVE_FEED_DATA
#if BYTESWAP_FIELDS_EARLY
        printf("%X --> %X:", hpp->hr_fd.fd_src.s_addr, hpp->hr_fd.fd_dst.s_addr);
#else
        printf("%X --> %X:", ntohl(hpp->hr_fd.fd_src.s_addr), ntohl(hpp->hr_fd.fd_dst.s_addr));
#endif
        printf("%d pkts, %.0f bytes,",hpp->hr_packets,(double)hpp->hr_bytes);
#if BYTESWAP_FIELDS_EARLY
        printf("%d sport, %d dport,",hpp->hr_fd.fd_sport,hpp->hr_fd.fd_dport);
#else
        printf("%d sport, %d dport,",ntohs(hpp->hr_fd.fd_sport),ntohs(hpp->hr_fd.fd_dport));
#endif
        printf("proto: %d\n",hpp->hr_fd.fd_proto);
#endif

        /* pack the flowstats into the current packet */
        packet[ packet_fullness++ ] = hpp->hr_fs;

        /* output packet if it is full */
        if( packet_fullness >= FLOWS_PER_UDP )
        {
            sock_write(&msock, (byte *)&packet, packet_fullness*sizeof(struct flowstats));
#if DEBUG_ACTIVE_FEED
            printf("wrote %d flowstats to UDP socket\n", packet_fullness);
#endif
            packet_fullness = 0;
        }

        /* remove it from A queue */
        del_active_queue(hpp);
    }

    /* flush packet if it is not empty */
    if( packet_fullness > 0 )
    {
        sock_write(&msock, (byte *)&packet, packet_fullness*sizeof(struct flowstats));
#if DEBUG_ACTIVE_FEED
        printf("wrote %d trailing flowstats to UDP socket\n", packet_fullness);
#endif
    }
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

// this code doesn't have anything special to do
// when keys are pressed
//
void flows_keystroke( int keystroke )
{
}
// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par

#endif  // DO_FOREGROUND_STATS



// for diagnostic use
//
void dump_flow( void *sock, struct hashrec *hp )
{
    char scratch1[ 20 ];
    char scratch2[ 20 ];

    sock_printf( sock, "hashrec at 0x%.8X\n", hp );
    sock_printf( sock, "  packets: %d,   bytes: %.0f\n", hp->hr_packets, (double)hp->hr_bytes );
    sock_printf( sock, "  start time: %.6f,   access time: %.6f\n", hp->hr_start.get_seconds(), hp->hr_access.get_seconds() );
#if BYTESWAP_FIELDS_EARLY
    sock_printf( sock, "  srcIP: %s,   dstIP: %s\n",
        inet_ntoa( scratch1, hp->hr_fd.fd_src.s_addr ),
        inet_ntoa( scratch2, hp->hr_fd.fd_dst.s_addr )
    );
    sock_printf( sock, "  iface: %d,   proto: %d,   sport: %d,   dport: %d\n", hp->hr_iface, hp->hr_fd.fd_proto, hp->hr_fd.fd_sport, hp->hr_fd.fd_dport );
#else
    sock_printf( sock, "  srcIP: %s,   dstIP: %s\n",
        inet_ntoa( scratch1, ntohl( hp->hr_fd.fd_src.s_addr ) ),
        inet_ntoa( scratch2, ntohl( hp->hr_fd.fd_dst.s_addr ) )
    );
    sock_printf( sock, "  iface: %d,   proto: %d,   sport: %d,   dport: %d\n", hp->hr_iface, hp->hr_fd.fd_proto, ntohs(hp->hr_fd.fd_sport), ntohs(hp->hr_fd.fd_dport) );
#endif
#if REASM_IP_FRAGS
    sock_printf( sock, "  ident: %d,   offset: %d\n", ntohs(hp->hr_ident), hp->hr_len );
    sock_printf( sock, "  fragment overlaps: %d,   fragment holes: %d\n", hp->hr_olaps, hp->hr_holes );
#endif
#if KEEP_ATM_HEADER
    sock_printf( sock, "  ATM GFC:%d VPI:%d VCI:%d OAM:%d CE:%d LAST:%d CLP:%d\n",
        hp->hr_atm.bits.GFC,
        hp->hr_atm.bits.VPI,
        hp->hr_atm.bits.VCI,
        hp->hr_atm.bits.OAM,
        hp->hr_atm.bits.CE,
        hp->hr_atm.bits.LAST,
        hp->hr_atm.bits.CLP
    );
#endif
#if EXTRA_TCP_STATS
    if( hp->hr_fd.fd_proto == IPPROTO_TCP )
    {
        sock_printf( sock, "  TCP length > IP length: %u\n", hp->hr_badlen );
        sock_printf( sock, "  TCP flags: %u\n", hp->hr_tcpflags );
        sock_printf( sock, "  SEQ min: %u,   max: %u,   prev: %u\n", hp->hr_minseq, hp->hr_maxseq, hp->hr_prevseq );
        sock_printf( sock, "  SEQ hole min: %u,   hole max: %u\n", hp->hr_minhole, hp->hr_maxhole );
        sock_printf( sock, "  SEQ single holes: %u,   fill failures: %d,   multiple holes: %u\n", hp->hr_1holes, hp->hr_holepkts, hp->hr_nholes );
        sock_printf( sock, "  SEQ duplicates: %d,   backtracks: %d\n", hp->hr_dupseqs, hp->hr_backseqs );
        sock_printf( sock, "  ACK min: %u,   max: %u,   prev: %u\n", hp->hr_minack, hp->hr_maxack, hp->hr_prevack );
        sock_printf( sock, "  ACK duplicate packets: %d\n", hp->hr_dupacks );
        sock_printf( sock, "  RX window max: %u,   prev: %u\n", hp->hr_maxwin, hp->hr_prevwin );

#if CELLS_PER_PACKET > 1
        if( hp->hr_tcpopts & fs_tcp_opt_win_scale )
            sock_printf( sock, "  RX window scaling: %u bits\n", hp->hr_wscale );
        else
            sock_printf( sock, "  RX window scaling option not seen in SYN packet\n" );

        if( hp->hr_tcpopts & fs_tcp_opt_win_scale )
            sock_printf( sock, "  RX MSS: %u\n", hp->hr_mss );
        else
            sock_printf( sock, "  RX MSS option not seen in SYN packet\n" );

        sock_printf( sock, "  TCP other options: %s%s%s%s%s%s%s\n",
            hp->hr_tcpopts & fs_tcp_opt_sack_permitted ? "SACKPERMITTED " : "",
            hp->hr_tcpopts & fs_tcp_opt_sack ? "SACK " : "",
            hp->hr_tcpopts & fs_tcp_opt_timestamp ? "TIMESTAMP " : "",
            hp->hr_tcpopts & fs_tcp_opt_ttcp_cc ? "TTCP_CC " : "",
            hp->hr_tcpopts & fs_tcp_opt_ttcp_cc_new ? "TTCP_CC_NEW " : "",
            hp->hr_tcpopts & fs_tcp_opt_ttcp_cc_echo ? "TTCP_CC_ECHO " : "",
            hp->hr_tcpopts & fs_tcp_opt_unknown ? "UNKNOWN " : "",
            hp->hr_tcpopts & fs_tcp_opt_reserved ? "RESERVED " : ""
        );

        sock_printf( sock, "  TCP %s options had too short a length\n",
            hp->hr_tcpopts & fs_tcp_opt_too_short ? "some" : "no" );

        sock_printf( sock, "  TCP %s options were truncated by capture\n",
            hp->hr_tcpopts & fs_tcp_opt_truncated ? "some" : "no" );
#endif

        Bit32 good_bytes = hp->hr_maxseq - hp->hr_minseq;
        Bit64 total_bytes = hp->hr_bytes - (hp->hr_packets * (sizeof(ip) + sizeof(tcphdr)));
        if( hp->hr_tcpflags & TH_SYN )
            total_bytes = total_bytes + 1;
        if( hp->hr_tcpflags & TH_FIN )
            total_bytes = total_bytes + 1;

        // ???should nholes, 1holes, and holepkts fields
        // affect this value???
        //
        Bit32 good_packets = hp->hr_packets - hp->hr_backseqs - hp->hr_dupseqs;

        sock_printf( sock, "  total forward  goodput: %u bytes (%g%%),   %d packets (%g%%)\n",
            good_bytes,
            total_bytes==0
                ? 100
                : 100 * good_bytes / total_bytes,
            good_packets,
            hp->hr_packets==0
                ? 100
                : 100.0 * good_packets / hp->hr_packets
        );

        sock_printf( sock, "  total backward goodput: %u bytes (?%%),  ?-%d packets (?%%)\n",
            hp->hr_maxack - hp->hr_minack, hp->hr_dupacks );
    }
#endif
    sock_printf( sock, "\n" );
}
