/*
 * definitions for queue routines
 */

#ifndef _QUEUE_H
#define _QUEUE_H



/* conditional compilation switches
 */
#define EXTRA_TCP_STATS     TRUE
#define KEEP_ATM_HEADER     TRUE
#define REASM_IP_FRAGS      TRUE
#define KEEP_FRAG_SQUARES   FALSE



#include "../wattcp/tcp.h"              /* struct in_addr */
#include "../integrat/constant.h"       /* CARD_CLOCK_HEADER */
#include CARD_CLOCK_HEADER              /* class CardClock */
#include "../integrat/subnet.h"
#if KEEP_ATM_HEADER
#   include "../aal5/atm.h"             /* struct AtmHeaderCoerce */
#endif



struct queue_entry {
    struct hashrec *prev;
    struct hashrec *next;
};

struct queue {
    struct hashrec *head;
    struct hashrec *tail;
    int num_entries;
};

void insque( struct hashrec *qe );
void delque( struct hashrec *qe );
void endque( struct hashrec *qe );
void dumpq( void *sock );

/* klt - 10/31/96 */
void update_active_queue( struct hashrec *qe );
void del_active_queue( struct hashrec *qe );

/* Joel Apisdorf - 3/5/98 */
void insert_fragment_queue( struct hashrec *qe );
void delete_fragment_queue( struct hashrec *qe );

extern struct queue Q, A;
#if REASM_IP_FRAGS
    extern struct queue F;
#endif



/*
 * flowdata is what makes a flow unique
 *
 * length field here is just for the first packet,
 * or the most recently received fragmented packet
 */

struct flowdata {
    struct in_addr  fd_src;     /* SRC IP Address */
    struct in_addr  fd_dst;     /* DST IP Address */
    Bit16       fd_len;         /* Pkt length (from IP) */
    Bit8        fd_proto;       /* Protocol ID (IP) */
    Bit8        fd_iface;       /* interface index */
    Bit16       fd_sport;       /* Src. Port (TCP or UDP) */
    Bit16       fd_dport;       /* Dest. Port (TCP or UDP) */
#if REASM_IP_FRAGS
    Bit16       fd_ident;       /* IP packet identification */
    Bit8        fd_holes;       /* How many fragments left holes? */
    Bit8        fd_olaps;       /* how many fragments overlapped? */
#endif
    u_int       fd_index;       /* bucket number */
#if KEEP_ATM_HEADER
    AtmHeaderCoerce fd_atm;     /* ATM header */
#endif
};



/* what we know about all packets in a flow, so far
 *
 * all fields are in host byte order
 * when sent due to active feed or flow tracing,
 * except those that aren't swapped early
 */

struct flowstats {
    struct flowdata fs_fd;
    CardClock fs_start;             /* time flow established */
    CardClock fs_access;            /* flow last accessed */
    u_int fs_packets;               /* tot # packets on this flow */
    Bit64 fs_bytes;                 /* tot # bytes on this flow */
#if EXTRA_TCP_STATS
    Bit32 fs_tcp_seq_min;           /*  left edge of flow (inclusive) */
    Bit32 fs_tcp_seq_max;           /* right edge of flow (exclusive) */
    Bit32 fs_tcp_seq_hole_min;      /* left  edge of hole (inclusive) (==maxhole when filled) */
    Bit32 fs_tcp_seq_hole_max;      /* right edge of hole (exclusive) */
    Bit32 fs_tcp_seq_prev;          /* what TCP sequence # was seen last? */
    Bit32 fs_tcp_ack_min;           /* left  edge of backflow (inclusive) */
    Bit32 fs_tcp_ack_max;           /* right edge of backflow (exclusive) */
    Bit32 fs_tcp_ack_prev;          /* what TCP acknowledge # was seen last? */
    Bit8  fs_tcp_seq_n_holes;       /* how many holes did we detect after first multi? */
    Bit8  fs_tcp_seq_1_holes;       /* how many times have we had a single hole? */
    Bit8  fs_tcp_seq_hole_pkts;     /* how many packets did we get that didn't fill the hole? */
    Bit8  fs_tcp_seq_backs;         /* how many times was seq in pkt < max? */
    Bit8  fs_tcp_seq_dups;          /* how many times did seq definitely overlap old data? */
    Bit8  fs_tcp_ack_dups;          /* how many times were ack, seq and rwnd same as prev pkt?    */
    Bit8  fs_tcp_flags;             /* logical OR of flags from all TCP headers in flow */
    Bit8  fs_tcp_bad_length;        /* how many packets claimed TCP options beyond end of packet? */   /* should this be global??? */
    u_short fs_tcp_win_prev;        /* receive window in previous packet */
    u_short fs_tcp_win_max;         /* highest receive window seen */
#if CELLS_PER_PACKET > 1
#if HOST_IS_BIG_ENDIAN
    Bit16 fs_tcp_options   : 12;    /* which TCP options did we see in SYN packet? */
    Bit16 fs_tcp_win_scale : 4;     /* from TCP window scaling option in SYN packet */
#else
    Bit16 fs_tcp_win_scale : 4;     /* from TCP window scaling option in SYN packet */
    Bit16 fs_tcp_options   : 12;    /* which TCP options did we see in SYN packet? */
#endif
    Bit16 fs_tcp_max_segment;       /* from TCP MSS option in SYN packet */

#   define fs_tcp_opt_reserved          2048    /* "R"   for future OCxMON use */
#   define fs_tcp_opt_ttcp_cc_echo      1024    /* "E"   T/TCP connection count SYN-ACK echo */
#   define fs_tcp_opt_ttcp_cc_new       512     /* "N"   T/TCP connection count new SYN */
#   define fs_tcp_opt_ttcp_cc           256     /* "C"   T/TCP connection count */
#   define fs_tcp_opt_too_short         128     /* "H"   0 or 1 or too-small-for-option length seen (caused abort) */
#   define fs_tcp_opt_truncated         64      /* "D"   some options (or parts thereof) were in missing 2nd cell */
#   define fs_tcp_opt_unknown           32      /* "U"   any other option with length>=2 skipped (like T/TCP CC CC.NEW CC.ECHO) */
#   define fs_tcp_opt_timestamp         16      /* "T"   non-SYN detection not implemented yet */
#   define fs_tcp_opt_max_segment       8       /* "M"   */
#   define fs_tcp_opt_win_scale         4       /* "W"   */
#   define fs_tcp_opt_sack              2       /* "S"   non-SYN detection not implemented yet */
#   define fs_tcp_opt_sack_permitted    1       /* "P"   */
#endif
#endif
};



/*
 * hashrecord contains:
 *      flow stats
 *      forward/backward links for 2x-link hash chains
 */

struct hashrec {
    struct queue_entry hr_qlink;    /* earlier access times first */
    struct queue_entry hr_alink;    /* klt - 10/31/96 - active feed, least recently tagged first (not tagged again if touched again before sending) */
    struct queue_entry hr_clink;    /* hash table collision resolution, earlier start times first */
#if REASM_IP_FRAGS
    struct queue_entry hr_flink;    /* only has IP fragments */
#endif
    struct flowstats hr_fs;

#   define hr_fd        hr_fs.fs_fd
#   define hr_start     hr_fs.fs_start
#   define hr_access    hr_fs.fs_access
#   define hr_packets   hr_fs.fs_packets
#   define hr_bytes     hr_fs.fs_bytes
#   define hr_index     hr_fs.fs_fd.fd_index
#   define hr_minseq    hr_fs.fs_tcp_seq_min
#   define hr_maxseq    hr_fs.fs_tcp_seq_max
#   define hr_prevseq   hr_fs.fs_tcp_seq_prev
#   define hr_minhole   hr_fs.fs_tcp_seq_hole_min
#   define hr_maxhole   hr_fs.fs_tcp_seq_hole_max
#   define hr_nholes    hr_fs.fs_tcp_seq_n_holes
#   define hr_1holes    hr_fs.fs_tcp_seq_1_holes
#   define hr_holepkts  hr_fs.fs_tcp_seq_hole_pkts
#   define hr_dupseqs   hr_fs.fs_tcp_seq_dups
#   define hr_backseqs  hr_fs.fs_tcp_seq_backs
#   define hr_minack    hr_fs.fs_tcp_ack_min
#   define hr_maxack    hr_fs.fs_tcp_ack_max
#   define hr_prevack   hr_fs.fs_tcp_ack_prev
#   define hr_dupacks   hr_fs.fs_tcp_ack_dups
#   define hr_prevwin   hr_fs.fs_tcp_win_prev
#   define hr_maxwin    hr_fs.fs_tcp_win_max
#   define hr_wscale    hr_fs.fs_tcp_win_scale
#   define hr_mss       hr_fs.fs_tcp_max_segment
#   define hr_tcpopts   hr_fs.fs_tcp_options
#   define hr_tcpflags  hr_fs.fs_tcp_flags
#   define hr_badlen    hr_fs.fs_tcp_bad_length
#   define hr_iface     hr_fs.fs_fd.fd_iface
#   define hr_olaps     hr_fs.fs_fd.fd_olaps
#   define hr_holes     hr_fs.fs_fd.fd_holes
#   define hr_ident     hr_fs.fs_fd.fd_ident
#   define hr_len       hr_fs.fs_fd.fd_len
#   define hr_atm       hr_fs.fs_fd.fd_atm
#   define hr_src       hr_fs.fs_fd.fd_src.s_addr
#   define hr_dst       hr_fs.fs_fd.fd_dst.s_addr
#   define hr_sport     hr_fs.fs_fd.fd_sport
#   define hr_dport     hr_fs.fs_fd.fd_dport
#   define hr_proto     hr_fs.fs_fd.fd_proto
};



#define MAX_FRAG_ATMS   15

struct fraghashrec {
#if KEEP_ATM_HEADER
    u_int n_atm : 4;    /* how many ATM headers are filled in, minus one (max means overflow) */
    u_int iface : 4;    /* which cardindex were all these flows on? */
#else
    u_int iface : 8;    /* which cardindex were all these flows on? */
#endif
    u_int proto : 8;
    u_int dport : 16;   /* destination (TCP or UDP or ICMP) port */
    u_int dst;          /* destination IP address */
#if KEEP_ATM_HEADER
    AtmHeaderCoerce atm[ MAX_FRAG_ATMS ];
#endif
    u_int holes, olaps;
    u_int flows, packets;
    Bit64 bytes;
    double duration;
#if KEEP_FRAG_SQUARES
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
#endif
    struct fraghashrec *next;
};



struct porthashrec {
    u_int padding : 8; /* SGI & Sun require, PC wants even multiple of 4 bytes */
    u_int proto : 8;
    u_int sport : 16;  /* Src. Port (TCP or UDP) */
    u_int dport : 16;  /* Dest. Port (TCP or UDP) */
    u_int flows, packets;
    Bit64 bytes;
    double duration;
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
    struct porthashrec *next;
};



struct nethashrec {
    u_short src;                /* AS number */
    u_short dst;                /* AS number */
    u_int flows, packets;
    Bit64 bytes;
    double duration;
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
    struct nethashrec *next;
};



enum hosthashgroup
{
    WebServer = 0,
    SnmpRequest,
    PingRequest,
    FtpData,
    FtpControl,
    Other,
    HostGroupCount
};

struct hosthashrec {
    u_int ip;
    u_char proto;
    u_char group;
    u_short port;
    u_int flows, packets;
    Bit64 bytes;
    double duration;
    double p2sum, b2sum, d2sum; /* sum sq's to get var */
    double pbsum, bdsum, pdsum; /* sum prod's to get covar */
    struct hosthashrec *next;
};



/* which flows do we want to know about when they expire?
 *
 * all fields are in network byte order when sent on network
 */
struct flowselector {
    Bit8        iface;       /* interface index */
    Bit8        proto;       /* IP protocol ID */
    Bit16       behavior;    /* see fsel_XXX #define's below */
    Bit16       sport;       /* TCP/UDP/ICMP source port */
    Bit16       dport;       /* TCP/UDP destination port */
    Bit32       src;         /* IP source address */
    Bit32       dst;         /* IP destination address */
#if KEEP_ATM_HEADER
    AtmHeaderCoerce atm;     /* ATM GFC/VPI/VCI */
#endif

#define fsel_ignore_iface    1
#define fsel_ignore_proto    2
#define fsel_ignore_sport    4
#define fsel_ignore_dport    8
#define fsel_ignore_src      16
#define fsel_ignore_dst      32
#define fsel_ignore_gfc      64
#define fsel_ignore_vpi      128
#define fsel_ignore_vci      256

#define fsel_output_text     32768      /* cleared causes binary output */
};



#endif
