// some notes on memory management:
//
// I have chosen to call malloc() explicitly in this module,
// and want  subnet.c to be compiled to use flat_alloc()
// and want readbgp.c to be compiled to map flat_alloc() to malloc()
// and want environment variable SET DPMIMEM=MAXMEM 62400
// and want 82439HX extended cachability turned off in BIOS setup
//
// all the above should all force
//     routing table info (line numbers, address, mask, and AS path),
//     program code,
//     DPMI32 server code,
//     DOS,
//     BIOS, and
//     list of IP addresses to look up
// to be in cachable memory below the 64 MB line, and
//     the radix lookup nodes
// to be in uncacheable memory above the 64 MB line
//
// because the radix nodes will exhibit no locality of reference
// (except the root table) and is better off not waiting for a
// whole cache line (32 bytes) to be loaded from DRAM when the
// inevitable cache misses occur
//
// result: 308.6 nS per IP address lookup
//
//
//
// alternative to the above: SET DPMIMEM=MAXMEM 62272
// so the root node (128 KB) is cached, but other nodes aren't
//
// result: 231.4 nS per IP address lookup
//
//
//
// OK I give up, put the whole mess in cachable memory,
// still using flat_alloc(), by SET DPMIMEM=MAXMEM 32768
//
// result: 184.7 nS per IP address lookup
//
//
//
// now try allocating subnet_pool statically, instead of using
// flat_alloc() or malloc()
//
// result: 186   nS per IP address lookup
//



#include <mem.h>                        // memset()
#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "trace.h"
#include "..\intel\byteswap.h"          // htons() htonl()
#include "buffer.h"                     // BufferBlock COMPL_RING_ENTRIES
#include "readbgp.h"                    // read_bgp_file()
#include "subnet.h"                     // Subnet FindSubnet()
#include "dostats.h"                    // nonzero()
#include "..\pci\pci.h"                 // pci_is_bios_present() pci_set_cache_control()
#include "aal5.h"                       // struct cellhashrec



// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define TRACE_ALREADY_SWAPPED           TRUE
#define WHINE_ABOUT_UNROUTABLES         FALSE
#define LIST_ROUTES_FOUND               FALSE
#define DISABLE_L2_CACHING_GT_64_MB     FALSE
#define DISABLE_L1_CACHE_DURING_FIND    FALSE
#define SUBTRACT_CALIBRATION            TRUE
#define PRECALIBRATE_MANY_TIMES         TRUE
#define CALIBRATE_MANY_TIMES            FALSE
#define SHOW_RAW_SUBNETS                FALSE
#define SHOW_CLEAN_SUBNETS              FALSE
#define SHOW_TRACE_PROGRESS             TRUE
#define SHOW_FIND_PROGRESS              TRUE
#define USE_HARDCODED_CALIBRATION       TRUE
#define EXTRA_FIND_PASS                 FALSE
#define FORCE_ROOT_NODE_INTO_CACHE      TRUE
#define ROUTE_DELETION_TESTING          TRUE




// manifest constants
//
#define MAX_IP_PACKETS                  (123 * COMPL_RING_ENTRIES)
#define MAX_TIME_HISTOGRAM              100000
#define HARDCODED_CALIBRATION           145e-9
//#define HARDCODED_CALIBRATION           4200e-9



// global variables
//
Bit8 header[ SECTOR_SIZE ];
traceblock *buffer = (traceblock *) header;
BufferBlock *block;
typedef struct _AddressPair AddressPair;
struct _AddressPair
{
    Bit32 dst;
    Bit32 src;
} *ip_addresses;
Bit32 timings[ MAX_SUBNET_MEMREFS ][ MAX_TIME_HISTOGRAM ];


int main( int argc, char *argv[] )
{
    const char *program   = argv[ 0 ];
    const char *tracefile = argv[ 1 ];
    const char *bgpfile   = argv[ 2 ];
#if WANT_TIMES
    const char *outfile   = argv[ 3 ];
    const char *cardstr   = argv[ 4 ];
#else
    const char *cardstr   = argv[ 3 ];
#endif
    FILE *in_handle;
    FILE *out_handle;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 block_count = 0;
    Bit32 valid_packets = 0;
    Bit32 non_ip_packets = 0;
    Bit32 wrong_cardnum_packets = 0;
    Bit32 wrong_src_net_packets = 0;
    Bit32 wrong_dst_net_packets = 0;
    Bit32 when_temp;
    struct cellhashrec packet;
    Bit16 cardnum_wanted;
    Subnet *route;
    Subnet route_to_delete;
    char ip_string[ 30 ];
    PentiumClock before, after;
    PentiumClock calibration;
    PentiumClock total;
    double elapsed, elapsed2;
#if DISABLE_L2_CACHING_GT_64_MB || DISABLE_L1_CACHE_DURING_FIND
    Bit8 old_pci_cache_control;
#endif
    union { PciCacheControl pci_cache_control; Bit8 pci_cache_control_alias; };
    Bit8 value;
    Bit16 handle;
    unsigned index;
    unsigned refcnt;

    fprintf( stdout, "each record is %d bytes\n", sizeof(struct tracerec) );
    fprintf( stdout, "ICMP alone accounts for %d bytes\n", sizeof(struct icmp) );
    fprintf( stdout, "%d records in a block\n\n", COMPL_RING_ENTRIES );

    if( argc < 3 )
    {
        fprintf( stderr, "calculates routes for blocks in an OC3MON trace file\n" );
        fprintf( stderr, "usage: %s trace_file bgp_file {hex-cardnum}\n", program );
        return 1;
    }

    if( argc >= 4 )
    {
        result = sscanf( cardstr, "%hX", &cardnum_wanted );
        if( result != 1 )
        {
            fprintf( stderr, "unable to convert cardnum '%s' to integer\n", cardstr );
            return 4;
        }
    }
    else
        cardnum_wanted = 0;

    // we are talking to the chipset itself, not any of the PCI cards
    //
    handle = 0;

    // show current chipset cache settings
    //
#if DISABLE_L2_CACHING_GT_64_MB || DISABLE_L1_CACHE_DURING_FIND
    old_pci_cache_control =
#endif
    pci_cache_control_alias = pci_get_cache_control( handle );
    printf( "CPU cache control register was 0x%2.2X (first level cache %s,"
        " secondary cache miss/invalidate %sforced,"
#if INTEL_82439HX
        " L2 cache up to %s,"
#endif
        " NA# signal %s, SRAM type %d, secondary cache size %dK)\n",
        pci_cache_control_alias,
        pci_cache_control.first_level_cache_enable ? "enabled" : "disabled",
        pci_cache_control.secondary_cache_force_miss_or_invalidate ? "is " : "not ",
#if INTEL_82439HX
        pci_cache_control.extended_cachability_enable ? "512 MB" : "64 MB",
#endif
        pci_cache_control.na_pound_signal_disable ? "disabled" : "enabled",
        pci_cache_control.static_ram_type,
        pci_cache_control.secondary_cache_size * 256
    );

#if DISABLE_L2_CACHING_GT_64_MB
    // now change the chipset cache settings
    //
    pci_cache_control.extended_cachability_enable = FALSE;
    pci_set_cache_control( handle, pci_cache_control_alias );
    value = pci_get_cache_control( handle );
    if( value != pci_cache_control_alias )
        printf( "unable to set CPU cache control register to 0x%2.2X; is now 0x%2.2X\n",
            pci_cache_control_alias, value );
    else
        printf( "disabled new caching beyond 64 MB\n" );

    // I wish I could flush the data cache here,
    // but the WBINVD Pentium (and 486) instruction
    // would cause a protection fault, since I am
    // running in ring 3 (least priveleged) not 0
    //
    //__emit__( 0x0f, 0x09 );
#endif

    /* initialize our memory allocator
     */
    flat_init();

    /* initialize the subnet extraction module
     */
    InitSubnet();

    /* parse the main BGP table into a radix trie that points at subnets
     */
    result = read_bgp_file( bgpfile );
    if( result )
    {
        fprintf( stderr, "error %d reading BGP file '%s'\n", result, bgpfile );
        return 5;
    }

#if ROUTE_DELETION_TESTING
    /* dump the raw routing table before deletion
    DumpRawSubnets();
     */

    /* try deleting a subnet
     */
    route_to_delete.addr = htonl( inet_addr( "24.0.0.0" ) );
    route_to_delete.mask = htonl( get_mask_from_length( 14 ) );
    result = DeleteSubnet( &route_to_delete );
    if( result == FALSE )
        fprintf( stderr, "error deleting route\n" );

    /* dump the raw routing table after deletion
    DumpRawSubnets();
     */
#endif

    /* allocate buffers for IP addresses we look up in the second pass
     */
    ip_addresses = (AddressPair *) malloc( MAX_IP_PACKETS * sizeof( AddressPair ) );
    if( !ip_addresses )
    {
        fprintf( stderr, "error: cannot allocate IP address array\n" );
        return 6;
    }

#if SHOW_RAW_SUBNETS
    DumpRawSubnets();
#endif
#if SHOW_CLEAN_SUBNETS
    PrintSubnets( (char *)ip_addresses, MAX_IP_PACKETS*sizeof(AddressPair) );
    fprintf( stdout, "%s\n", (char *)ip_addresses );
#endif

    block = new BufferBlock;

    CardClock::sync_time.set();

    in_handle = fopen( tracefile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

#if SHOW_TRACE_PROGRESS
    /* print the beginning part of the status line,
     * so the user knows what phase we are in
     */
    fprintf( stderr, "\nreading tracefile byte %9u", 0 );
#endif

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        block_count++;
#if SHOW_TRACE_PROGRESS
        fprintf( stderr, "\b\b\b\b\b\b\b\b\b%9u", block_count*sizeof(struct traceblock) );
#endif

        // read the next block header
        //
        result = fread( header, sizeof(header), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d header\n", errno, block_count );

            break;
        }

        // read the trace records into the BufferBlock
        //
        // note that this must NOT be conditional on our
        // liking the block, or we will be unable to read
        // the next block!
        //
        result = fread( block->buffers, COMPL_RING_ENTRIES*sizeof(tracerec), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d data\n", errno, block_count );

            break;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the usage count and card number
        //
        buffer->used = htonl( buffer->used );
        buffer->cardnum = htons( buffer->cardnum );
#endif

        // must remember usage count
        //
        // at this point it is in host byte order
        //
        used = buffer->used;

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            fprintf( stderr, "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        if( cardnum_wanted!=0 && buffer->cardnum!=cardnum_wanted )
        {
            fprintf( stderr, "ignoring cardnum 0x%.4X\n", buffer->cardnum );
            wrong_cardnum_packets += used;
            continue;
        }

        // copy the fields we like to the BufferBlock
        //
        block->cardnum = buffer->cardnum;
        block->cardindex = buffer->cardindex;
        block->used = buffer->used;
        block->queried = 0;

        // for all records in the BufferBlock
        //
        for( count=0; count<block->used; count++ )
        {
            packet.first = (tracerec *) (block->buffers + count*BUFFER_SIZE);
            packet.second = NULL;
            packet.last = (tracerec_last *) packet.first;

            /* if the packet is not valid IP...
             */
            if( packet.tr_lsap != LSAP_SNAP
            ||  packet.tr_ctl != CTL_SNAP
            ||  packet.tr_oui[0] != 0
            ||  packet.tr_oui[1] != 0
            ||  packet.tr_oui[2] != 0
            ||  packet.tr_etype != htons( ETHERTYPE_IP ) )
            {
                non_ip_packets++;
                continue;
            }

            // don't bother swqapping the words of the
            // ATM header and the timestamp because we
            // aren't going to reference those fields
            //

            // the rest of the data is IP and TCP headers,
            // and was always in network byte order!
            //

            // remember the source and destination IP addresses
            // for later lookup in the routing table
            //
            ip_addresses[ valid_packets ].dst = packet.tr_dst;
            ip_addresses[ valid_packets ].src = packet.tr_src;

            // count only the packets that we saved addresses from
            //
            valid_packets++;
        }
    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    fclose( in_handle );
    delete block;

    // show tracefile statistics
    //
    fprintf( stdout, "\n" );
    fprintf( stdout, "%lu blocks examined\n", block_count );
    fprintf( stdout, "%lu valid packets found\n", valid_packets );
    fprintf( stdout, "%lu non-IP packets found\n", non_ip_packets );
    if( cardnum_wanted != 0 )
        fprintf( stdout, "%lu packets without cardnum 0x%.4X found\n", wrong_cardnum_packets, cardnum_wanted );

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        fprintf( stderr, "Sorry, PCI BIOS is not present.\n" );
        return 7;
    }

    // no packets looked up yet
    //
    total = (Bit32) 0;
    for( refcnt=0; refcnt<MAX_SUBNET_MEMREFS; refcnt++ )
        for( count=0; count<MAX_TIME_HISTOGRAM; count++ )
            timings[ refcnt ][ count ] = 0;

    // find out how much time it takes to measure the time
    //
    // in case you don't feel like counting them, the 10
    // lines below each have 10 calls to calibration.set()
    //
#if PRECALIBRATE_MANY_TIMES
    before.set(); before.set(); before.set(); before.set(); before.set(); before.set(); before.set(); before.set(); before.set(); before.set();
#else
    before.set();
#endif
#if CALIBRATE_MANY_TIMES
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
    calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set(); calibration.set();
#else
    calibration.set();
#endif
    calibration -= before;
#if CALIBRATE_MANY_TIMES
    calibration.low /= 100;
#endif
    elapsed = calibration.get_seconds() * 1e9;
    fprintf( stdout, "\n" );
    fprintf( stdout, "calibration says %5.1f nS for measurement overhead\n", elapsed );
    fprintf( stdout, "raw elapsed low=%d high=%d\n", calibration.low, calibration.high );
#if USE_HARDCODED_CALIBRATION
    calibration.put_seconds( HARDCODED_CALIBRATION );
#endif

#if EXTRA_FIND_PASS
#if SHOW_FIND_PROGRESS
    /* print the beginning part of the status line,
     * so the user knows what phase we are in
     */
    fprintf( stderr, "pass1 finding IP addresses in packet %9d", 0 );
#endif

    // for all IP addresses we got in the above phase...
    //
    for( count=0; count<valid_packets; count++ )
    {
#if SHOW_FIND_PROGRESS
        if( !(count % 10000) )

            /* show the current packet number on stderr
             */
            fprintf( stderr, "\b\b\b\b\b\b\b\b\b%9d", count );
#endif

        // look up the routing of the both addresses
        //
        route = FindSubnet( ip_addresses[count].dst );
        route = FindSubnet( ip_addresses[count].src );
    }

#if SHOW_FIND_PROGRESS
    /* show the last packet number on stderr,
     * in case it wasn't an even multiple of 10000
     */
    fprintf( stderr, "\b\b\b\b\b\b\b\b\b%9d\n", count );
#endif
#endif  /* EXTRA_FIND_PASS */

#if FORCE_ROOT_NODE_INTO_CACHE
    ForceSubnetCaching();
#endif

#if DISABLE_L1_CACHE_DURING_FIND
    // now change the chipset cache settings
    //
    pci_cache_control.first_level_cache_enable = FALSE;
    pci_set_cache_control( handle, pci_cache_control_alias );
    value = pci_get_cache_control( handle );
    if( value != pci_cache_control_alias )
        printf( "unable to set CPU cache control register to 0x%2.2X; is now 0x%2.2X\n",
            pci_cache_control_alias, value );
    else
        printf( "disabled all new caching\n" );
#endif

#if EXTRA_FIND_PASS && SHOW_FIND_PROGRESS
    fprintf( stderr, "pass2 " );
#endif

    // remember when loop began for gross measurement
    //
    before.set();

#if SHOW_FIND_PROGRESS
    /* print the beginning part of the status line,
     * so the user knows what phase we are in
     */
    fprintf( stderr, "finding IP addresses in packet %9d", 0 );
#endif

    // for all IP addresses we got in the above phase...
    //
    for( count=0; count<valid_packets; count++ )
    {
#if SHOW_FIND_PROGRESS
        if( !(count % 10000) )

            /* show the current packet number on stderr
             */
            fprintf( stderr, "\b\b\b\b\b\b\b\b\b%9d", count );
#endif

        // look up the routing of the destination IP address
        //
        route = FindSubnet( ip_addresses[count].dst );

        // get the timing for the node memory accesses of the find
        //
        after = GetLastFindSubnetTiming();

        // get the memory reference count of the find
        //
        refcnt = GetLastFindSubnetRefcnt() - 1;

#if SUBTRACT_CALIBRATION
        // try to eliminate measurement overhead
        //
        after -= calibration;
#endif

        // update the net average
        //
        total += after;

        // update the histogram based upon tenths of nanoseconds
        //
        index = after.get_seconds() * 10E9;
        if( index > MAX_TIME_HISTOGRAM - 1 )
            index = MAX_TIME_HISTOGRAM - 1;
        timings[ refcnt ][ index ]++;

        // check the results of the lookup
        //
        if( !route )
        {
            wrong_dst_net_packets++;

#if WHINE_ABOUT_UNROUTABLES
            printf( "packet %7d dst IP address %-15s unroutable\n",
                count, inet_ntoa( ip_string, ntohl( ip_addresses[count].dst ) ) );
#endif
        }
#if LIST_ROUTES_FOUND
        else
            printf( "%s/%d ", inet_ntoa(ip_string,ntohl(route->addr)), bits_in_mask(route->mask) );
#endif

        // look up the routing of the source IP address
        // and do the same measurements as we did above
        // for the dest IP address
        //
        route = FindSubnet( ip_addresses[count].src );
        after = GetLastFindSubnetTiming();
        refcnt = GetLastFindSubnetRefcnt() - 1;
#if SUBTRACT_CALIBRATION
        after -= calibration;
#endif
        total += after;
        index = after.get_seconds() * 10E9;
        if( index > MAX_TIME_HISTOGRAM - 1 )
            index = MAX_TIME_HISTOGRAM - 1;
        timings[ refcnt ][ index ]++;
        if( !route )
        {
            wrong_src_net_packets++;

#if WHINE_ABOUT_UNROUTABLES
            printf( "packet %7d src IP address %-15s unroutable\n",
                count, inet_ntoa( ip_string, ntohl( ip_addresses[count].src ) ) );
#endif
        }
#if LIST_ROUTES_FOUND
        else
            printf( "%s/%d\n", inet_ntoa(ip_string,ntohl(route->addr)), bits_in_mask(route->mask) );
#endif
    }

#if SHOW_FIND_PROGRESS
    /* show the last packet number on stderr,
     * in case it wasn't an even multiple of 10000
     */
    fprintf( stderr, "\b\b\b\b\b\b\b\b\b%9d", count );
#endif

    // calculate gross elapsed time
    //
    after.set();
    after -= before;

    // we don't need any of that memory we allocated anymore
    //
    flat_exit();

#if DISABLE_L2_CACHING_GT_64_MB || DISABLE_L1_CACHE_DURING_FIND
    // now restore the old chipset cache settings
    //
    pci_set_cache_control( handle, old_pci_cache_control );
    value = pci_get_cache_control( handle );
    if( value != old_pci_cache_control )
        printf( "unable to restore CPU cache control register to 0x%2.2X; is now 0x%2.2X\n",
            old_pci_cache_control, value );
    else
        printf( "restored CPU cache control register to 0x%2.2X\n", value );
#endif

    // calculate net elapsed time and print it and gross
    //
    fprintf( stdout, "\n" );
    elapsed = total.get_seconds();
    fprintf( stdout, "%g seconds to look up %d IP addresses\n", elapsed, 2*valid_packets );
    elapsed2 = after.get_seconds();
    fprintf( stdout, "%g seconds from start to end of loop, so %g seconds must have\n"
       " been taken up by measurement, byte swapping, looping, and function calling\n",
       elapsed2, elapsed2-elapsed );
    elapsed = elapsed * 5e8 / nonzero(valid_packets);
    fprintf( stdout, "average %5.1f nanoseconds per IP address\n", elapsed );

    // print the histogram of individual elapsed times,
    // except for the times with a zero count
    //
    for( refcnt=0; refcnt<MAX_SUBNET_MEMREFS; refcnt++ )
    {
        fprintf( stdout, "\nwhen find needed %d memory references:\n", refcnt+1 );
        for( count=0; count<MAX_TIME_HISTOGRAM; count++ )
            if( timings[ refcnt ][ count ] > 0 )
                fprintf( stdout, "%6.1f nS seen on %7d lookups\n", (double)count/10, timings[refcnt][count] );
    }

    // show routing statistics
    //
    fprintf( stdout, "\n" );
    fprintf( stdout, "%3d IP packets had unroutable src address\n", wrong_src_net_packets );
    fprintf( stdout, "%3d IP packets had unroutable dst address\n", wrong_dst_net_packets );
    ShowSubnetStatistics();

    // return successfully
    //
    return 0;
}

