/*
 * util
 *
 *  this file contains a number of things..
 *
 *  the "objtypes" array contains different update
 *  functions which depend on args supplied by the
 *  user, and they basically delete or adjust various
 *  fields in the flowdata structure so that subsequent
 *  hashes and compares match.  This is a bit easier
 *  than having a separate hash and compare function
 *  depending on what the user types
 */

#include <sys/types.h>
#include <string.h>
#include <stdio.h>

#if __unix__
#   include <sys/time.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#   include "types.h"
#else
//#   include <winsock.h>
#   include "..\types\types.h"
#endif

#include "trace.h"
#include "util.h"




// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



static void zero_fields( struct flowdata *fd );

struct objdesc objtypes[]  = {
    "pq", update_pq,                    // must be first
    "sh", update_sh,
    "sn", update_sn,
    "dh", update_dh,
    "dn", update_dn,
    "np", update_np,
    "hp", update_hp
};



int findobj( char *cp )
{
    register struct objdesc *p = objtypes;

    if (!cp)
        return (-1)
        ;
    while (p < objtypes + NOBJ) {
        if (strcmp(p->option, cp) == 0)
            return (p - objtypes);
        p++;
    }
    return (-1);
}



void dumpobjtypes()
{
    register struct objdesc *p;

    for (p = objtypes; p < objtypes + NOBJ; p++)
        printf("%s ", p->option);
    printf("\n");
}



/* "port quad"
 */
void update_pq( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    fd->fd_len   = tr->tr_len;
    fd->fd_proto = tr->tr_proto;
    fd->fd_src.s_addr   = tr->tr_src;
    fd->fd_dst.s_addr   = tr->tr_dst;
    fd->fd_iface = interface;

    if (fd->fd_proto == IPPROTO_TCP || fd->fd_proto == IPPROTO_UDP)
    {
        fd->fd_sport = tr->tr_sport;
        fd->fd_dport = tr->tr_dport;
        return;
    }

    fd->fd_dport = 0;
    if (fd->fd_proto == IPPROTO_ICMP)
        fd->fd_sport = tr->tr_sport;        /* actually copies type and code */
    else
        fd->fd_sport = 0;
}



/* "source host"
 */
void update_sh( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    zero_fields(fd);
    fd->fd_src.s_addr = tr->tr_src;
    fd->fd_len = tr->tr_len;
    fd->fd_iface = interface;
}



/* "destination host"
 */
void update_dh( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    zero_fields(fd);
    fd->fd_dst.s_addr = tr->tr_dst;
    fd->fd_len = tr->tr_len;
    fd->fd_iface = interface;
}



/* "destination network"
 */
void update_dn( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    u_int net=(u_int)(tr->tr_dst)&0xffffff00;
    if(net < 0xc0000000){net = net & 0xffff0000;}
    if(net < 0x80000000){net = net & 0xff000000;}
    zero_fields(fd);
    fd->fd_dst.s_addr = net;
    fd->fd_len = tr->tr_len;
    fd->fd_iface = interface;
}



/* "source network"
 *
 * I know that the interface parm isn't used
 * and I don't want the compiler to nag me about it
 */
void update_sn( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    u_int net=(u_int)(tr->tr_src)&0xffffff00;
    if(net < 0xc0000000){net = net & 0xffff0000;}
    if(net < 0x80000000){net = net & 0xff000000;}
    zero_fields(fd);
    fd->fd_src.s_addr = net;
    fd->fd_len = tr->tr_len;
    fd->fd_iface = interface;
}



/* "network pair"
 */
void update_np( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    u_int dstnet=(u_int)(tr->tr_dst)&0xffffff00;
    u_int srcnet=(u_int)(tr->tr_src)&0xffffff00;
    if(srcnet < 0xc0000000){srcnet = srcnet & 0xffff0000;}
    if(srcnet < 0x80000000){srcnet = srcnet & 0xff000000;}
    if(dstnet < 0xc0000000){dstnet = dstnet & 0xffff0000;}
    if(dstnet < 0x80000000){dstnet = dstnet & 0xff000000;}
    zero_fields(fd);
    fd->fd_src.s_addr = srcnet;
    fd->fd_dst.s_addr = dstnet;
    fd->fd_len = tr->tr_len;
    fd->fd_iface = interface;
}



/* "host pair"
 */
void update_hp( cellhashrec *tr, flowdata *fd, Bit8 interface )
{
    zero_fields(fd);
    fd->fd_src.s_addr = tr->tr_src;
    fd->fd_dst.s_addr = tr->tr_dst;
    fd->fd_len = tr->tr_len;
    fd->fd_iface = interface;
}



#endif      // DO_FOREGROUND_STATS



static void zero_fields( struct flowdata *fd )
{
    memset(fd, '\0', sizeof(*fd));
}
