#include "..\wattcp\copyrigh.h"
#include "..\intel\timestmp.h"
#include "..\wattcp\wattcp.h"
#include <mem.h>

/*
 * ICMP - RFC 792
 */

static char *unreach[] = {
        "Network Unreachable",
        "Host Unreachable",
        "Protocol Unreachable",
        "Port Unreachable",
        "Fragmentation needed and DF set",
        "Source Route Failed" };
static char *exceed[] = {
        "TTL exceeded in transit",
        "Frag ReAsm time exceeded" };

static char *redirect[] = {
        "Redirect for Network",
        "Redirect for Host",
        "Redirect for TOS and Network",
        "Redirect for TOS and Host" };

/* constants */


typedef struct icmp_unused {
    byte        type;
    byte        code;
    word        checksum;
    longword    unused;
    in_Header   ip;
    byte        spares[ 8 ];
};

typedef struct icmp_pointer {
    byte        type;
    byte        code;
    word        checksum;
    byte        pointer;
    byte        unused[ 3 ];
    in_Header   ip;
};

typedef struct icmp_ip {
    byte        type;
    byte        code;
    word        checksum;
    longword    ipaddr;
    in_Header   ip;
};

typedef struct icmp_echo {
    byte        type;
    byte        code;
    word        checksum;
    word        identifier;
    word        sequence;
};

typedef struct icmp_timestamp {
    byte        type;
    byte        code;
    word        checksum;
    word        identifier;
    word        sequence;
    longword    original;       /* original timestamp */
    longword    receive;        /* receive timestamp */
    longword    transmit;       /* transmit timestamp */
};

typedef struct icmp_info {
    byte        type;
    byte        code;
    word        checksum;
    word        identifier;
    word        sequence;
};

typedef union  {
        struct icmp_unused      unused;
        struct icmp_pointer     pointer;
        struct icmp_ip          ip;
        struct icmp_echo        echo;
        struct icmp_timestamp   timestamp;
        struct icmp_info        info;
} icmp_pkt;

typedef struct _pkt {
    in_Header   in;
    icmp_pkt    icmp;
    in_Header   data;
};

static word icmp_id = 0;

static longword ping_host_cache = 0;
static word ping_identifier_cache = 0;
static word ping_sequence_cache;
static PentiumClock ping_delta_when_cache;
extern word multihomes;

/* you must be C++ code to call this
 */
int _chk_ping( longword host, word *identifier, word *sequence, PentiumClock *delta_when )
{
    if ( ping_host_cache != host )
        return -1;

    ping_host_cache = 0;
    *identifier = ping_identifier_cache;
    *sequence = ping_sequence_cache;
    *delta_when = ping_delta_when_cache;

    return 0;
}


// turn off compiler warning about unused parameter icmp
// since I know it won't be used and I must have
// it because calling code cannot be changed
//
#pragma warn -par

static void icmp_print( icmp_pkt *icmp, char *msg)
{
    byte *dest = (byte *) &icmp->unused.ip.destination;
    udp_Header *udp = (udp_Header *) icmp->unused.spares;

    printf( "ICMP: %s %d.%d.%d.%d port %d\n", msg, dest[0], dest[1], dest[2], dest[3], ntohs(udp->dstPort) );
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par

/*
 *
 */
static struct _pkt *icmp_Format( longword destip )
{
    eth_address dest;
    char *temp;

    /* we use arp rather than supplied hardware address */
    /* after first ping this will still be in cache */

    if ( !_arp_resolve( destip , &dest, 0 ))
        return( NULL );                         /* unable to find address */
    return( (struct _pkt*)_eth_formatpacket( &dest, 8 ));
}

/*
 * icmp_Reply - format a reply packet
 *            - note that src and dest are NETWORK order not host!!!!
 */
static void icmp_Reply( struct _pkt *p, longword src, longword dest, int icmp_length )
{
    in_Header *ip;
    icmp_pkt *icmp;

    ip = &p->in;
    memset( ip, 0, sizeof( in_Header ));
    icmp = &p->icmp;

    /* finish the icmp checksum portion */
    icmp->unused.checksum = 0;
    icmp->unused.checksum = ~checksum( icmp, icmp_length );

    /* encapsulate into a nice ip packet */
    ip->ver = 4;
    ip->hdrlen = 5;
    ip->length = htons( sizeof( in_Header ) + icmp_length);
    ip->tos = 0;
    ip->identification = htons( icmp_id ++);  /* not using ip id */
//    ip->frag = 0;
    ip->ttl = in_DefaultTTL;
    ip->proto = IPPROTO_ICMP;
    ip->checksum = 0;
    ip->source = src;
    ip->destination = dest;
    ip->checksum = ~ checksum( ip, sizeof( in_Header ));

    _eth_send( ntohs( ip->length ));
}

/* called from tcp_tick() in PCTCP.C
 * so it must have a C calling sequence and be external
 */
extern "C"
icmp_handler( in_Header *ip )
{
    icmp_pkt *icmp, *newicmp;
    struct _pkt *pkt;
    int len, code;
    in_Header *ret;
    PentiumClock *when;

    len = in_GetHdrlenBytes( ip );
    icmp = (icmp_pkt*)((byte *)ip + len);
    len = ntohs( ip->length ) - len;
    if ( checksum( icmp, len ) != 0xffff ) {
        outs("ICMP received with bad checksum\n\r");
        return( 1 );
    }

    code = icmp->unused.code;
    ret = & (icmp->ip.ip);

    switch ( icmp->unused.type) {
        case 0 : /* icmp echo reply received */
                /* icmp_print( icmp, "received icmp echo reply"); */

                /* check if we were waiting for it */
                ping_host_cache = ntohl( ip->source );
                ping_identifier_cache = icmp->echo.identifier;
                ping_sequence_cache = icmp->echo.sequence;
                when = (PentiumClock *) (1 + &icmp->echo);
                ping_delta_when_cache.set();
                ping_delta_when_cache -= *when;
                /* do more */
                break;

        case 3 : /* destination unreachable message */
                if (code < 6) {
                    icmp_print( icmp, unreach[ code ]);

                    /* handle udp or tcp socket */
                    if (ret->proto == IPPROTO_TCP)
                        _tcp_cancel( ret, 1, unreach[ code ], 0 );
                    if (ret->proto == IPPROTO_UDP)
                        _udp_cancel( ret );
                }
                break;

        case 4  : /* source quench */
                if (debug_on > 0 ) icmp_print( icmp, "Source Quench");
                if (ret->proto == IPPROTO_TCP)
                    _tcp_cancel( ret, 2, NULL, 0 );
                break;

        case 5  : /* redirect */
                if (code < 4) {
                    if (ret->proto == IPPROTO_TCP)
                        /* do it to some socket guy */
                        _tcp_cancel( ret, 5, NULL, icmp->ip.ipaddr );

                    if (debug_on > 0 ) icmp_print( icmp, redirect[ code ]);
                }
                break;

        case 8  : /* icmp echo request */
                /* icmp_print( icmp, "PING requested of us"); */

                // don't reply if the request was made by ourselves
                // such as a problem with Etherslip pktdrvr
                if  ( (longword) (ip->destination - htonl( my_ip_addr)) > multihomes )
                    return( 1 );

                // do arp and create packet
                /* format the packet with the request's hardware address */
                pkt = (struct _pkt*) _eth_formatpacket( _eth_hardware( (byte *)ip ), 8 );

                newicmp = &pkt->icmp;

                movmem( icmp, newicmp, len );
                newicmp->echo.type = 0;
                newicmp->echo.code = code;

                /* use supplied ip values in case we ever multi-home */
                /* note that ip values are still in network order */
                icmp_Reply( pkt,ip->destination, ip->source, len );

                /* icmp_print( newicmp, "PING reply sent"); */

                break;

        case 11 : /* time exceeded message */
                if (code < 2 ) {
                    icmp_print( icmp, exceed[ code ]);
                    if ((ret->proto == IPPROTO_TCP) && (code != 1))
                        _tcp_cancel( ret, 1, NULL, 0 );
                }
                break;

        case 12 : /* parameter problem message */
                icmp_print( icmp, "IP Parameter problem");
                break;

        case 13 : /* timestamp message */
                icmp_print( icmp, "Timestamp message");
                /* send reply */
                break;

        case 14 : /* timestamp reply */
                icmp_print( icmp, "Timestamp reply");
                /* should store */
                break;

        case 15 : /* info request */
                icmp_print( icmp,"Info requested");
                /* send reply */
                break;

        case 16 : /* info reply */
                icmp_print( icmp,"Info reply");
                break;

    }
    return( 1 );
}

/* you must be C++ code to call this
 */
int _ping( longword host, word identifier, word sequence )
{
    _pkt *p;
    eth_address dest;
    in_Header *ip;
    icmp_echo *icmp;
    PentiumClock *when;

    if ((host & 0xff) == 0xff ) {
        outs( "Cannot ping a network!\n\r");
        return( -1 );
    }
    if ( ! _arp_resolve( host, &dest, 0 )) {
        outs( "Cannot resolve host's hardware address\n\r");
        return( -1 );
    }

    if (debug_on) {
        outs("\n\rDEBUG: destination hardware :");
        outhexes( (byte *)&dest, 6 );
        outs("\n\r");
    }
    p = (struct _pkt*)_eth_formatpacket( &dest, 8 );

    ip = &p->in;
    memset( ip, 0, sizeof( in_Header ));
    icmp = &p->icmp.echo;
    when = (PentiumClock *) (1 + icmp);

    icmp->type = 8;
    icmp->code = 0;
    icmp->identifier = identifier;
    icmp->sequence = sequence;
    when->set();

    /* finish the icmp checksum portion */
    icmp->checksum = 0;
    icmp->checksum = ~checksum( icmp, sizeof(PentiumClock) + sizeof(struct icmp_echo));

    /* encapsulate into a nice ip packet */
    ip->ver = 4;
    ip->hdrlen = 5;
    ip->length = htons( sizeof(in_Header) + sizeof(struct icmp_echo) + sizeof(PentiumClock) );
    ip->tos = 0;
    ip->identification = htons( icmp_id ++);  /* not using ip id */
//    ip->frag = 0;
    ip->ttl = in_DefaultTTL;
    ip->proto = IPPROTO_ICMP;
    ip->checksum = 0;
    ip->source = htonl( my_ip_addr );
    ip->destination = htonl( host );
    ip->checksum = ~ checksum( ip, sizeof( in_Header ));

    return( _eth_send( ntohs( ip->length )));
}

